# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.
#

from abc import ABC
from abc import abstractmethod
from enum import Enum
from pathlib import Path
from iotcs.common import Closeable
from iotcs.exception import ArgumentException
from iotcs.exception import StateException

# oracle.iot.client

# TODO: Add the following when there is an enterprise pkg.
# The `AbstractVirtualDevice` API is common to both the enterprise
# client and the device client. The semantics of the API are also the
# same. The processing model on the enterprise client is different,
# however, from the processing model on the device client.
# On the enterprise client, an enterprise application calls the
# `AbstractVirtualDevice} set methods in order to affect a change on
# the physical device. The enterprise client application receives
# confirmation of that the change was accepted via a
# `ChangeCallback`.
#     The `ChangeCallback` indicates to the enterprise client that
# the value of the attribute on the device has changed. This callback may
# come as a result of the enterprise client calling `set` on the
# attribute, or it may come unsolicited as a result of the device-client
# updating the value from the physical device.
#
#     The enterprise client may also receive an
# `ErrorCallback` if the attribute value change was not accepted. This
# could be because the request timed out, the connection failed, or the
# change was explicitly rejected. The `ErrorCallback` indicates to
# the enterprise client that it should roll-back the attribute value to
# the last known value.


class AbstractVirtualDevice(ABC):
    """
    `AbstractVirtualDevice` is a representation of a device model.

    The device implements a ddevice model. A device model is a specification of
    the attributes, formats, and resources available on the device.

    The application calls :meth:`.set` on an `AbstractVirtualDevice` attribute
    and results in sending a message to the server. It calls
    :meth:`.get` to obtain the value of an attribute.

    The :class:`ErrorCallback <.AbstractVirtualDevice.ErrorCallback>` is
    invoked if there is an error sending the message to the server.
    See :meth:`setOnError <.AbstractVirtualDevice.setOnError>`.

    The :class:`.ChangeCallback` is invoked when the server receives a request
    to change a device attribute. It forwards the request to the device and the
    callback is invoked.
    See :meth:`setOnChange <.AbstractVirtualDevice.setOnChange>`.

    The `AbstractVirtualDevice` supports sending more than one message at
    time using the :meth:`.update` and :meth:`.finish` methods.
    """

    # TODO: make into an iterator
    class NamedValue(ABC):
        """
        This class encapsulates and `name/value` pair related to an event.

        In some cases the name is the name of an attribute, and value is the
        attribute's value and in some cases it is the name of a field in a
        format, or the name of an action, when an event is raised related to
        an alert or an action.
        """

        __slots__ = ()

        @abstractmethod
        def getName(self):
            """
            Return the name of the value.

            :return: the name of the value
            """
            pass

        @abstractmethod
        def getValue(self):
            """
            Return the value.

            :return: the value
            """
            pass

        @abstractmethod
        def next(self):
            """
            Return the next :class:`.NamedValue` in the list, or `None`.

            Return the next :class:`.NamedValue` in the list or `None` if there
            are no more in the list.
            :return: the next :class:`.NamedValue` instance or `None`
            """
            pass

    class Event(ABC):
        """
        The base class for an event.

        See :class:`.ChangeCallback`,
        :class:`.AbstractVirtualDevice.ErrorCallback`,
        :class:`.ChangeEvent`, :class:`.ErrorEvent`
        """

        __slots__ = ()

        @abstractmethod
        def getVirtualDevice(self):
            """
            Return the :class:`.VirtualDevice` instance.

            :return: the :class:`.VirtualDevice` instance, never `None`
            """
            pass

        @abstractmethod
        def getNamedValue(self):
            """
            Return the :class:`.NamedValue` instance.

            :return: the :class:`.NamedValue` instance, never `None`
            """
            pass

    class ChangeEvent(Event):
        """
        A Change Event.

        This class encapsulates information about an attribute change. An
        instance of this class is passed to :meth:`ChangeCallback.onChange`
        method. The :class:`.NamedValue` instance behaves like a linked list
        containing all the changed attributes and their values.

        The :class:`.VirtualDevice` instance is the virtual device whose
        attributes have changed.

        See :class:`.ChangeCallback`, :class:`.Event`
        """

        __slots__ = ()

        def getVirtualDevice(self):
            """
            Return the :class:`.VirtualDevice` instance.

            :return: the :class:`.VirtualDevice` instance, never `None`
            """
            pass

        def getNamedValue(self):
            """
            Return the :class:`.NamedValue` instance.

            :return: the :class:`.NamedValue` instance, never 'None`
            """
            pass

    class ErrorEvent(Event):
        """
        An Error event.

        The event instance that is passed to :meth:`.onError` to indicate an
        error has occurred when setting one or more attributes, or an error
        occurred raising an :class:`.Alert`.

        See :class:`ErrorCallback <.AbstractVirtualDevice.ErrorCallback>`,
        :class:`.Event`
        """

        __slots__ = ()

        @abstractmethod
        def getMessage(self):
            """
            Return the error message.

            :return: the error message
            """

    class ChangeCallback(ABC):
        """
        An attribute value change callback interface.

        This interface provides notification that the value
        of an attribute has changed.

        See :meth:`setOnChange <AbstractVirtualDevice.setOnChange>`
        """

        __slots__ = ()

        def onChange(self, event):
            """
            Invoke `onChange` when an attribute value changes.

            This method is invoked when the value of an attribute
            has changed. If an error occurs attempting to change the value,
            the new value is not set.

            :param event: the :class:`.ChangeEvent` change event
            """
            pass

    class ErrorCallback(ABC):
        """
        An error callback interface.

        This interface provides notification that an error when
        setting the value of an attribute.

        See :meth:`setOnError <.AbstractVirtualDevice.setOnError>`
        """

        __slots__ = ()

        def onError(self, event):
            """
            Invoke `onError` when there is an error.

            This method is invoked when an error occurs setting the value of an
            attribute or raising an :class:`.Alert`.
            :param event: The :class:`.ErrorEvent` error event, never `None`
            """
            pass

    __slots__ = ()

    # TODO: Does this return DeviceModelImpl ?
    @abstractmethod
    def getDeviceModel(self):
        """
        Return the :class:`.DeviceModel` instance for this device object.

        :return: the device model :class:`.DeviceModel`
        """
        pass

    @abstractmethod
    def getEndpointId(self):
        """
        Return the endpoint id of the device.

        The endpoint id is the identifier that the server creates when a
        registered device is activated.

        :return: the device's endpoint id
        """

    # TODO: Add the following to the get doc when there is an enterprise pkg
    # For an enterprise client, no REST API call is made to the server.

    @abstractmethod
    def get(self, attributeName):
        """
        Return the value of `attributeName`.

        This method returns the current value held by the virtual device model
        and no call is made to the physical device.

        :param attributeName: the name of an attribute from the device type
            model
        :return: the attribute value
        :raises `ValueError`: if `attributeName` is not in the model,
            or `attributeName` is `None`
        """
        pass

    # TODO: Add the following to getLastKnown doc when there is an enterprise
    # TODO: package
    # In the case where the enterprise client sets the value of an attribute,
    # the last known value is the value before an
    # `setOnChange(ChangeCallback)` `onChange` event is received for the
    # attribute. In the event the attribute value
    # could not be set, the current value is rolled back to the last known
    # value.

    @abstractmethod
    def getLastKnown(self, attributeName):
        """
        Return the last known value of `attributeName`.

        The last known value is always the same as the
        current value. The last known value is maintained by the
        :class:`.AbstractVirtualDevice` instance.
        No REST API call is made to the server.

        :param attributeName: the name of an attribute from the device type
            model
        :return: the attribute value
        :raises `ValueError`: if `attributeName` is not in the model,
            or `attributeName` is `None`
        """
        pass

    # TODO: Add the following to set doc when there is an enterprise package
    # For an enterprise client, an endpoint resource REST API call is made
    # to the server.

    # TODO: Note that "iotcs.shared.message.DATA" is not part of the
    # TODO: documentation set.

    @abstractmethod
    def set(self, attributeName, value):
        """
        Set the value of `attributeName` to `value`.

        This method is used by the application to synchronize the
        :class:`.AbstractVirtualDevice` state with the physical device and
        results in an endpoint :class:`.DATA` message REST API call to the
        server.

        `value` is validated according to the constraints in the
        :class:`.DeviceModel`. If `value` is not valid, a
        `ValueError` exception is raised.

        :param attributeName: the name of an attribute from the device type
            model
        :param value: the value to set
        :return: this AbstractVirtualDevice instance
        :raises `ValueError`: if `value` is not valid,
            `attributeName` is not in the model, or `attributeName` is `None`.
            or if `attributeName` is not writable.
        """
        pass

    @abstractmethod
    def update(self):
        """
        Set a mark for beginning an update transaction.

        If the `update` method had not been called and there is no open update
        transaction, calls to :meth:`.set` sends the new value to the server.
        The `update` method opens an update transaction and future calls to
        :meth:`.set` do not immediatelly send the new value to the server,
        but accumulates the changes to be sent in one call to the server.

        The accumulated changes in an open update transaction are sent to the
        server when the :meth:`.finish` method is called. :meth:`.finish` sends
        the updated values to the server and closes the open update
        transaction.

        For example::

            virtualDevice.update().set("min", 10).set("max", 20).finish()

        See :meth:`.finish`

        :return: this `AbstractVirtualDevice` instance
        """
        pass

    @abstractmethod
    def finish(self):
        """
        Send the values in an update transaction to the server.

        The :meth:`.update` method must have been called before calling
        `finish`. The update transaction mark is cleared.

        See :meth:`.update()`
        """
        pass

    @abstractmethod
    def setOnChange(self, callback, attributeName=None):
        """
        Set a :class:`.ChangeCallback` instance.

        If `attributeName` is `None`, `callback` is invoked when
        the value of any attribute changes.

        If `attributeName` is not None, `callback` is invoked when the value
        of `attributeName` changes.

        If `callback` is `None`, the callback instance is removed and no
        callback is invoked for attribute value changes.

        :param attributeName: the name of an attribute from the device model
        :param callback: a :class:`.ChangeCallback` instance
        :raises `ValueError`: if `attributeName` is not in the model,
            or `attributeName` is `None`
        """
        pass

    @abstractmethod
    def setOnError(self, callback, attributeName=None):
        """
        Set a :class:`ErrorCallback <.AbstractVirtualDevice.ErrorCallback>` instance.

        If `attributeName` is `None`, `callback` is invoked if an
        error occurs when setting any attribute or an error occurs
        when raising an :class:`.Alert`.

        If `attributeName` is not None, `callback` is invoked if an error
        occurs when setting the value for `attributeName`
        or an error occurs when raising an :class:`.Alert`.

        If `callback` is `None`, the callback instance is removed and no
        callback is invoked if an error occurs.

        :param attributeName: the name of an attribute from the device model
        :param callback: a
            :class:`ErrorCallback <.AbstractVirtualDevice.ErrorCallback>`
            instance
        :raises `ValueError`: if `attributeName` is not in the model,
            or `attributeName` is `None`
        """
        pass


# TODO: Add the following to the class doc when there is an enterprise pkg
# TODO: or an enterprise application.

class Client(Closeable):
    """
    Client of the Oracle IoT Cloud Service.

    A client is a directly-connected device, or a gateway.

    See :class:`.DirectlyConnectedDevice`, :class:`.GatewayDevice`
    """

    @abstractmethod
    def createVirtualDevice(self, deviceId, deviceModel):
        """
        Create an :class:`.AbstractVirtualDevice` instance.

        A instance of :class:`.AbstractVirtualDevice` is created with the
        the given :class:`.DeviceModel` instance and the given device
        identifier. This method creates a new :class:`.VirtualDevice` instance
        with the given parameters. The client library does not cache created
        VirtualDevice instances.

        :param deviceId: the device identifier of the device being modeled
        :param deviceModel: an instance of :class:`.DeviceModel` that
            this device implements.
        :raises `ValueError`: if the device does not implement the model,
            if `deviceId` or `deviceModel` are `None`.
        :return: a :class:`.VirtualDevice` instance
        """
        pass

    @abstractmethod
    def getDeviceModel(self, deviceModelUrn):
        """
        Return the :class:`.DeviceModel` for `deviceModelUrn`.

        This method returns None if there is no device model identified by
        `deviceModelUrn`. This method will also return `None` if the model is a
        *draft model* and the :mod:`.config` property
        *allow_draft_device_models* in section *device_model* is set to `False`
        (the default).

        :param deviceModelUrn: the URN of the device model
        :raises `ValueError`: if `deviceModelUrn` is `None` or there is an I/O
            error when communicating with the server, or when key or signature
            algorithm class cannot be loaded, or the key is not in the trusted
            assets store, or the private key is invalid
        :return: a :class:`.DeviceModel` instance or `None`
        """
        pass

    @abstractmethod
    def createStorageObject(self, name, contentType):
        """
        Create a new `StorageObject` with `name` and `contentType`.

        If `contentType` is `None`, the mime-type defaults to
        "application/octet-stream".

        :param name: the unique name to be used to reference the content in
            storage
        :param contentType: The mime-type of the content
        :return: a :class:`.StorageObject`
        :raises: :class:`.TransportException` if there is an Exception raised
            by the runtime, or an abnormal response from the storage cloud
        :raises: :class:`.SecurityException` if there is an exception
            establishing a secure connection to the storage cloud
        """
        pass

    @abstractmethod
    def createStorageObject(self, name, contentType):
        """

        Create a new :class:`.StorageObject`. with `name` and `contentType`.

                If `contentType` is `None`, the mime-type defaults to
        "application/octet-stream".

        :param name: the unique name to be used to reference the content in
            storage
        :param contentType: The mime-type of the content
        :return: a :class:`.StorageObject`
        :raises: :class:`.TransportException` if there is an Exception raised
            by the runtime, or an abnormal response from the storage cloud
        :raises: :class:`.SecurityException` if there is an exception
            establishing a secure connection to the storage cloud
        """
        pass

    def __init__(self):
        """Create a new `Client` instance."""
        Closeable.__init__(self)


class ExternalObject():
    """
    ExternalObject represents the value of a URI type in a device model.

    The application is responsible for uploading/downloading the content
    referred to by the URI.
    """

    __slots__ = (
        '__uri'
    )

    def __init__(self, uri):
        """
        Create an `ExternalObject`.

        :param uri: The URI
        """
        self.__uri = uri

    def getURI(self):
        """
        Get the URI value.

        :return: the URI
        """
        return self.__uri


class DeviceModel(ABC):
    """
    Detailed information on a device model.

    A device model is a specification
    of the attributes, formats, and resources available on the device.
    """

    __slots__ = ()

    @abstractmethod
    def getURN(self):
        """
        Return the URN of the device model.

        :return: the URN of the model
        """
        pass

    @abstractmethod
    def getName(self):
        """
        Return the name of the device model.

        :return: the device model name
        """
        pass

    @abstractmethod
    def getDescription(self):
        """
        Return the free form description of the device model.

        :return: the description of the model
        """
        pass


class StorageObject(ExternalObject):
    """
    StorageObject provides information about content in the cloud storage.
    
    Implementation for `<.StorageObject>`.
    
    .. code::python
    
        # Upload example
        lenna = directlyConnectedDevice.createStorageObject("lenna.jpg",
                                                            "image/jpeg")
        lenna.setInputPath("../images/lenna.jpg")
    
        # onSync is called when the content referenced by the storageObject
        # is in sync with the storage cloud, or the sync has failed.
    
        class UploadSyncCallback(SyncCallback):
            def __init__(self):
                super(UploadSyncCallback, self).__init__()
    
            def onSync(self, event):
                storageObject = event.getSource()
                if storageObject.getSyncStatus() ==
                        StorageObject.SyncStatus.IN_SYNC:
                    # image was uploaded and can be deleted
                elif storageObject.getSyncStatus() ==
                        StorageObject.SyncStatus.SYNC_FAILED):
                    # image was not uploaded, take action!
    
        lenna.setOnSync(UploadSyncCallback())
        virtualDevice.set("image", lenna);
    
        # Download example
        # onChange is called when the attribute value changes.
    
        class DownloadSyncCallback(SyncCallback):
            def __init__(self):
                super(DownloadSyncCallback, self).__init__()
    
            def onSync(self, event):
                storageObject = event.getSource()
                if storageObject.getSyncStatus() ==
                        StorageObject.SyncStatus.IN_SYNC:
                    # image was download and can be used
                elif storageObject.getSyncStatus() ==
                        StorageObject.SyncStatus.SYNC_FAILED):
                    # image was not was not downloaded, take action!
    
    
        class OnChangeCallback(VirtualDevice.ChangeCallback):
            def __init__(self):
                super(OnChangeCallback, self).__init__()
    
            def onChange(event):
                namedValue = event.getNamedValue()
                storageObject = namedValue.getValue()
                if storageObject.getContentLength() < availableDiskSpace:
                    # syncTo will kick off the async download of the content
                    storageObject.setOnSync(DownloadSyncCallback());
                        StorageObject storageObject = event.getSource();
                        if storageObject.getSyncStatus() == SyncStatus.IN_SYNC:
                            # image was downloaded and can now be used
                        elif storageObject.getSyncStatus() ==
                                SyncStatus.SYNC_FAILED:
                            # image was not downloaded, take action!
    
                    storageObject.setOutputPath("../downloads" +
                                                storageObject.getName())
                    storageObject.sync()
    
    
        virtualdevice.setOnChange(OnChangeCallback())
    """

    class SyncStatus(Enum):
        """The storage cloud content status."""

        IN_SYNC = 'IN_SYNC'
        """The upload or download is complete."""

        SYNC_FAILED = 'SYNC_FAILED'
        """The upload or download failed."""

        NOT_IN_SYNC = 'NOT_IN_SYNC'
        """The content has not been uploaded or downloaded."""

        SYNC_PENDING = 'SYNC_PENDING'
        """An upload or download is pending."""

    class SyncEvent(object):
        """
        An event passed to the :meth:`.setOnSync` :class:`.SyncCallback`.

        This event occurs when an attribute value's content refers to
        a :class:`.StorageObject` and has been successfully
        synchronized, or has failed to be synchronized.
        """

        @abstractmethod
        def getVirtualDevice(self):
            """
            Return the virtual device that is the source of the event.

            :return: the virtual device, or `None` if :meth:`.sync` was
                called independently
            """
            pass

        @abstractmethod
        def getName(self):
            """
            Return the name of the attribute, action, or format.

            Return the name of the attribute, action, or format that
            this event is associated with.

            :return: the name, or `None` if :meth:`.sync` was called
                independently
            """
            pass

        @abstractmethod
        def getSource(self):
            """
            Return the :class:`.StorageObject` source of this event.

            :return the storageObject, never `None`
            """
            pass

    class SyncCallback(object):
        """
        A callback interface for receiving a :class:`.SyncEvent`.

        This callback is invoked when an attribute value refers to
        a :class:`.StorageObject` and has been successfully synchronized
        or has failed to be synchronized with the cloud storage service.
        See :meth:`setOnSync <.setOnSync>`
        """

        @abstractmethod
        def onSync(self, event):
            """
            Call when content is synchronized.

            This method is called when a :class:`.StorageObject` attribute
            value has been successfully synchronized, or has failed
            to be synchronized.

            This method never be invoked by the client library with a
            `None` argument.
            :param event: The synchronization event
            """
            pass

    __slots__ = (
        'delegate',
        'syncStatus',
        'syncCallback',
        'inputPath',
        'outputPath'
    )

    def __init__(self, delegate):
        """
        Create a :class:`.StorageObject`.

        :param delegate: the StorageObject delegate implementation.
        """
        super(StorageObject, self).__init__(delegate.getURI())
        self.delegate = delegate
        self.syncStatus = StorageObject.SyncStatus.NOT_IN_SYNC
        self.syncCallback = None
        self.inputPath = None
        self.outputPath = None

    def getName(self):
        """
        Return the the name of this object in the storage cloud.

        It is the name and path of the file that was uploaded to the
        storage cloud.

        :return the name of this object in the storage cloud
        """
        return self.delegate.getName()

    def getContentType(self):
        """
        Return the mime-type of the content.

        See `IANA Media Types <http://www.iana.org/assignments/media-types/media-types.xhtml>`_ # noqa: E501
        :return: The mime-type of the content
        """
        return self.delegate.getContentType()

    def getEncoding(self):
        """
        Return the compression scheme of the content.

        :return: the compression scheme of the content, or `None` if
            the content is not compressed
        """
        return self.delegate.getEncoding()

    def getLength(self):
        """
        Return the length of the content in bytes.

        This is the number of bytes required to upload or download the content.

        :return: The length of the content, or -1 if unknown
        """
        return self.delegate.getLength()

    def getDate(self):
        """
        Return the date and time the content was created or last modified.

        Return the date and time the content was created or last modified in
        the cloud storage in ISO 8601 format. This may be `None` if the
        content has not been uploaded.

        :return: The date the content was last modified in cloud storage
            in ISO 8601 format, or `None` if the content has not been
            uploaded.
        """
        return self.delegate.getDate()

    def getSyncStatus(self):
        """
        Return the :class:`.SyncStatus` of this :class:`.StorageObject`.

        :return:  the :class:`.SyncStatus` of this :class:`.StorageObject`.
        """
        return self.syncStatus

    def getInputPath(self):
        """
        Return  the input path.

        :return the input path, or `None` if not set
        """
        return self.inputPath

    def setInputPath(self, inputPath):
        """
        Set the input path for uploading content to the storage cloud.

        The implementation allows for either the input path to be set, or
        the output path to be set, but not both.
        If `inputPath` is not `None`, the output path will be set to `None`.
        If `inputPath` is not `None` and does not equal the current input path,
        `syncStatus` will be reset to
        :attr:`NOT_IN_SYNC <.SyncStatus.NOT_IN_SYNC>`
        This method raises `ValueError` if :meth:`.getSyncStatus` returns
        :attr:`SYNC_PENDING <.SyncStatus.SYNC_PENDING>`.

        :param inputPath: the path from which to read the content for upload
        :raises OSError: if the input path cannot be read
        :raises: :class:`.StateException` if called when sync
            status is :attr:`SYNC_PENDING <.SyncStatus.SYNC_PENDING>`.
        """
        if self.syncStatus == self.SyncStatus.SYNC_PENDING:
            raise ValueError("sync pending")
        if self.inputPath is None or not self.inputPath == inputPath:
            self.inputPath = inputPath
            if self.inputPath is not None:
                self.outputPath = None
                self.syncStatus = self.SyncStatus.NOT_IN_SYNC
                # Use pathib.Path objects for "stream" memberts
                # TODO: Proably should check "readable"
                path = Path(self.inputPath)
                if not path.exists() or not path.is_file():
                    raise OSError(f'StorageObject.setInputPath(): '
                                  f'File {inputPath} does not exist')
                self.delegate.setInputStream(path)

    def getOutputPath(self):
        """
        Return the output path.

        :return: the output path, or `None` if not set
        """
        return self.outputPath

    def setOutputPath(self, outputPath):
        """
        Set the output path for downloading content from the storage cloud.

        The implementation allows for either the output path to be set, or
        the input path to be set, but not both.
        If `outputPath` is not `None`, the input path will be set to `None`.
        If `outputPath` is not `None` and does not equal the current output

        `syncStatus` will be reset to
        :attr:`NOT_IN_SYNC <.SyncStatus.NOT_IN_SYNC>`.
        This method raises `ValueError` if :meth:`.getSyncStatus` returns
        :attr:`SYNC_PENDING <.SyncStatus.SYNC_PENDING>`.

        :param outputPath: the path to which the content will be downloaded.
            If the path does not already exist, it will be created.
        :raises OSError: if the output path cannot be written
        :raises: :class:`.StateException` if called when sync
            status is :attr:`SYNC_PENDING <.SyncStatus.SYNC_PENDING>`.
        """
        if self.syncStatus == self.SyncStatus.SYNC_PENDING:
            raise ValueError("sync pending")
        if self.outputPath is None or not self.outputPath == outputPath:
            self.outputPath = outputPath
            if self.outputPath is not None and len(self.outputPath) != 0:
                self.inputPath = None
                self.syncStatus = self.SyncStatus.NOT_IN_SYNC

                path = Path(self.outputPath)
                if path.exists():
                    raise OSError(f'StorageObject.setOutputPath: '
                                  f'File {outputPath} exists.')

                parent = path.parent
                if parent != '.':
                    try:
                        parent.makedir(parents=True, exist_ok=True)
                    except Exception as e:
                        raise OSError(f'StorageObject.setOutputPath: '
                            f'Cannot create path {parent} for {outputPath}, {e}')  # noqa: E128,E501

                # Use pathib.Path objects for "stream" members
                self.delegate.setOutputStream(path)

    def setOnSync(self, callback):
        """
        Set a :class:`.StorageObject.SyncCallback`.

        The callback is invoked when the content referred to by this
        :class:`.StorageObject` is synchronized.

        :param callback: a :class:`.StorageObject.SyncCallback` to invoke when
            there is an error setting a value, if `None`, the existing
            `syncCallback` is removed.
        """
        self.syncCallback = callback

    def __str__(self):
        """Return the str representation of the `StorageObject`."""
        return self.getURI()

    def setSyncStatus(self, syncStatus):
        """
        Set the syncStatus.

        Used internally by the client library. Not intended for general use.
        :param syncStatus: the :class:`.StorageObject.SyncStatus`
        """
        self.syncStatus = syncStatus

    def getSyncCallback(self):
        """
        Return `syncCallback`.

        Used internally by the client library. Not intended for general
        use.
        :return: the SyncCallback
        """
        return self.syncCallback

    def setCustomMetadata(self, key, value):
        """
        Sets the metadata for the StorageObject.

        All metdata will be added to Storage Cloud Service as custom metadata
        with the *X-Object-Meta-KeyName* header.
        :param key: The metadata key
        :param value: The metadata value
        :raises: :class:`.Argument` if the key or value is `None` or empty
        """
        self.delegate.setCustomMetadata(key, value)

    def getCustomMetadata(self):
        """
        This method return unmodifiable copy of metadata.
        :return: a `dict` of metadata. It may be empty.
        """
        return self.delegate.getCustomMetadata()

    @abstractmethod
    def sync(self):
        """
        Notify the library to sync content with the storage cloud.

        The direction of the sync, upload or download, depends on whether
        :meth:`inputPath <.setInputPath>` or the
        :meth:`outputPath <.setOutputPath>` have been set.
        This method does not start any uploads or downloads if
        :meth:`sycn status <.getSyncStatus>` is other than
        :attr:`NOT_IN_SYNC <.SyncStatus.NOT_IN_SYNC>`.

        This is a non-blocking call. The sync is performed on a separate
        thread on this :class:`.StorageObject`.

        If the input path cannot be read, or the output path cannot be
        written, an `OSError` is raised. Any I/O exceptions
        during the background sync are reported through the virtual
        device :class:`error callback <.AbstractVirtualDevice.ErrorCallback>`
        of the virtual device.

        :raises: :class:`.StateException` if both input path
            and output path are `None`
        """
        pass
