# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and 
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.
#

from abc import ABC
from abc import abstractmethod

import datetime
from enum import Enum
import json
import logging
import numbers
import os
import sys
from urllib.parse import quote

from iotcs.client import DeviceModel
from iotcs.client import ExternalObject
from iotcs.exception import ArgumentException
from iotcs.exception import StateException
from iotcs.common import Filter
from iotcs.common import RestApi
from iotcs.common import StatusCode
from iotcs.common import Util


class DeviceModelAttribute():
    """
    DeviceModelAttribute is the model of an attribute in a `DeviceModel`.
    """


    class Access(Enum):
        """
        Access rules for an attribute.
            - READ_ONLY The attribute value can only be read.
            - WRITE_ONLY The attribute value can only be set.
            - READ_WRITE The attribute value can be set and can be read
            - EXECUTABLE The attribute is a action
        """
        READ_ONLY = 0
        WRITE_ONLY = 1
        READ_WRITE = 2
        EXECUTABLE = 3

    # TODO: What exception is raised if constructed from a String that does not
    # TODO: match ?
    class Type(Enum):
        """
        An enumeration of the data types an attribute may contain.
        """
        NUMBER = 0
        STRING = 1
        URI = 2
        BOOLEAN = 3
        INTEGER = 4
        DATETIME = 5

        @classmethod
        def cast(cls, attributeType, value, convert=False):
            """
            Return value as a primitive type for the `dmaType`,
            :param dmaType: the value `DeviceModelAttribute.Type`
            :param value: the value to cast to the appropriate primitive
            :param convert: if `True` convert the value if possible to the
            expected type.
            :return: value as a primitive type for the `dmaType`
            :raises: :class:`.ArgumentError` if `dmaType` is unknown or
                `value` cannot be cast to the expected primitive.
            """
            if attributeType is None or value is None:
                if convert is True:
                    return None
                raise ArgumentException(f'dmaType or value must not be None')

            castvalue = value
            exception = None
            expectedtype = None
            try:
                if attributeType == DeviceModelAttribute.Type.BOOLEAN:
                    if not isinstance(value, (bool,str,)):
                        if convert is True:
                            if isinstance(value, numbers.Number):
                                return False if value == 0 else True
                            else:
                                return False
                        expectedtype = bool.__name__
                    else:
                        castvalue = bool(value)
                elif attributeType == DeviceModelAttribute.Type.INTEGER:
                    if not isinstance(value, (int,str,)):
                        if convert is True:
                            if isinstance(value, numbers.Number):
                                return int(value)
                        expectedtype = int.__name__
                    else:
                        castvalue = int(value)
                elif attributeType == DeviceModelAttribute.Type.NUMBER:
                    if not isinstance(value, (numbers.Number,str)):
                        if convert is True:
                            return None
                        expectedtype = f'{int.__name__} or {float.__name__}'
                    else:
                        castvalue = value
                elif attributeType == DeviceModelAttribute.Type.DATETIME:
                    if not isinstance(value, (int,datetime.datetime,)):
                        if convert is True:
                            if isinstance(value, numbers.Number):
                                return int(value)
                        expectedtype = f'{int.__name__} or {datetime.datetime.__name__}'
                    else:
                        if isinstance(value, datetime.datetime):
                            ts = value.timestamp() * 1000
                            castvalue = int(ts)
                        else:
                            castvalue = int(value)
                elif attributeType == DeviceModelAttribute.Type.STRING:
                    if not isinstance(value, (str,)):
                        if convert is True:
                            try:
                                return str(value)
                            except:
                                return None
                        expectedtype = str.__name__
                    else:
                        castvalue = str(value)
                elif attributeType == DeviceModelAttribute.Type.URI:
                    if not isinstance(value,(str,)):
                        if convert is True:
                            return None
                        expectedtype = str.__name__
                    else:
                        castvalue = value
                else:
                    exception = \
                        ArgumentException(f'cast(): unknown attribute type '
                            f'{str(attributeType)}')
            except Exception as e:
                if convert is True:
                    return None
                raise ArgumentException(
                    f'cast(): value {value} for {attributeType.name} '
                    f'is not valid: {e}')
            if convert is False:
                if exception is not None:
                    raise exception

                if expectedtype is not None:
                    raise ArgumentException(f'cast(): value for '
                        f'{attributeType.name} is not {expectedtype}')

            return castvalue

    __slots__ = (
        '__name',
        '__urn',
        '__description',
        '__type',
        '__lowerBound',
        '__upperBound',
        '__access',
        '__alias',
        '__defaultValue'
    )

    def __init__(self, urn, name, description, type, lowerBound, upperBound,
                 access, alias, defaultValue=None):
        self.__urn = urn
        self.__name = name
        self.__description = description
        self.__type = type
        self.__lowerBound = lowerBound
        self.__upperBound = upperBound
        self.__access = access
        self.__alias = alias
        self.__defaultValue = defaultValue

    def getName(self):
        """
        Get the attribute name.
        :return: the attribute name from the device model
        """
        return self.__name

    def getUrn(self):
        """
        Get the URN of the device type model this attribute comes from.
        :return: the URN of the device type model
        """
        return self.__urn

    def getDescription(self):
        """
        A human friendly description of the attribute. If the model does not
        contain a description, this method will return an empty String.
        :return: the attribute description, or an empty string
        """
        return self.__description

    def getType(self):
        """
        The data type of the attribute. If the access type of the attribute is
        executable, this method will return `None`.
        :return: the attribute's data type, or `None`
        """
        return self.__type

    def getLowerBound(self):
        """
        For `Type.NUMBER` only, give the lower bound of the
        acceptable range of values. `None` is always returned for attributes
        other than `NUMBER` type.
        :return: a number, or `None` if no lower bound has been set
        """
        return self.__lowerBound

    def getUpperBound(self):
        """
        For ` Type.NUMBER` only, give the upper bound of the
        acceptable range of values. Null is always returned for attributes
        other than `NUMBER` type.
        :return: a number, or `None` if no upper bound has been set
        """
        return self.__upperBound

    def getAccess(self):
        """
        Return the access rules for the attribute. The default is READ-ONLY
        :return: the access rule for the attribute
        """
        return self.__access

    def getDefaultValue(self):
        """
        Get the default value of the attribute as defined by the device model.
        If there is no `defaultValue` for the attribute in the device model,
        then this method will return `None`. The value `None` is *not* a
        default value.
        :return: the default value of the attribute, or `None` if no default
                is defined
        """
        return self.__defaultValue

    # If expectedType is not nil then the value must conform to this type.
    # In the case of URI it might be "ExternalObject" and in the case of
    # DATETIME it might be Date or Int.
    def validate(self, value):
        result = False
        if self is DeviceModelAttribute.Type.BOOLEAN:
            result = isinstance(value, bool)
        elif self is DeviceModelAttribute.Type.NUMBER:
            result = isinstance(value, numbers.Number)
        elif self is DeviceModelAttribute.Type.DATETIME:
            result = isinstance(value, (datetime.datetime, numbers.Number,))
        elif self is DeviceModelAttribute.Type.INTEGER:
            result = isinstance(value, numbers.Number) and isinstance(value, int)
        elif self is DeviceModelAttribute.Type.STRING:
            result = isinstance(value, str)
        elif self is DeviceModelAttribute.Type.URI:
            result = isinstance(value, (str, ExternalObject,))

        return result



class DeviceModelImpl(DeviceModel):
    """The implementation of :class:`.DeviceModel`."""

    __slots__ = (
        '__urn',
        '__name',
        '__description',
        '__attributeMap',
        '__actionMap',
        '__formatMap'
    )

    def __init__(self, urn, name, description, attributes, actions, formats):
        """
        Create a `DeviceModel` instance.

        :param urn: the URN that identifies a `DeviceModel`
        :param name: the name for this `DeviceModel`
        :param description: the human readable description for this
            `DeviceModel`
        :param attributes: a `list` of :class:`.DeviceModelAttribute`s for
            this `DeviceModel`
        :param actions: a `list` of :class:`.DeviceModelAction`s for
            this `DeviceModel`
        :param formats: a `list` of :class:`.DeviceModelFormat`s for this
            `DeviceModel`
        """
        super(DeviceModelImpl, self).__init__()

        self.__urn = urn
        self.__name = name
        self.__description = description
        self.__attributeMap = dict()
        self.__actionMap = dict()
        self.__formatMap = dict()

        if attributes is not None:
            self.__attributeMap = dict()
            for attribute in attributes:
                self.__attributeMap[attribute.getName()] = attribute

        if actions is not None:
            self.__actionMap = dict()
            for action in actions:
                self.__actionMap[action.getName()] = action

        if formats is not None:
            self.__formatMap = dict()
            for formaturn in formats:
                self.__formatMap[formaturn.getURN()] = formaturn

    def getDeviceModelFormat(self, urn):
        """
        Return the `DeviceModelFormat` for `urn`.

        :param urn: the urn that identifies a `DeviceModelFormat` of this
            `DeviceModel`
        :return: the `DeviceModelFormat` for `urn`
        """
        if urn in self.__formatMap:
            return self.__formatMap[urn]
        else:
            raise ValueError("{} does not exist in device model".format(urn))

    def getURN(self):
        """
        Return the URN that identified this `DeviceModel`.

        :return: the URN that identifies this `DeviceModel`
        """
        return self.__urn

    def getName(self):
        """
        Return the name of this `DeviceModel`.

        :return: the name of this `DeviceModel`
        """
        return self.__name

    def getDescription(self):
        """
        Return the human readable description of this `DeviceModel`.

        :return: the human readable description of this `DeviceModel`.
        """
        return self.__description

    def getDeviceModelAttributes(self):
        """
        Return the `DeviceModelAttribute`s of this :class:`.DeviceModel`.

        :return: a `dict` of *{attributename: :class:`.DeviceModelAttribute`}*
        """
        return self.__attributeMap

    def getDeviceModelActions(self):
        """
        Return the :class:`.DeviceModelAction`s of this :class:`.DeviceModel`.

        :return: a `dict` of *{actionname: :class:`.DeviceModelAction`}*
        """
        return self.__actionMap

    def getDeviceModelFormats(self):
        """
        Return the :class:`.DeviceModelFormat`'s of this :class:`.DeviceModel`.

        :return: a `dict` of *{formaturn: :class:`.DeviceModelFormat`}*
        """
        return self.__formatMap

    def __str__(self):
        """Return a string representation of :class:`.DeviceModel`."""
        deviceModelString = f'\
            urn =\t{self.__urn},\n \
            \tname = {self.__name},\n \
            \tdescription = {self.__description},\n \
            \tattributes = ['

        comma = ""
        attrsString = ""
        for (name, attribute) in self.__attributeMap:

            s = f' \n\t{{ {attribute.__str__()} }}'
            attrsString = f'{attrsString}{comma}{s}'
            comma = ","
        if attrsString != "":
            deviceModelString = f'{deviceModelString}{attrsString}\n\t'

        deviceModelString = f'{deviceModelString}],\n\tactions = ['
        actionString = ""
        comma = ""
        for (name, action) in self.__actionMap:
            s = f'\n\t{{ {action.__str__()} }}'
            actionString = f'{actionString}{comma}{s}'
            comma = ","
        if actionString != "":
            deviceModelString = f'{deviceModelString}{actionString}\n\t'

        deviceModelString = f'{deviceModelString}],\n\tformats = ['
        comma = ""
        formatsString = ""
        for (urn, format) in self.__formatMap:
            s = f'\n\t{{ {format.__str__()} }}'
            formatsString = f'{formatsString}{comma}{s}'
            comma = ","
        deviceModelString = f'{deviceModelString}{formatsString}\n\t]'

        return deviceModelString


# TODO: Should this be in "impl" ? JCL is iot.client.DeviceModelFormat
class DeviceModelFormat():
    """`DeviceModelFormat` is the model of a message in a `DeviceModel`."""

    class Type(Enum):
        """
        The message type.

        Each type corresponds to one type of message.
        """

        DATA = 1,
        ALERT = 2

        def __int__(self):
            """Return the `int` value of `DeviceModelFormat.Type`."""
            return self.value

    class Field():
        """Describes a field of a message."""

        __slots__ = (
            '__type',
            '__name',
            '__description',
            '__optional'
        )

        def __init__(self, name, description, strType, optional):
            """
            Create a `DeviceModelFormat.Field`.

            :param name:  the `Field` name
            :param description:  the `Field` description
            :param strType: the string name of a `DeviceModelAttribute.Type`
            :param optional: if `False` the field must exist
            """
            # TODO: should this catch and throw ValueError if strType is
            # TODO: invalid ?
            self.__type = DeviceModelAttribute.Type[strType]
            self.__name = name
            self.__description = description
            self.__optional = optional

        def getType(self):
            """
            Return the :class:`.DeviceModelAttribute.Type` `Field` type.

            :return: the :class:`.DeviceModelAttribute.Type` `Field` type
            """
            return self.__type

        def getName(self):
            """
            Return the `Field` name.

            :return: the `Field` name
            """
            return self.__name

        def isOptional(self):
            """
            Return `True` if this `Field` is not required.

            :return: `True` if this `Field` is not required else `False`.
            """
            return self.__optional

        def getDescription(self):
            """
            Return the human readable description of this `Field`.

            :return: the human readable description of this `Field`.
            """
            return self.__description

        def __str__(self):
            """Return a string representation of `DeviceModelFormat.Field`."""
            return str(f'name = {self.__name} \
                , description = {self.__description} \
                , type = {self.__type} \
                , optional = {self.__optional}')

    __slots__ = (
        '__type',
        '__description',
        '__name',
        '__urn',
        '__fields'
    )

    # def __init__(self, jsonFormat):
    #     #print(json.dumps(format))
    #     type = jsonFormat["type"]
    #     if type == "ALERT":
    #         self.__type = DeviceModelFormat.Type.ALERT
    #     elif (type == "DATA"):
    #         self.__type = DeviceModelFormat.Type.DATA
    #     else:
    #         raise ValueError("Device model format type invalid:" + type)
    #
    #     self.__description = jsonFormat["description"]
    #     self.__deviceModel = jsonFormat["deviceModel"]
    #     self.__name = jsonFormat["name"]
    #     self.__sourceId = jsonFormat["sourceId"]
    #     self.__sourceType = jsonFormat["sourceType"]
    #     self.__urn = jsonFormat["urn"]
    #
    #     fields = jsonFormat["value"]["fields"]
    #     self.__fieldMap = dict()
    #     for fieldData in fields:
    #         fieldName = fieldData["name"]
    #         field = DeviceModelField(fieldData)
    #         self.__fieldMap[fieldName] = field

    def __init__(self, urn, name, description, formatType, fields):
        """
        Create an instance of `DeviceModelFormat`.

        :param urn: the urn of the message format
        :param name: the message format name
        :param description: a message format human readable description
        :param type: the `DeviceModelFormat.Type`
        :param fields: the fields of this message format.
        :raises: `KeyError` if `type` is not a valid `DeviceModelFormat.Type`
        """
        self.__name = name
        self.__urn = urn
        self.__description = description
        self.__type = formatType
        self.__fields = fields

    def getType(self):
        """
        Return the :class:.`DeviceModelFormat.Type` of the message.

        :eturn: the :class:.`DeviceModelFormat.Type` of the message
        """
        return self.__type

    def getName(self):
        """
        Return the message format name.

        :return: the message format name
        """
        return self.__name

    def getDescription(self):
        """
        Return the human readable description of this device model format.

        A human readable description of the message format. If the format does
        not contain a description, this method will return an empty String.

        :return: the message format description, or an empty string
        """
        return self.__description if self.__description is not None else ""

    def getURN(self):
        """
        Get the URN of the message format.

        :return: the message format URN
        """
        return self.__urn

    # TODO: there should be a "getField"
    # def getField(self, fieldName):
    #     for field in __fields:
    #         if field.getName() == fieldName
    #             return field
    #     return None

    def getFields(self):
        """
        Get the format of fields.

        :return: list of field formats
        """
        return self.__fields

    def __str__(self):
        """Return a string representation of :class:`.DeviceModelFormat`."""
        comma = ""
        strFields = ""
        for field in self.__fields:
            strFields = str(f'{strFields}\n{{{field}}}{comma}')
            comma = ","

        return str(f'name = {self.__name}` \
            , description = {self.__description} \
            , type = {self.__type} \
            ,\n fields = [{strFields}\n]')


class DeviceModelAction():
    """A :class:`.DeviceModel` action model."""

    class Argument():
        """A :class:`.DeviceModelAction` argument model."""

        __slots__ = (
            '__name',
            '__argType',
            '__lowerBound',
            '__upperBound',
            '__defaultValue'
        )

        def __init__(self, name, argType, description=None, lowerBound=None,
                     upperBound=None, defaultValue=None):
            self.__name = name
            self.__argType = argType
            if self.__argType is DeviceModelAttribute.Type.INTEGER or \
                    self.__argType is DeviceModelAttribute.Type.NUMBER:
                self.__lowerBound = lowerBound
                self.__upperBound = upperBound
            else:
                self.__lowerBound = None
                self.__upperBound = None

            self.__defaultValue = defaultValue

        def getName(self):
            """
            Return the :class:`.Argument` name.

            :return: the action argument name
            """
            return self.__name

        def getArgType(self):
            """
            Return the data type of the :class:`.Argument` action argument.

            If the action does not take an argument, return `None`.
            :return: the :class:`.DeviceModelAction` :class:`.Argument`
                :class:`.DeviceModelAttribute.Type` data type, or `None`
            :return: a :class:`.DeviceModelAttribute.Type` data type, or `None`
            """
            return self.__argType

        def getLowerBound(self):
            """
            Return the :class:`.Argument` lower bound.

            Only :class:`.NUMBER` and :class:`.INTEGER` :class:`.DeviceModelAttribute.Type`
            arguments have a lower bound. For other attribute types `None` is returned.

            :return: a :class:`.Number` or `None` if no lower bound has been set, or
            :method:`.getArgType` returns a type other than :class:`.NUMBER` or
            :class:`.INTEGER`
            """
            return self.__lowerBound

        def getUpperBound(self):
            """
            Return the :class:`.Argument` upper bound.

            Only :class:`.NUMBER` and :class:`.INTEGER` :class:`.DeviceModelAttribute.Type`
            arguments have an upper bound. For other attribute types `None` is returned.

            :return: a :class:`.Number` or `None` if no upper bound has been set, or
            :method:`.getArgType` returns a type other than :class:`.NUMBER` or
            :class:`.INTEGER`
            """
            return self.__upperBound

        def getDefaultValue(self):
            """
            Return the argument's default value.

            If a default value is defined in the :class:`.DeviceModel`, that value is
            returned. If there is no for the argument in the device model, then `None`
            is returned.
            The value `None` is *not* a default value.

            :return: the argument's default value or `None` if no default value is
            defined
            """
            return self.__defaultValue

        def checkBounds(self, value):

            # Assumption here is that lowerBound <= upperBound
            if value is None and (self.getUpperBound() is not None or \
                    self.getLowerBound() is not None):
                raise ArgumentException(f'DeviceModelAction.Argument.checkBounds'
                                        f'{self.getName()} value None is out of range')
            val = float(value)
            upperBound = self.getUpperBound()
            if upperBound is not None:
                upper = float(upperBound)
                if val > upper:
                    raise ArgumentException(f'DeviceModelAction.Argument.checkBounds'
                        f'{self.getName()} out of range {val} > {upper}')

            lowerBound = self.getLowerBound()
            if lowerBound is not None:
                lower = float(lowerBound)
                if val < lower:
                    raise ArgumentException(f'DeviceModelAction.Argument.checkBounds'
                                            f'{self.getName()} out of range {val} < {lower}')

            return

        def __str__(self):
            return f'{{ name = {self.__name}, '\
                f'type = {self.__argType if self.__argType is not None else "None"}, '\
                f'lowerBound = {self.__lowerBound if self.__lowerBound is not None else "None"}, '\
                f'upperBound = {self.__upperBound if self.__upperBound is not None else "None"}, '\
                f'defaultValue = {self.__defaultValue if self.__defaultValue is not None else "None"}}}'

        @classmethod
        def fromJson(cls, jsonObject):
            name = "value"
            if "name" in jsonObject:
                name = jsonObject["name"]
            argtype = None
            if "type" in jsonObject:
                argtypestr = jsonObject["type"]
                try:
                    argtype = DeviceModelAttribute.Type[argtypestr]
                except:
                    raise ArgumentException(
                        f'DeviceModelAction.Argument.fromJson(): Invalid type {argtypestr}')
            min = None
            max = None
            if "range" in jsonObject:
                try:
                    rangestr = jsonObject["range"]
                    range = rangestr.split(",")
                    min = float(range[0])
                    max = float(range[1])
                except:
                    raise ArgumentException(
                        f'DeviceModelAction.Argument.fromJson(): Invalid range {rangestr}')
            defaultValue = None
            if "defaultValue" in jsonObject:
                defaultValue = jsonObject["defaultValue"]
            return DeviceModelAction.Argument(name=name, argType=argtype, lowerBound=min,
                                              upperBound=max, defaultValue=defaultValue)


    __slots__ = (
        '__name',
        '__description',
        '__argumentList'
    )

    def __init__(self, name, description, argumentList=None):
        """
        Create an instance of DeviceModelAction.

        :param name:  the action name
        :param description: the action human readable description
        :param argumentList: a `list` of :class:`.Argument` action arguments.
        """
        self.__name = name
        self.__description = description
        self.__argumentList = argumentList

    def getName(self):
        """
        Get the action name.

        :return: the action name from the device model
        """
        return self.__name

    def getDescription(self):
        """
        Return the human readable description of this action.

        If the model does not contain a description, this method will return
        an empty String.

        :return: the action description, or an empty string
        """
        return self.__description

    def getArguments(self):
        """
        Return the action :class:`.Argument` arguments as `list`.

        :return: the action :class:`.Argument` arguments else `None`
        """
        return self.__argumentList

class DeviceModelParser(object):
    """
    Parse JSON into a :class:`iotcs.client.DeviceModel`.

    JSON in this context is a JSON formatted String or a readable stream
    """

    @classmethod
    def fromJson(cls, jsonObject):
        """
        Create a `DeviceModel` instance.

        Return a :class:`iotcs.client.DeviceModel` from a JSON formatted
        String
        :param jsonString: a JSON formatted String
        :return: a :class:`iotcs.client.DeviceModel` instance
        :raises: json exceptions if `jsonString` cannot be decoded
        :raises: ValueError if required values are not present, or invalid
        """
        urn = None
        try:
            urn = jsonObject["urn"]
            if urn is None or 0 == len(urn):
                raise ValueError("urn cannot be null or empty")
        except Exception as e:
            raise ValueError("urn is required: %s", str(e))

        name = None
        try:
            name = jsonObject["name"]
            if name is None or 0 == len(name):
                raise ValueError("name cannot be null or empty")
        except Exception as e:
            raise ValueError("name is required: %s", str(e))

        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]

        #  Attributes
        attributes = []
        existingAttrNames = []
        if "attributes" in jsonObject:
            jsonAttrs = jsonObject["attributes"]
            if jsonAttrs is not None:
                for jattr in jsonAttrs:
                    attribute = \
                        cls.createDeviceModelAttribute(urn,
                                                       existingAttrNames,
                                                       jattr)
                    attributes.append(attribute)
                    existingAttrNames.append(attribute.getName())

        #  Actions
        actions = []
        existingActionsNames = []
        if "actions" in jsonObject:
            jsonActions = jsonObject["actions"]
            if jsonActions is not None:
                for jaction in jsonActions:
                    action = cls.createDeviceModelAction(existingActionsNames,
                                                         jaction)
                    actions.append(action)
                    existingActionsNames.append(action.getName())

        #  Formats
        formats = []
        if "formats" in jsonObject:
            jsonFormats = jsonObject["formats"]
            if jsonFormats is not None:
                for jformat in jsonFormats:
                    format = cls.createDeviceModelFormat(jformat)
                    formats.append(format)

        return DeviceModelImpl(urn, name, description, attributes, actions,
                               formats)

    @classmethod
    def createDeviceModelAttribute(cls, urn, existingFieldNames, jsonObject):
        """
        Create and return a :class:`.DeviceModelAttribute`.

        Return a :class:`iotcs.shared.DeviceModelAttribute from the json
        object `root`
        :param urn:
        :param existingFieldNames:
        :param jsonObject: The device model attribute in json object format
        :return: a `DeviceModelAttribute`
        :raises: ValueError if required value is missing or value is invalid
        """
        # TODO: Isn't name required ?
        name = None
        if "name" in jsonObject:
            name = jsonObject["name"]
            if name in existingFieldNames:
                raise ValueError("duplicate attribute: %s ", name)

        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]

        alias = None
        if "alias" in jsonObject:
            alias = jsonObject["alias"]

        access = None
        if "writable" in jsonObject:
            access = DeviceModelAttribute.Access.READ_WRITE \
                if bool(jsonObject["writable"]) else \
                DeviceModelAttribute.Access.READ_ONLY
        else:
            # TODO: is the the appropriate default ?
            access = DeviceModelAttribute.Access.READ_ONLY

        attrType = None
        if "type" in jsonObject:
            try:
                t = jsonObject["type"]
                if t is not None:
                    attrType = DeviceModelAttribute.Type[t]
            except Exception as e:
                raise ValueError("Illegal type: %s, %s", t, str(e))

        min = None
        max = None
        if "range" in jsonObject:
            range = jsonObject["range"]
            try:
                if range is not None:
                    strings = range.split(",")
                    try:
                        min = int(strings[0])
                    except ValueError as ve:
                        min = float(strings[0])

                    try:
                        max = int(strings[1])
                    except ValueError as ve:
                        max = float(strings[1])

            except Exception as e:
                raise ValueError("Invalid range: " + range)

        # TODO: I am assuming these are "real" types from the json loads.
        defaultValue = None
        badValue = True
        if "defaultValue" in jsonObject:
            defaultValue = jsonObject["defaultValue"]
            if defaultValue is not None:
                if attrType == DeviceModelAttribute.Type.INTEGER and \
                        isinstance(defaultValue, int):
                    badValue = False
                elif attrType == DeviceModelAttribute.Type.NUMBER and \
                        isinstance(defaultValue, float):
                    badValue = False
                elif attrType == DeviceModelAttribute.Type.STRING and \
                        isinstance(defaultValue, str):
                    badValue = False
                elif attrType == DeviceModelAttribute.Type.BOOLEAN and \
                        isinstance(defaultValue, bool):
                    badValue = False
                elif attrType == DeviceModelAttribute.Type.DATETIME and \
                        isinstance(defaultValue, int):
                    # TODO: don't convert ?
                    # TODO: defaultValue = Date((defaultValue).longValue())
                    badValue = False
                elif attrType == DeviceModelAttribute.Type.URI and \
                        isinstance(defaultValue, str):
                    # TODO: Only HL API uses defaults, so only HL object is
                    # TODO: supported for now StorageObject case?
                    defaultValue = ExternalObject(str(defaultValue))
                    badValue = False
            if badValue:
                raise ValueError("Invalid defaultValue: ", defaultValue)

        return DeviceModelAttribute(urn, name, description, attrType, min, max,
                                    access, alias, defaultValue)

    @classmethod
    def createDeviceModelAction(cls, existingActionNames, jsonObject):
        """
        Create and return a :class:`.DeviceModelAction` instance.

        Return a :class:`iotcs.client.impl.DeviceModelAction`
        from a python json object
        :param existingActionNames: previously processed actions
        :param jsonObject: a python object representing a json
        :class:`iotib.client.impl.DeviceModelAction`
        :return: a :class:`iotcs.client.impl.DeviceModelAction` instance
        :raises: ValueError if the action exists in `existingActionNames` or
        jsonObject` has an invalid
        :class:`iotcs.client.impl.DeviceModelAction.Type` argument type
        """
        # TODO: Need to know the required fields
        name = None
        if "name" in jsonObject:
            name = jsonObject["name"]
            if name in existingActionNames:
                raise StateException(
                    "DeviceModelParser.createDeviceModelAction(): duplicate action: %s", name)

        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]

        argumentList = list()
        if "arguments" in jsonObject:
            jsonArguments = jsonObject["arguments"]
            for argument in jsonArguments:
                actionarg = DeviceModelAction.Argument.fromJson(argument)
                argumentList.append(actionarg)
        elif "argType" in jsonObject:
            # Legacy... anonymous arg... name it value
            argType = None
            argTypeStr = "None"
            try:
                argTypeStr = jsonObject["argType"]
                if argTypeStr is not None:
                    argType = DeviceModelAttribute.Type[argTypeStr]
            except Exception as e:
                raise StateException(
                    f'DeviceModelParser.createDeviceModelAction(): Invalid action type: {argTypeStr}')
            argument = DeviceModelAction.Argument("value", argType)
            argumentList.append(argument)

        return DeviceModelAction(name, description, argumentList)

    @classmethod
    def createDeviceModelFormat(cls, jsonObject):
        """
        Create and return a :class:`.DeviceModelFormat` instance.

        Return a :class:`iotcs.client.impl.DeviceModelFormat` from a python
        json object
        :param jsonObject: a python object representing a json
        :class:`iotcs.client.impl.DeviceModelFormat`
        :return: a :class:`iotcs.client.impl.DeviceModelFormat` instance
        """
        urn = None
        if "urn" in jsonObject:
            urn = jsonObject["urn"]

        name = None
        if "name" in jsonObject:
            name = jsonObject["name"]

        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]

        # TODO: Exception should probably be caught if type is not valid.
        formatType = None
        if "type" in jsonObject:
            t = jsonObject["type"]
            formatType = DeviceModelFormat.Type[t]

        fields = []
        value = None
        if "value" in jsonObject:
            value = jsonObject["value"]

            jsonFields = None
            if "fields" in value:
                jsonFields = value["fields"]
                for f in jsonFields:
                    field = cls.createMsgFormatField(f)
                    fields.append(field)

        return DeviceModelFormat(urn, name, description, formatType, fields)

    @classmethod
    def createMsgFormatField(cls, jsonObject):
        """
        Create and return a :class:`.DeviceModelFormat.Field`.

        :param jsonObject: a python object representing a json
            :class:`iotcs.client.impl.DeviceModelFormat.Field`
        :return: a :class:`iotcs.client.impl.DeviceModrlFormat.Field` instance
        """
        # TODO: Need to know what fields are required
        name = None
        if "name" in jsonObject:
            name = jsonObject["name"]

        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]

        formatType = None
        if "type" in jsonObject:
            formatType = jsonObject["type"]

        optional = None
        if "optional" in jsonObject:
            optional = bool(jsonObject["optional"])

        return DeviceModelFormat.Field(name, description, formatType, optional)


# TODO: Should this be ABC so that it cannot be institated ?
# TODO: class or static methods ?
class DeviceModelFactory():
    """DeviceModel Factory."""

    __dmConfig = Util.getConfig()
    __DM_PREFIX = __dmConfig.get("device_model", "prefix")
    __localStorePathname = __dmConfig.get("device_model", "device_model_store")
    __FILE_TYPE = __dmConfig.get("device_model", "suffix")
    __ALLOW_DRAFT_MODELS = __dmConfig["device_model"].getboolean("allow_draft_device_models")

    @classmethod
    def checkLocalStorePath(cls, localStorePathname):
        """
        Verify the local store path.

        If device_model_store is set, then make sure it is readable/writable,
        or try to create the directory if it does not exist.

        :param localStorePathname: the device model store directory path
        :return: the absolute path to the device model store directory else
            'None' if the local storing of device models is not configured or
            the directory cannot be read or written
        """
        if localStorePathname is None or "" == localStorePathname:
            return None
        dmstorepath = os.path(localStorePathname)
        if os.path.exists(dmstorepath):
            if not os.path.isDir(dmstorepath):
                cls.getLogger().warning(
                    "Local device model store is not a directory: %s ",
                    localStorePathname)
                return None
            if os.access(dmstorepath, os.W_OK):
                return os.path.abspath(dmstorepath)
            if not os.access(dmstorepath, os.R_OK):
                cls.getLogger().warning(
                    "Cannot read from local device model store: %s",
                    localStorePathname)
            if not os.access(dmstorepath, os.W_OK):
                cls.getLogger().warning(
                    "Cannot write to local device model store: %s",
                    localStorePathname)
        elif os.mkdir(dmstorepath):
            return os.path.abspath(dmstorepath)
        else:
            cls.getLogger().warning(
                "Cannot create local device model store: %s ",
                localStorePathname)
        return None

    @classmethod
    def getDeviceModel(cls, secureConnection, urn, aid=None):
        """
        Get a device model from the IoT server.

        :param secureConnection: a :class:`iotcs.client.SecureConnection`
            connection to the server.
        :param urn: the URN of desired device model
        :param aid: the IoT application identifier or None

        :return: device model or null if not found

        :raises: NullPointerException if an argument is null
        :raises: IOException if the request failed
        :raises: GeneralSecurityException when key or signature algorithm class
            cannot be loaded, or the key is not in the trusted assets store,
            or the private key is invalid
        """
        path = None
        if cls.checkLocalStorePath(cls.__localStorePathname) is not None:
            reader = None
            try:
                encoded = quote(urn, encoding="utf-8")
                filename = cls.__DM_PREFIX + encoded + cls.__FILE_TYPE
                path = os.path.join(cls.__localStorePathname, filename)
                reader = open(path, 'r', encoding="utf-8")
                return DeviceModelParser.fromJson(reader)
            except FileNotFoundError as e:
                pass
            except OSError as e:
                cls.getLogger().error(f'Cannot get model from {path}',
                                      sys.exc_info())
                raise e
                #  The model is not local, so get it from the server
            except UnicodeError as e:
                #  UTF-8 is a required encoding so we should never get here.
                cls.getLogger().critical(f'Cannot get model from {path}',
                                         exc_info=1)
                raise e
            finally:
                if reader is not None:
                    try:
                        reader.close()
                    except IOError as ignored:
                        pass
        uri = None
        queryParams = None
        restApi = RestApi.V2.getReqRoot()
        if aid is not None:
            f = Filter.eq("urn", urn)
            uri = f'{restApi}/apps/{aid}/deviceModels/?q={f.toJson()}'
        else:
            uri = f'{restApi}/deviceModels/{urn}'
            queryParams = {'expand': 'formats'}

        data = cls.getObject(secureConnection, uri, queryParams)
        if data is None:
            #  Model not found
            return None
        dm = None
        try:
            jsonObject = json.loads(str(data, "UTF-8"))
            if aid is not None:
                #  we have received a collection of Device Models
                items = jsonObject["items"]
                if 0 > len(items):
                    dmJson = items[0]
                    dm = DeviceModelParser.fromJson(dmJson)
                    if cls.__localStorePathname is not None:
                        data = json.dumps(dmJson)
            else:
                #  we have received a single Device Model
                if not cls.__ALLOW_DRAFT_MODELS and jsonObject["draft"]:
                    return None
                dm = DeviceModelParser.fromJson(jsonObject)
        except Exception as e:
            #  could be JSONException or ClassCastException if server response
            #  is incorrect
            raise IOError("GET " + uri.__str__() + ": " + str(e))
        if path is not None:
            with open(path, "w") as dmfile:
                dmfile.write(data)

        return dm

    @classmethod
    def getObject(cls, secureConnection, uri, queryParams):
        """
        Get the model object from the server defined by the `uri`.

        :param secureConnection: the :class:`iotcs.client.SecureConnection`
            to make requests
        :param uri: the device model resource uri
        :param queryParams: `dict` of query parameters
        :return: the device model as JSON
        """
        res = secureConnection.get(uri, queryParams)
        status = res.getStatus()
        if status == StatusCode.NOT_FOUND.value:
            return None
        if status != StatusCode.OK.value:
            raise IOError(res.getVerboseStatus("GET", uri))
        data = res.getData()
        if data is None:
            raise IOError("GET " + uri + " failed: no data received")
        return data

    @classmethod
    def getLogger(cls):
        """Return a logger for `DeviceModelFactory` log messages."""
        return logging.getLogger("iotcs.client.DeviceModelFactory")
