# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
import base64
import collections
import copy
from datetime import datetime
from enum import Enum
import json
import logging
import sys
import uuid

from iotcs.common import StatusCode
from iotcs.common import TimeManager
from iotcs.common import Util



class Message(ABC):
# TODO: Why aren't these embedded classes of Message ?

    class Priority(Enum):
        """The priorty of the message.
        Deliviery of a message with a higher priority will take precedence over
        delivery of a message with lower priority.
        """
        LOWEST = 0
        LOW = 1
        MEDIUM = 2
        HIGH = 3
        HIGHEST = 4

        def __int__(self):
            return self.value


    class Reliability(Enum):
        """The required reliability of delivery of a message.
        NO_GUARANTEE: Message will not be persisted, but will be retried a default number of times
        BEST_EFFORT: Message will be persisted, and will be retried twice the default number of times
        GUARANTEED_DELIVERY: Message will be persisted, and will be retried until successfully delivered
        """
        NO_GUARANTEE = 1
        BEST_EFFORT = 2
        GUARANTEED_DELIVERY = 3

    class Severity(Enum):
        """The required reliability of delivery of a message.
        NO_GUARANTEE: Message will not be persisted, but will be retried a default number of times
        BEST_EFFORT: Message will be persisted, and will be retried twice the default number of times
        GUARANTEED_DELIVERY: Message will be persisted, and will be retried until successfully delivered
        """
        CRITICAL = 1
        SIGNIFICANT = 2
        NORMAL = 3
        LOW = 4

    class Type(Enum):
        """
        " The message type. Each type corresponds to one type of message.
        """
        DATA = "DATA"
        REQUEST = "REQUEST"
        RESPONSE = "RESPONSE"
        RESOURCE = "RESOURCES_REPORT"
        ALERT = "ALERT"

    class Direction(Enum):
        """
        " The message direction with respect to the device.
        """
        FROM_DEVICE = 0
        TO_DEVICE = 1

    """An abstract class for the different types of messages sent to,
    or received from, the messaging server.
    """

    # Constant defines maximum length of string key in Json
    MAX_KEY_LENGTH = 2048

    #Constant defines maximum length of string value in Json
    MAX_STRING_VALUE_LENGTH = 64 * 1024

    class Encoder(json.JSONEncoder):
        """An encoder for encoding a Message to JSON"""
        #def default(self, o):
        def encode(self, o):
            dct = dict()
            for attr in Message.__slots__:
                if attr == 'remainingRetries':
                    continue
                if attr == '_ordinal':
                    continue
                value = getattr(o, attr, None)
                if value == None:
                    continue
                if isinstance(value, Enum):
                    value = value.name
                # To distiguish from builtin "id"
                if attr == "msgId":
                    attr = "id"
                elif attr == "properties":
                    # TODO: Only if "expand" is true, where should that be set and used ?
                    allprops = value.getAllProperties()
                    if len(allprops) > 0:
                        value = value.toJson()
                    else:
                        continue
                elif attr == "diagnostics":
                    value =  value
                dct[attr] = value
            return dct

    # These are the allowable attributes of a Message.
    __slots__ = (
        'msgId',  # unique id assigned by server
        'clientId',  # id assigned by client
        'source',  # endpoint id of the device that created the message
        'destination',  # endpoint id of the device receiving the message
        'priority',  # message Priority
        'reliability',  # message Reliability
        'eventTime',  # message event time
        'properties',  # extra information for the message
        'sender',  # the endpoint of the device sending the message
        'diagnostics',  # message diagnostics
        'direction',  # the message Direction
        'receivedTime',  # the time the message was received (epoch value in nanoseconds)
        'sentTime',  # the time the message was sent (epoch value in nanoseconds)
        '_ordinal', # maintains order of messages when messages are created less than a millisecond apart
        'remainingRetries' # the number of retry attempts remaining
    )

    # TODO: Should the constructor ensure that required attributes are specified ? i.e. not given default values ?
    def __init__(self, msgId=None, source=None, destination=None, clientId=None,
                 reliability=Reliability.BEST_EFFORT, priority=Priority.LOW,
                 properties=None, diagnostics=None,
                 eventTime=None, sender=None, direction=None,
                 receivedTime=None, sentTime=None):
        """
        Constructs a Message object. Only subclasses of Message should be constructed.
        source: the endpoint id of the device sending the message.
        clientId: an id that may be used by the client to track the message.
        If clientId is not provided, one will be generated.
        reliability: the required reliability of the message
        priority: the required priority of the message
        eventTime: the time the event occurred. If eventTime is not provided,
        the value defaults to the current time.

        Validation occurs when converting to and from json
        """

        # TODO: Message constructor builder.source is required
        self.source = source
        self.msgId = msgId
        # TODO: Message constructor builder.destination is set to "" if null
        self.destination = destination

        if clientId is None:
            clientId = uuid.uuid4().hex.__str__()
        self.clientId = clientId

        if reliability is None:
            reliability = Message.Reliability.BEST_EFFORT
        self.reliability = reliability

        if priority is None:
            priority = Message.Priority.LOW
        self.priority = priority

        # TODO: Message constructor builder.destination is set to "" if null
        if sender is  None:
            sender = ""
        self.sender = sender

        # TODO: When does this get set ?
        #if direction is not None:
        self.direction = direction

        if eventTime is None:
            eventTime = TimeManager.currentTimeMillis()
        self.eventTime = eventTime

        # TODO: Not set here in JCL
        # if receivedTime is None:
        #     receivedTime = TimeManager.currentTimeMillis()
        self.receivedTime = receivedTime

        if sentTime is not None:
            self.sentTime = sentTime

        # TODO: JCL increases monotonocally until MAX_INT
        self._ordinal = datetime.now().toordinal()

        #if sentTime is not None:
        self.sentTime =  sentTime

        # TODO: frozen set
        self.diagnostics = diagnostics

        if properties is None:
            properties = MessageProperties()
        else:
            print("properties is not None")
        self.properties = properties

        # TODO: Implement retries
        self.remainingRetries = 0

    # TODO: Setters
    def getId(self):
        return self.msgId

    def getClientId(self):
        return self.clientId

    def getSource(self):
        return self.source

    def getDestination(self):
        return self.destination

    def getPriority(self):
        return self.priority

    def getReliability(self):
        return self.reliability

    def getEventTime(self):
        return self.eventTime

    def getProperties(self):
        return self.properties

    def getSender(self):
        return self.sender

    def getDiagnostics(self):
        return self.diagnostics

    def getDirection(self):
        return self.direction

    def getReceivedTime(self):
        return self.receivedTime

    def getSentTime(self):
        return self.receivedTime

    def getRemainingRetries(self):
        return self.remainingRetries \
            if self.getReliability() != Message.Reliability.GUARANTEED_DELIVERY \
            else sys.maxsize

    def setRemainingRetries(self, retries):
        self.remainingRetries = retries

    def diagnostic(self, name, value):
        if self.diagnostics is None:
            self.diagnostics = dict()
        self.diagnostics[name] = value
        return self

    def toJson(self):
        # TODO: Do we want a json object (python object) returned ?
        # TODO: Why does json.dumps in this return dict and not string as the doc says ?
        """
        Serialize the message to dict
        """
        return json.dumps(self, cls=Message.Encoder)

    def fromJson(self, jsonObject):
        """
        Set values in this instance from the jsonObject
        Typical use case Message().fromJson(jsonObject)
        :param jsonObject:
        """
        source = None
        if "source" in jsonObject:
            source = jsonObject["source"]
        destination = None
        if "destination" in jsonObject:
            destination = jsonObject["destination"]
        if (source is None or len(source) == 0) and \
                (destination is None or len(destination) == 0):
            raise ValueError(f'Message.fromJson: both source and destination must not be None or empty')
        if source is not None:
            self.source = source
        if destination is not None:
            self.destination = destination

        if "id" in jsonObject:
            self.msgId = jsonObject["id"]
        if "clientId" in jsonObject:
            self.clientId = jsonObject["clientId"]

        if "sender" in jsonObject:
            self.sender = jsonObject["sender"]

        # TODO: remove second condition after RequestMessage from Server has source

        priority = None
        if "priority" in jsonObject:
            priority = jsonObject["priority"]
            try:
                priority = Message.Priority[priority]
            except Exception as e:
                raise ValueError(f'Message.fromJson: invalid Priority {priority}')
        if priority is None:
            raise ValueError(f'Message.fromJson: priority must not be None')
        self.priority = priority

        reliability = None
        if "reliability" in jsonObject:
            reliability = jsonObject["reliability"]
            try:
                reliability = Message.Reliability[reliability]
            except Exception as e:
                raise ValueError(f'Message.fromJson: invalid Reliability {reliability}')
        if reliability is None:
            raise ValueError(f'Message.fromJson: reliability must not be None')
        self.reliability = reliability

        if "eventTime" in jsonObject:
            self.eventTime = jsonObject["eventTime"]

        # This is a python object and can be used as is, but need to convert to MessageProperties
        # { "key0" : [ "v0", ..., "vN" ], ..., "keyN" : [ "x0", ..., "xN"] }
        if "properties" in jsonObject:
            self.properties = MessageProperties(jsonObject["properties"])

        # This is a python object and can be used as is
        # { "key0" : <Number>|<Bookean>|<String>, ..., "keyN" : <Number>|<Bookean>|<String> }
        diagnostics = None
        if "diagnostics" in jsonObject:
            diagnostics = jsonObject["diagnostics"]
            self.diagnostics = diagnostics

        direction = None
        if "direction" in jsonObject:
            direction = jsonObject["direction"]
            try:
                direction = Message.Direction[direction]
            except Exception as e:
                raise ValueError(f'Message.fromJson: invalid Direction {direction}')
            self.direction = direction

        if "receivedTime" in jsonObject:
            self.receivedTime = jsonObject["receivedTime"]

        if "sentTime" in jsonObject:
            self.sentTime = jsonObject["sentTime"]

        return self

    # def __setattr__(self, key, value):
    #     if "sentTime" == key:
    #         if value == None:
    #             value = TimeManager.currentTimeMillis()
    #         super().__setattr__("sentTime", value)
    #
    #     raise AttributeError("'" + key + "' may not be set")

    def __str__(self):
        return json.dumps(self.toJson())

    @staticmethod
    def compareTo(a, b):
        if (a == b):
            return 0
        elif (a > b):
            return 1
        else:
            return -1

    @staticmethod
    def compareMessages(msg1, msg2):
        """
        Messages are the same if the object references are equal, Priority is equal, eventTimes are equal,
        Reliability is equal and ordinals are equal.

        msg1 is greater than msg2 if msg1.ordingal is greater than msg2.ordinal, conversely less than,
        otherwise equal

        :param msg1:
        :param msg2:
        :return: 0 if msg1 == msg2, 1 if msg1 > msg2, -1 if msg1 < msg2
        """

        #print("prio1:" + int(msg1.priority) + ",prio2:" + int(msg2.priority))
        c = Message.compareTo(msg1.priority.value, msg2.priority.value)

        #If they are the same priority, take the one that was created first
        if (c == 0):
            c = Message.compareTo(msg1.eventTime, msg2.eventTime)

        #If they are still the same, take the one with higher reliability.
        if (c == 0):
            c = Message.compareTo(msg1.reliability, msg2.reliability)

        # TODO: Is "ordinal" really necessary ? What about JCL ?
        #If they are still the same, take the one that was created first.
        if (c == 0):
            lc = msg1._ordinal - msg2._ordinal
            if (lc > 0):
                c = 1
            elif (lc < 0):
                c = -1
            else:
                c = 0 # this would mean o1 == o2! This shouldn't happen.
        return c


    def __lt__(self, other):
        return Message.compareMessages(self, other) < 0

    def __eq__(self, other):
        if self is other:
            return True
        if self.__class__ != other.__class__:
            return False
        if self.msgId != other.msgId:
            return False
        return self.clientId == other.clientId

    @staticmethod
    def checkKeyLengthAndRaiseError(key):
        keybytes = bytes(key, encoding='utf-8')
        if len(keybytes) >= Message.MAX_KEY_LENGTH:
            raise ValueError(msg = f'key \'{key}\' length {len(keybytes)} exceeds {Message.MAX_KEY_LENGTH} bytes')

    @staticmethod
    def checkValueLengthAndRaiseError(value):
        valuebytes = bytes(value, encoding='utf-8')
        if len(valuebytes) >= Message.MAX_STRING_VALUE_LENGTH:
            raise ValueError(msg = f'value \'{value}\' length {len(valuebytes)} exceeds {Message.MAX_STRING_VALUE_LENGTH}')

_DATA_MESSAGE_RELIABILITY = Message.Reliability.BEST_EFFORT
try:
    _DATA_MESSAGE_RELIABILITY =\
        Message.Reliability[Util.getConfig()["device_messaging"]["data_message_reliability"]]
except:
    pass

_ALERT_MESSAGE_RELIABILITY = Message.Reliability.GUARANTEED_DELIVERY
try:
    _ALERT_MESSAGE_RELIABILITY =\
        Message.Reliability[Util.getConfig()["device_messaging"]["alert_message_reliability"]]
except:
    pass

class AlertMessage(Message):
    """AlertMessage extends Message class. It can be used for sending alert data to IoT Server."""

    class Severity(Enum):
        """
        The `AlertMessage` severity, one of `LOW`, `NORMAL`, `SIGNIFICANT`, `CRITICAL`
        """
        LOW = 4
        NORMAL = 3
        SIGNIFICANT = 2
        CRITICAL = 1

    class Encoder(Message.Encoder):
        """
        An encoder for encoding a AlertMessage to JSON
        """
        def encode(self, o):

            # add the format attribute/value pair
            format = getattr(o, "format", None)
            if format == None:
                raise ValueError("'format' not found")

            # the iterable repr of the common attributes in a message
            msg = super().encode(o)
            msg["type"] = "ALERT"

            # the data message adds a payload containing
            # a format and data items to the message
            payload = dict()
            msg["payload"] = payload
            payload["format"] = format

            description = getattr(o, "description", None)
            if description is None:
                description = ""
            payload["description"] = description

            severity = getattr(o, "severity", None)
            if (isinstance(severity, Enum)):
                payload["severity"] = severity.name
            else:
                raise ValueError("'severity' not found")

            # add the data items as key/value pairs
            items = getattr(o, "items", None)
            if items == None:
                raise ValueError("'items' not found")

            if len(items) > 1:
                data = dict()
                for dataItem in items:
                    data[dataItem.key] = dataItem.value
                payload["data"] = data
            else:
                item = items[0]
                payload["data"] = {item.key : item.value}

            return msg

    # These are the allowable attributes of a AlertMessage.
    __slots__ = (
       'format', # the format of the alert message
       'items',   # the data items
       'description', # Description of the Alert message
       'severity' # Severity of the Alert
    )

    def __init__(self, format=None, items=None, description=None, severity=Severity.SIGNIFICANT,
                 msgId=None, source=None, destination=None, clientId=None,
                 reliability=_ALERT_MESSAGE_RELIABILITY, priority=Message.Priority.HIGHEST,
                 properties=None, diagnostics=None,
                 eventTime=None, sender=None,
                 direction=None, receivedTime=None, sentTime=None):
        """
        Construct an AlertMessage.
        See `Message` for description of `Message` paraemters
        AlertMessage parameters
        format: the URN of the message format in the data model.
        items: one or more DataItems to be included in the message
        severity: defaults to `Severity.SIGNIFICANT`

        :raises: ValueError if `format` is None
        """
        # if format == None:
        #     raise ValueError("'format' may not be None")
        #
        # if items == None:
        #     raise ValueError("'items' may not be None")

        # Message requires soruce
        super(AlertMessage, self).__init__(msgId=msgId, source=source, destination=destination, clientId=clientId,
            reliability=reliability, priority=priority, properties=properties, diagnostics=diagnostics,
            eventTime=eventTime, sender=sender, direction=direction, receivedTime=receivedTime, sentTime=sentTime)

        self.format = format
        self.description = description

        # TODO: must be Severity
        self.severity = severity
        if items is not None:
            alertitems = copy.deepcopy(items)
            self.items = alertitems if isinstance(alertitems, list) else [alertitems]
        else:
            self.items = []

    def getSeverity(self):
        """
        Gets the severity of this alert
        :return: the severity of this alert
        """
        return self.severity

    def setSeverity(self, severity):
        self.severity = severity
        return self

    def getType(self):
        return Message.Type.ALERT

    def getDescription(self):
        """
        Gets the description for this alert.
        :return: the description for this alert.
        """
        return self.description

    def setDescription(self, description):
        self.description = description
        return self

    def getFormat(self):
        """
        Get the name of the format for data.
        :return: data format, never {@code null}.
        """
        return self.format

    def setFormat(self, format):
        self.format = format
        return self

    # TODO: this should probably be in Message
    # TODO: other getters/setters
    def setSource(self, source):
        self.source = source
        return self

    def getDataItems(self):
        """
        Get message payload as {@link List} of {@link DataItem}s.
        :return: Payload that is carried by {@link AlertMessage}, never {@code null}.
        """
        return copy.deepcopy(self.items)

    def dataItem(self, key, value):
        dataItem = DataItem(key, value)
        self.items.append(dataItem)
        return self

    def toJson(self):
        """

        :return:
        """
        return json.dumps(self, cls=AlertMessage.Encoder)

    def fromJson(self, jsonObject):
        super().fromJson(jsonObject)
        payload = None
        if "payload" in jsonObject:
            payload = jsonObject["payload"]
        if payload is None:
            return self
        if "format" in payload:
            self.format = payload["format"]
        if "description" in payload:
            self.description = payload["description"]
        severity = None
        if "severity" in payload:
            try:
                severity = Message.Severity[payload["severity"]]
            except:
                pass
        if severity is None:
            self.severity = Message.Severity.SIGNIFICANT
        else:
            self.severity = severity

        data = None
        if "data" in payload:
            data = payload["data"]
            for key in data.keys():
                value = data[key]
                self.dataItem(key, value)
        return self

class DataItem(object):
    """DataItem is a key/value pair for data in DataMessage.
    """

    class Type(Enum):
        STRING = 0
        DOUBLE = 1
        BOOLEAN = 2

    # These are the allowable attributes of a DataItem.
    __slots__ = (
        'key',
        'value'
    )

    def __init__(self, key, value):
        """Construct a DataItem.
        key: the name of the attribute or field
        value: the value of the attribute or field
        """
        if key == None:
            raise ValueError("'key' may not be None")
        self.key = key

        if value == None:
            raise ValueError("'value' may not be None")
        self.value = value

    def getKey(self):
        return self.key

    def getValue(self):
        return self.value

class DataMessage(Message):
    """
    DataMessage extends Message class. It can be used for sending sensor data to IoT Server.
    """
    class Encoder(Message.Encoder):

        def encode(self, o):

            # add the format attribute/value pair
            format = getattr(o, "format", None)
            if format == None:
                raise ValueError("'format' not found")

            # the iterable repr of the common attributes in a message
            msg = super().encode(o)
            msg["type"] = "DATA"

            # the data message adds a payload containing
            # a format and data items to the message
            payload = dict()
            msg["payload"] = payload
            payload["format"] = format

            # add the data items as key/value pairs
            items = getattr(o, "items", None)
            if items == None:
                raise ValueError("'items' not found")

            data = dict()
            for item in items:
                    data[item.key] = item.value
            payload["data"] = data

            description = getattr(o, "description", None)
            if description is not None:
                payload["description"] = description

            msg["payload"] = payload

            return msg

    # These are the allowable attributes of a DataMessage.
    __slots__ = (
        'format', # the format of the data message
        'items', # the data items
        'msgType' # The `iotcs.shared.message.Message.Type`
    )

    def __init__(self, format=None, items=None,
                 # Message parameters
                 msgId=None, source=None, destination=None, clientId=None,
                 reliability=_DATA_MESSAGE_RELIABILITY, priority=None, properties=None,
                 diagnostics=None, eventTime=None, sender=None,
                 direction=Message.Direction.FROM_DEVICE,
                 receivedTime=None, sentTime=None):
        """
        Return a :class:`.DataMessage`.

        :param format:
        :param items: `list` of `iotcs.shared.message.DataItem`
        :param msgId:
        :param source:
        :param destination:
        :param clientId:
        :param reliability:
        :param priority:
        :param properties:
        :param diagnostics:
        :param eventTime:
        :param sender:
        :param direction:
        :param receivedTime:
        :param sentTime:
        """

        super(DataMessage, self).__init__(msgId=msgId, source=source, destination=destination,
            clientId=clientId,reliability=reliability, priority=priority,
            properties=properties, diagnostics=diagnostics,
            eventTime=eventTime, sender=sender, direction=direction, receivedTime=receivedTime,
            sentTime=sentTime)

        # TODO: format is required by "toJson" why is this not required here ?
        # TODO: to support DataMessage().fromJson() ?
        # TODO: Does a DataMessage actually need a format if there are no items ?
        # TODO: But if there are not items what is the message ?
        self.format = format

        if items is not None:
            dataitems = copy.deepcopy(items)
            self.items = dataitems if isinstance(dataitems, collections.Container) else [dataitems]
        else:
            self.items = []

        self.msgType = Message.Type.DATA

    def getFormat(self):
        return self.format

    def getDataItems(self):
        return self.items

    def getType(self):
        return self.msgType

    def fromJson(self, jsonObject):
        payload = None
        if "payload" in jsonObject:
            payload = jsonObject["payload"]

        super().fromJson(jsonObject)
        if payload is None:
            return self

        if "format" in payload:
            self.format = payload["format"]

        data = None
        if "data" in payload:
            data = payload["data"]
            for key in data.keys():
                value = data[key]
                self.addItem(key, value)

        return self

    def addItem(self, key, value):
        if isinstance(value, (int, float,)):
            self.items.append(DataItem(key, float(value)))
        elif isinstance(value, (str)):
            self.items.append(DataItem(key, value))
        elif isinstance(value, (bool)):
            self.items.append(DataItem(key, bool(value)))
        else:
            logging.getLogger("DataMessage").error(
                msg = f'DataMessage.addItem(): unknown type {type(value)} of key {key}')
        return self

    def toJson(self):
        return json.dumps(self, cls=DataMessage.Encoder)

    def add(self, dataItem):
        self.items.append(dataItem)


# com.oracle.iot.client.message

class MessageProperties():
    """
    MessageProperties contains a table of key and list of values pairs for 
    extra information to send. This class is immutable
    """


    # `A Dictionary` for message properties. It is a key/list of value
    # pairs
    ___slots___ = {
        'propertiesTable'
    }

    def __init__(self, properties=None):
        """

        :param properties:
        """
        if properties is None:
            properties = dict()
        self.propertiesTable = properties

    def addValues(self, key, values):
        """
        Add a new key/values pair. If the key exists, adds the values to the
        list of existing values. If the key does not exist, puts the key and
        values to the table. Key or values cannot be {@code null}. Key cannot
        be empty or long string. Values cannot contain long strings. Maximum
        length for key is `Message.Utils.MAX_KEY_LENGTH` bytes, maximum length
        for any value is `Message.Utils.MAX_STRING_VALUE_LENGTH` bytes. The
        length is measured after the string is encoded using UTF-8 encoding.

        :param  key: property key
        :param  values: property values as `list()` or single value
        :return: MessageProperties.Builder
        :raises: ValueError for `None` key or if the values are `None` or any
        item in values is `None`.
        :raises: ValueError for empty or long key and for any long item in
                        values.
        """
        self.validateKey(key, 'MessageProperties.addValue()')
        self.validateValues(values, 'MessageProperties.addValue()')

        existingValues = None
        if key in self.propertiesTable:
            existingValues = self.propertiesTable.get(key)
        else:
            existingValues = []

        if isinstance(values, (list)):
            existingValues.extend(values)
        else:
            existingValues.append(values)

        self.propertiesTable[key] = existingValues
        return self

    def validateKey(self, msgprefix, key):
        """
        Raises `ValueError` with appropriate message for failure
        if `key` is `None`, 0 length, or its length exceeds `Message.MAX_KEY_LENGTH` bytes after
        encoing to bytes as `utf-8` `str`
        :param msgprefix: `ValueError` message prefix
        :param key: the key to validate
        :raises: `ValueError` if key is invalid.
        """
        if key is None:
            raise ValueError(msg = f'{msgprefix}: key must not be None')
        if len(key) == 0:
            raise ValueError(msg = f'{msgprefix}: key must not be 0 length')
        try:
            Message.checkKeyLengthAndRaiseError(key)
        except Exception as e:
            raise ValueError(msg = f'{msgprefix}: {e}')

    def validateValues(self, msgprefix, values):
        """
        Raises `ValueError` with appropriate message for failure under the following conditions
        - `values` is None
        - `values` is a `list()` or any of its values (of if `values` is a single value)
           meets any of the following conditions
        - a value is None, 0 length or exceeds `Messsage.MAX_STRING_VALUE_LENGTH` after encoding to bytes as
          a `utf-8` `str`

        :param msgprefix: `ValueError` message prefix
        :param values: a single or `list()` of values
        :raises: `ValueError` if values is invalid.
        """
        if values is None:
            raise ValueError(msg = f'{msgprefix}: values must not be None')
        if len(values) == 0:
            raise ValueError(msg = f'{msgprefix}: values must not be 0 length')
        if isinstance(values, (list)):
            for v in values:
                try:
                    Message.checkValueLengthAndRaiseError(v)
                except Exception as e:
                    raise ValueError(msg = f'{msgprefix}: {e}')

    def copy(self, properties):
        """
        Copy another `MessageProperties` by adding all properties to the
        current `MessageProperties`
        :param properties: `MessageProperties` to copy
        :return: `MessageProperties.Builder`
        """
        for key in properties.getKeys():
            properties = properties.getProperties(key)
            if key not in self.propertiesTable:
                # Should probably copy
                self.propertiesTable[key] = properties
                continue

            existingProperties = self.propertiesTable[key]
            # Can't be None
            existingProperties.extend(properties)

        return self

    def containsKey(self, key):
        """
        Check if the properties contain the key.
        :param key: property key
        :return: `True` if the properties contain the key
        """
        return key in self.propertiesTable

    def getKeys(self):
        """
        Get a `set()` of the keys.
        :return: a set of the keys, never `None`
        """
        # TODO: frozen set
        return self.propertiesTable.keys()

    def getProperties(self, key):
        """
        Get a `list()` of values for a particular key.
        :param key: property key
        :return: a `list()` of `str` values
        """
        # TODO: frozen set
        return self.propertiesTable[key] if key in self.propertiesTable else None


    def getProperty(self, key, index=-1):
        """
        if index is greater than -1
        Get the first value for a particular key.
        :param key: property key
        :return: value, may return `None` if the key does not exist 
        or values assigned to this key is empty.

        if index is -1
        Get a specific value.
        :param key: property key
        :param index: index of the `list()` of the `str` values
        :return: value, may return `None` if the key does not exist or 
        index is out of range.
        """
        if key in self.propertiesTable:
            if index != -1:
                propertyValues = self.propertiesTable[key]
                if propertyValues is not None and len(propertyValues) > 0:
                    return propertyValues[0]
                else:
                    return None
            else:
                propertyValues = self.propertiesTable[key]
                if propertyValues is not None and len(propertyValues) > index:
                    return propertyValues[index]
                else:
                    return None
        else:
            return None

    def getAllProperties(self):
        """
        Get all properties.
        :return a `Dictionary` of key and list of values, never `None`
        """
        # TODO: frozen set
        return self.propertiesTable

    def __str__(self):
        """
        Method to print the message properties in JSON format.
        @return message properties in string format
        """
        return json.dumps(self.toJson())

    def toJson(self):
        """
        Method to export the message properties to a json object.
        :return message properties as json object (python object)
        """

        # properties = dict()
        # for k in self.getKeys():
        #     jsonValue = []
        #     values = self.getProperties(k)
        #     for v in values:
        #         jsonValue.append(v)
        #     try:
        #         properties[k] = jsonValue
        #     except Exception as e:
        #         raise ValueError(
        #             msg = f'MessageProperties.toJson(): caught Exception {e}')
        # return properties

        # Try this instead
        return json.loads(self.propertiesTable)

    def __eq__(self, o):
        """
        {@inheritDoc}
        """
        if not isinstance(o, MessageProperties):
            return False

        if id(self) == id(o):
            return True

        if o == None:
            return False

        if self.propertiesTable is None and o.propertiesTable is None:
            return True
        else:
            return self.propertiesTable == o.propertiesTable

    def __hash___(self):
        """
        {@inheritDoc}
        """
        return hash(self.propertiesTable)


class RequestMessage(Message):
    """ReuestMessage extends Message class. This can be used for processing request messages from server.
    """
    # Ecnoding the super class happens before we get here
    class Encoder(json.JSONEncoder):
        """An encoder for encoding a Message to JSON"""
        #def default(self, o):
        def encode(self, o):
            dct = dict()
            for attr in RequestMessage.__slots__:
                if attr == '_ordinal':
                    continue
                value = getattr(o, attr, None)
                if value == None:
                    continue
                if attr == "body":
                    value = base64.b64encode(value).decode('utf-8')
                if isinstance(value, Enum):
                    value = value.name
                dct[attr] = value
            return dct

    # These are the allowable attributes of a AlertMessage.
    __slots__ = (
        'url',
        'body',
        'method',
        'headers',
        'params',
        'msgType'
    )

    def __init__(self,
            # RequestMessage args
            url=None, method=None, headers=None, body=None, params=None,
            # Message args
            msgId=None, source=None, destination=None, clientId=None,
            reliability=Message.Reliability.BEST_EFFORT, priority=Message.Priority.LOW,
            properties=None, diagnostics=None,
            eventTime=None, sender=None, direction=None,
            receivedTime=None, sentTime=None):
        """

        :param method:
        :param url:
        :param body:
        :param headers:
        :param params:
        :raises: ValueError if 'url' or 'headers' is None
        """

        # if url == None:
        #     raise ValueError("'url' may not be None")
        # if headers == None:
        #     raise ValueError("'headers' may not be None")
        # if method == None:
        #     raise ValueError("'method' may not be None")

        # Is this equivalent to "super(...)" ?
        super().__init__(msgId=msgId, source=source, destination=destination, clientId=clientId,
            reliability=reliability, priority=priority, properties=properties, diagnostics=None,
            eventTime=eventTime, sender=sender, direction=direction,
            receivedTime=receivedTime, sentTime=sentTime)

        self.url = url
        # TODO: Should this be a copy ?
        self.headers = headers
        self.method = method
        # TODO: Should this be a copy ?
        self.params = params
        self.msgType = Message.Type.REQUEST

        # body is byte[]
        self.setBody(body)

    def getMethod(self):
        return self.method

    def getMethodOverride(self):
        methodOverrideHeaderName = "X-HTTP-Method-Override".upper()
        for header in self.headers:
            if methodOverrideHeaderName == header.upper():
                method = self.headers[header]
                # TODO: Must this be a list ?
                if isinstance(method, (list)):
                    return method[0]
                else:
                    return method

    def getBody(self):
        return self.body

    def setBody(self, body):
        if body is None:
            self.body = bytearray()
        elif isinstance(body, (str)):
            # convert the str to bytearray
            self.body = bytearray(body, encoding='utf-8')
        elif isinstance(body, (dict)):
            # convert to json string, then to bytes
            bodystr = json.dumps(body)
            self.body = bytearray(bodystr, encoding='utf8')
        else:
            # assume it is already a bytearray
            self.body = body

    def getParams(self):
        return self.params

    def getURL(self):
        return self.url

    def getHeader(self, name):
        if name in self.headers.keys():
            return self.headers[name]
        else:
            return None

    def getType(self):
        return Message.Type.REQUEST

    def toJson(self):
        msgjson = super().toJson()
        reqmsgjson = json.dumps(self, cls=RequestMessage.Encoder)
        msgjson["payload"] = reqmsgjson
        return msgjson

    def fromJson(self, jsonObject):

        super().fromJson(jsonObject)
        if self.clientId is None or len(self.clientId) == 0:
            raise ValueError(f'{RequestMessage.fromJson.__qualname__}: clientId must be set for RequestMessage')

        payload = None
        if "payload" in jsonObject:
            payload = jsonObject["payload"]
        if payload is None:
            raise ValueError(f'{RequestMessage.fromJson.__qualname__}: request message payload is None')

        method = None
        if "method" in payload:
            method = payload["method"]
        if method is None:
            raise ValueError(f'{RequestMessage.fromJson.__qualname__}: request message method is None')
        self.method = method

        url = None
        if "url" in payload:
            url = payload["url"]
        if url is None:
            raise ValueError(f'{RequestMessage.fromJson.__qualname__}: request message url is None')
        self.url = url

        if "body" in payload:
            _body = payload["body"]
            try:
                self.body = base64.b64decode(str.encode(_body))
            except Exception as e:
                raise ValueError(f'{RequestMessage.fromJson.__qualname__}: cannot base64 decode request message body. {e}')

        if "headers" in payload:
            self.headers = payload["headers"]
        if "params" in payload:
            self.params = payload["params"]

        return self

    def __eq__(self, other):
        if self is other:
            return True
        if self.__class__ != other.__class__:
            return False
        if not super().__eq__(other):
            return False

        if self.body != other.body:
            return False
        if self.headers != other.headers:
            return False
        if self.method != other.method:
            return False
        if self.params != other.params:
            return False
        return self.url == other.url


class Resource():
    class Method(Enum):
        GET = 1
        POST = 2
        PUT = 4
        DELETE = 8
        PATCH = 10

    class Status(Enum):
        ADDED = 0
        REMOVED = 1

    __slots__ = (
        'endpointName', # The endpoint name or ID this resource is for.
        'name', # Resource name. This name should describe the resource in free form
        'status', # Resource status. `iotcs.shared.message.Resource.Status`
        'path', # Path of the resource
        'methods' # A list `iotcs.shared.message.Resource.Method` supported by the this resource
    )

    def __init__(self, name=None, path=None, endpointName=None, status=None, methods=None):
        # TODO: To be consistent with other classes with "fromJson" methods, don't require
        # TODO: init parameters, it is unclear why JCL Resource.Builder() allows its "fromJson"
        # TODO: to not require 'name' and 'path' since Resource(builder) will fail if name
        # TODO: and path have not been set.
        #if name is None or len(name) == 0 or path is None or len(path) == 0:
        #    raise ValueError('Resource name, path cannot be None or empty.')
        self.name = name
        self.path = path
        self.endpointName = endpointName
        if status is None:
            status = Resource.Status.ADDED
        self.status = status

        # if self.status != Resource.Status.REMOVED:
        #     if methods is None or len(methods) == 0:
        #         raise ValueError("Resource methods cannot be None or empty.")

        self.methods = set()
        if methods is not None:
            if isinstance(methods, (list)):
                self.methods.update(methods)
            else:
                self.methods.add(methods)

    def setEndpointName(self, endpointName):
        self.endpointName = endpointName
        return self

    def setPath(self, path):
        self.path = path
        return self

    def setName(self, name):
        self.name = name
        return self

    def setStatus(self, status):
        self.status = status
        return  self

    def setMethod(self, methods):
        """
        method can be list or single method
        :param method:
        :return:
        """
        self.methods.add(methods)
        return self

    def fromJson(self, jsonObject):
        """
        Return a `Resource` instance from `jsonObject`
        :param jsonObject:
        :return:
        """
        if "status" in jsonObject:
            try:
                self.status = Resource.Status[jsonObject["status"]]
            except Exception as e:
                # TODO: JCL does not react to this error, just makes status == REMOVED
                #raise ValueError(f'Resource.fromJson(): jsonObject["status"] = "{jsonObject["status"]}" is not valid.')
                self.status = Resource.Status.REMOVED
        if "methods" in jsonObject:
            methods = jsonObject["methods"]
            if methods is not None:
                for method in methods.split(','):
                    self.methods.add(Resource.Method[method])
        if "endpointName" in jsonObject:
            self.endpointName = jsonObject["endpointName"]
        if "path" in jsonObject:
            self.path = jsonObject["path"]

        return self

    def toJson(self):
        jsonObject = dict()
        if self.endpointName is not None and len(self.endpointName) > 0:
            jsonObject["endpointName"] = self.endpointName
        jsonObject["name"] = self.name
        jsonObject["path"] = self.path

        if self.methods is not None and len(self.methods) > 0:
            methods = ""
            comma = ""
            #for name, member in Resource.Method.__members__.items():
            for method in self.methods:
                methods = f'{methods}{comma}{method.name}'
                comma = ","
            jsonObject["methods"] = methods

        return jsonObject

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, (Resource)):
            return False

        if self.name != other.name:
            return False

        if self.path != other.path:
            return False

        if self.status != other.status:
            return False

        return self.methods is not None and self.methods == other.methods


    def __str__(self):
        return json.dumps(self.toJson())


class ResourceMessage(Message):
    """
    ReuestMessage extends Message class. This can be used for processing request messages from server.
    """

    class Type(Enum):
        UPDATE = 0
        DELETE = 1
        RECONCILIATION = 2

    # These are the allowable attributes of a AlertMessage.
    __slots__ = (
        'endpointName', # Endpoint ID, for the resources that shall be updated
        'resources', # List of resources described my message
        'reconciliationMark', # hash for resource available on client at the moment of message generation
        'resourceType' # The `iotcs.shared.message.ResourceMessage.Type` of this message.
    )

    def __init__(self, endpointName=None, resources=None, resourceType=None, reconciliationMark=None,
        # Message parameters
                 msgId=None, source=None, destination=None, clientId=None,
                 reliability=None, priority=None, properties=None, diagnostics=None,
                 eventTime=None, sender=None, direction=Message.Direction.FROM_DEVICE,
                 receivedTime=None, sentTime=None):

        """

        :param endpointName:
        :param resources:
        :param resourceType:
        :param reconciliationMark:
        :raises:
        """

        super(ResourceMessage, self).__init__(msgId=msgId, source=source, destination=destination,
            clientId=clientId, reliability=reliability,  priority=priority,
            properties=properties, diagnostics=diagnostics,
            eventTime=eventTime, sender=sender, direction=direction,
            receivedTime=receivedTime, sentTime=sentTime)

        # TODO: To be consistent with other classes with "fromJson" methods do not validate parameters
        # if endpointName is None or len(endpointName) == 0:
        #     raise ValueError('ResoureMessage "endpointName" cannot be None or empty')
        # if (resources is None or len(resources) == 0) and (resourceType == Resource.Type.UPDATE):
        #     raise ValueError('ResoureMessage "resources" cannot be None or empty if "resourceType" is UPDATE')

        self.endpointName = endpointName
        self.reconciliationMark = reconciliationMark if reconciliationMark is not None else ""
        self.resources = resources if resources is not None else []
        self.resourceType = resourceType

    def getMD5ofList(self, resStrings):
        """
        Calculates the MD5 hash value for a list of Strings.
        The String list is being alphabetical ordered before the calculation
        :param resStrings: List os Strings
        :return: hash value
        """
        pass

    def delete(self):
        """
        Set the `ResourceMessage` type to `iotcs.shared.message.ResourceMessage.Type.DELETE` and clear the
        resources list.
        :return: self
        """
        self.resourceType = ResourceMessage.Type.DELETE
        if len(self.resources) > 0:
            self.resources.clear()
        return self


    def register(self, resource):
        """
        Set the `ResourceMessage` type to `iotcs.shared.message.ResourceMessage.Type.UPDATE` and add the
        `resource` to the resources list
        :param resource: `iotcs.shared.message.Resource`
        :return: self
        """
        self.resourceType = ResourceMessage.Type.UPDATE
        self.resources.append(resource)
        return self

    def remove(self, resource):
        """
        Set the `ResourceMessage` type to `iotcs.shared.message.ResourceMessage.Type.DELETE` and add the
        `resource` to the resources list
        :param resource: `iotcs.shared.message.Resource`
        :return: self
        """
        self.resourceType = ResourceMessage.Type.REMOVED
        self.resources.append(resource)
        return self


    def add(self, resource):
        """
        `resource` may be one or a list of resources, are added to the resources list.
        :param resource: `iotcs.shared.message.Resource` or a list of `iotcs.shared.message.Resource`
        :return: self
        """
        self.resourceType = ResourceMessage.Type.UPDATE
        if isinstance(resource, (list)):
            self.resources.extend(resource)
        else:
            self.resources.append(resource)
        return self

    # TODO: How to identifiy the python path for RESOURCE ?
    def getResources(self):
        """
        Return the `list` of `iotcs.shared.message.Resource` in this `ResourceMessage`
        :return:
        """
        return self.resources

    def setReconciliationMark(self, mark):
        """
        Set the reconciliation mark.
        :param mark: reconciliation mark.
        :return: self
        """
        self.reconciliationMark = mark
        return self

    def getReconciliationMark(self):
        """

        :return: the reconciliation mark.
        """
        return self.reconciliationMark

    def setEndpointName(self, endpointName):
        """
        Set the endpointName for this message.
        :param endpointName: this message's endpointName
        :return: self
        """
        self.endpointName = endpointName
        return self

    def getEndpointName(self):
        """

        :return: the endpoint id for this `ResourceMessage`
        """
        return self.endpointName

    def reconcile(self):
        """
        Sets the `ResoureceMessage` type to `iotcs.shared.message.ResourceMessage.Type.RECONCILIATION`
        :return: self
        """
        self.resourceType = ResourceMessage.Type.RECONCILIATION
        return self

    def getType(self):
        """

        :return: `iotcs.shared.message.Message.Type.RESOURCE`
        """
        return Message.Type.RESOURCE

    def getMessageType(self):
        """
        Returns type of the message (Could be `iotcs.shared.message.ResourceMessage.UPDATE` or
        `iotcs.shared.message.ResourceMessage.Type.DELETE`)
        :return : `iotcs.shared.message.ResourceMessage.Type` for this `ResourceMessage`
        """
        return self.resourceType

    def fromJson(self, jsonObject):
        super(ResourceMessage, self).fromJson(jsonObject)
        try:
            payload = None
            if "payload" in jsonObject and jsonObject["payload"] is not None:
                payload = jsonObject["payload"]
            else:
                raise ValueError(f'payload must exist and not be None')
            payloadValue = None
            if "value" in payload and payload["value"] is not None:
                payloadValue = payload["value"]
            else:
                raise ValueError(f'payload["value"] must exist and not be None')
            rt = None
            if "reportType" in payloadValue and payloadValue["reportType"] is not None:
                rt = payloadValue["reportType"]
            else:
                raise ValueError(f'value["reportType"] must exist and not be None')
            try:
                self.resourceType = ResourceMessage.Type[rt]
            except Exception as e:
                raise ValueError(f'value["reportType"] = "{rt}" is invalid')
            # TODO: Is this required, don't think so
            if "reconciliationMark" in payloadValue and payloadValue["reconciliationMark"] is not None:
                self.reconciliationMark = payloadValue["reconciliationMark"]

            # TODO: Is this required ? JCL doesn't think so.
            resources = None
            if "resources" in payloadValue and payloadValue["resources"] is not None:
                resources = payloadValue["resources"]
            #else:
            #    raise ValueError('value["resources"] must exist and not be None')
            if resources is not None:
                for resource in resources:
                    self.resources.add(Resource().fromJson(resource))

        except Exception as e:
            raise ValueError(f'ResourceMessage.fromJson(): raised exception {e}')

        return self

    def toJson(self):

        jsonObject = super(ResourceMessage, self).toJson()

        payload = dict()
        value = dict()
        resource = []

        try:
            # This field must use "value" since it is NOT the same as name
            # name -> RESOURCE value -> RESOURCE_TYPE
            jsonObject["type"] =  Message.Type.RESOURCE.value
            payload["type"] = "JSON"
            value["reportType"] =  self.resourceType.name

            if self.reconciliationMark is not None and len(self.reconciliationMark) > 0:
                value["reconciliationMark"] = self.reconciliationMark
            value["endpointName"] = self.endpointName

            for res in self.resources:
                resource.append(res.toJson())

            value["resources"] = resource
            payload["value"] = value
            jsonObject["payload"] = payload
        except Exception as e:
            raise ValueError(f'ResourceMessage.toJson(): raised exception {e}')

        return jsonObject

    def __eq__(self, other):

        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, (ResourceMessage)):
            return False

        if not super(ResourceMessage, self).__eq__(other):
            return False

        if self.resourceType != other.resourceType:
            return False

        if self.resources != other.resources:
            return False

        return self.reconciliationMark == other.reconciliationMark


    def __str__(self):
        return json.dumps(self.toJson())

    class ReportResponse():
        class ResponseStatus(Enum):
            OK = 0
            BAD_REPORT = 1
            RECONCILIATION = 2

        __slots__ = (
            'endpointName',
            'status'
        )

        def __init__(self, endpointName=None, status=None):
            # TODO: To be consistent with other classes with json methods do not validate parameters
            #if endpointName is None or len(endpointName) == 0:
            #   raise ValueError(f'ResourceResponse endpointName cannot be None or empty')
            self.endpointName = endpointName
            if status is None:
                status = ResourceMessage.ReportResponse.ResponseStatus.OK
            self.status = status

        def getEndpointName(self):
            return self.endpointName

        def setEndpointName(self, endpointName):
            """
            Sets the name or ID of the resource's endpoint.
            :return:  self
            """
            self.endpointName = endpointName
            return self

        def getStatus(self):
            return self.status

        def setStatus(self, status):
            self.status = status
            return self

        def fromJson(self, jsonObject):
            if "status" in jsonObject and jsonObject["status"] is not None:
                try:
                    self.status = ResourceMessage.ReportResponse.ResponseStatus[jsonObject["status"]]
                except Exception as e:
                    raise ValueError(f'ReporResponse.fromJson(): invalid status "{jsonObject["status"]}"')
            else:
                raise ValueError(f'ReporResponse.fromJson(): status must exist and not be None')

            if "endpointName" in jsonObject and jsonObject["endpointName"] is not None and \
                    len(jsonObject["endpointName"]) > 0:
                self.endpointName = jsonObject["endpointName"]
            else:
                raise ValueError(f'ReporResponse.fromJson(): endpointName must exist and not be None or empty')

            return self

        def fromString(self, string):
            if string is None or len(string) == 0:
                return self
            try:
                return self.fromJson(json.loads(string))
            except Exception as e:
                raise ValueError(f'ReportResponse.fromString(): failed to convert to instance from {string}')

        def toJson(self):
            res = dict()
            if self.endpointName is None or len(self.endpointName) == 0:
                raise ValueError(f'ResourceResponse.toJson(): endpointName cannot be None or empty')

            res["endpointName"] = self.endpointName
            status = self.status
            if status is None:
                status = ResourceMessage.ReportResponse.ResponseStatus.OK
            res["status"] = status.name

            return res


class ResponseMessage(Message):
    """
    ReuestMessage extends Message class. This can be used for processing request messages from server.
    """

    class Encoder(Message.Encoder):
        """
        An encoder for encoding a ResponseMessage to JSON
        """
        def encode(self, o):

            # the iterable repr of the common attributes in a message
            msg = super().encode(o)
            msg["type"] = Message.Type.RESPONSE.name

            # the response message adds a body
            payload = {}
            payload["headers"] = getattr(o, "headers", {})
            payload["requestId"] = getattr(o, "requestId", None)
            statusCode = getattr(o, "statusCode", None)
            # TODO: Should there be a default ? This must be set right ?
            if statusCode is not None:
                statusCode = statusCode.value
            else:
                statusCode = StatusCode.NOT_IMPLEMENTED.value
            payload["statusCode"] = statusCode
            payload["url"] = getattr(o, "url", "")

            body = getattr(o, "body", None)
            if body is not None:
                # body is 'bytes-like' and b64 returns 'bytes'
                # decode to str for the json object
                body = base64.b64encode(body).decode('utf-8')
                payload["body"] = body
            else:
                # make empty str
                payload["body"] = ""

            msg["payload"] = payload
            #print(json.dumps(msg,indent=4))
            return msg

    # These are the allowable attributes of a ResponseMessage.
    __slots__ = (
        'url',
        'requestId',
        'statusCode',
        'headers',
        'body',
        'msgType'
    )

    def __init__(self, requestMessage=None, url=None, headers=None, body=None, statusCode=StatusCode.OK,
        # Message parameters
        msgId=None, source=None, destination=None, clientId=None,
        reliability=None, priority=None, properties=None, diagnostics=None,
        eventTime=None, sender=None, direction=Message.Direction.FROM_DEVICE,
        receivedTime=None, sentTime=None):
        """

        :param requestMessage:
        :param url:
        :param headers:
        :param body: bytearray or str or dict, if str or dict it is converted to bytearray,
        :param statusCode:
        :param msgId:
        :param source:
        :param destination:
        :param clientId:
        :param reliability:
        :param priority:
        :param properties:
        :param diagnostics:
        :param eventTime:
        :param sender:
        :param direction:
        :param receivedTime:
        :param sentTime:
        """

        # source, destination, clientId, reliability, priority, eventTime, sender, direction,
        # receivedTime, sentTime are derived from RequestMessage
        # TODO: What about "id" (msgid)
        super(ResponseMessage, self).__init__(msgId=msgId,
            source=None if requestMessage is None else requestMessage.destination,
            destination=None if requestMessage is None else requestMessage.source,
            clientId=clientId,
            reliability=reliability, priority=priority, properties=properties, diagnostics=diagnostics,
            eventTime=eventTime, sender=sender, direction=direction,
            receivedTime=receivedTime, sentTime=sentTime)

        self.requestId = None if requestMessage is None else requestMessage.msgId
        self.url = "" if url is None else url
        # TODO: make copy
        self.headers = {} if headers is None else headers
        self.statusCode = StatusCode.OK if statusCode is None else statusCode
        self.setBody(body)

        self.msgType = Message.Type.RESPONSE

    def setStatusCode(self, code):
        self.statusCode = code

    def getStatusCode(self):
        return self.statusCode

    def setBody(self, body):
        if body is None:
            self.body = bytearray()
        elif isinstance(body, (str)):
            # convert the str to bytearray
            self.body = bytearray(body, encoding='utf-8')
        elif isinstance(body, (dict)):
            # convert to json string, then to bytes
            bodystr = json.dumps(body)
            self.body = bytearray(bodystr, encoding='utf8')
        else:
            # assume it is already a bytearray
            self.body = body

    def getBody(self):
        return self.body

    def getType(self):
        return Message.Type.RESPONSE

    def toJson(self):
        """
        Serialize the Response to JSON
        """
        return json.dumps(self, cls=ResponseMessage.Encoder)

    # Not really needed as an api good for testing
    def fromJson(self, jsonObject):
        super().fromJson(jsonObject)
        payload = None
        if "payload" in jsonObject:
            payload = jsonObject["payload"]
        if payload is None:
            raise ValueError(f'invalid ResponseMessage no payload')

        if "type" in payload:
            msgType = payload["type"]
            if msgType != Message.Type.RESPONSE.name:
                raise ValueError(f'invalid type {msgType.name}')
            self.msgType = msgType
        if "headers" in payload:
            self.headers = payload["headers"]
        if "requestId" in payload:
            self.requestId = payload["requestId"]
        if "statusCode" in payload:
            statusCode = payload["statusCode"]
            try:
                statusCode = StatusCode(statusCode)
            except Exception as e:
                raise ValueError(f'invalid status code {statusCode}. {e}')
            self.statusCode = statusCode
        if "url" in payload:
            self.url = payload["url"]
        if "body" in payload:
            # expect a base64 encoded str, encode to bytes, then
            # base64 decode to bytes
            body = payload["body"]
            body = base64.b64decode(body.encode('utf-8'))
            self.body = body
        else:
            self.body = b''

        return self

    def __str__(self):
        return json.dumps(self.toJson())


