# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod
from copy import deepcopy
import datetime
import json
import logging
import numbers
import threading
import urllib


from iotcs.client import ExternalObject
from iotcs.client.model import DeviceModelAttribute
from iotcs.client.model import DeviceModelImpl
from iotcs.common import Pair
from iotcs.common import RestApi
from iotcs.common import StatusCode
from iotcs.common import TimeManager
from iotcs.common import Util
from iotcs.exception import ArgumentException
from iotcs.exception import SecurityException
from iotcs.exception import TransportException
from iotcs.message import AlertMessage
from iotcs.message import DataItem
from iotcs.message import DataMessage
from iotcs.message import Message
from iotcs.message import RequestMessage
from iotcs.message import ResponseMessage
from iotcs.messaging.policyfunctions import DeviceFunction
from iotcs.messaging.client.device.util import RequestDispatcher
from iotcs.persistence import PersistenceStoreManager

class DeviceAnalog(ABC):
    """DeviceAnalog encapsulates a device (endpoint id and attributes) with a model."""

    @abstractmethod
    def getEndpointId(self):
        """
        Return the endpoint id of the device.
        :return: the device endpoint id
        """

    @abstractmethod
    def getDeviceModel(self):
        """
        Return the device model of the device.
        :return: the device model
        """

    @abstractmethod
    def setAttributeValue(self, attribute, value):
        """
        Set the named attribute to the given value.

        :param attribute: the attribute to set
        :param value: the value of the attribute
        :raise ArgumentException: if the attribute is not in the device model,
            the value is `None`, or the value does not match the attribute type.
        """

    @abstractmethod
    def getAttributeValue(self, attribute):
        """
        Return the value of the named attribute.

        :param attribute: the attribute to get
        :return: the value of the attribute, possibly `None`
        """

    @abstractmethod
    def call(self, actionName, argumentValues):
        """
        Invoke an action.

        The `argumentValues` parameter may be empty if
        there are no arguments to pass, but will not be `None`.

        :param actionName: the name of the action to call
        :param argumentValues: the data, possibly `None` or empty, to pass to the action
        """

    @abstractmethod
    def queueMessage(self, message):
        """
        Queue a message for dispatch to the server.

        The implementation of this method may send the message directly
        without queuing.
        :param message: the message to be queued
        """

    #  TODO: this API doesn't really fit here, but needed for alertCondition

class DeviceAnalogImpl(DeviceAnalog):
    """An implementation of :class:`.DeviceAnalog`."""

    __slots__ = (
        'directlyConnectedDevice',
        'deviceModel',
        'endpointId',
        'attributeValueMap'
    )
    def __init__(self, directlyConnectedDevice, deviceModel, endpointId):
        """

        :param directlyConnectedDevice:
        :param deviceModel:
        :param endpointId:
        """
        super(DeviceAnalogImpl, self).__init__()
        self.directlyConnectedDevice = directlyConnectedDevice
        self.deviceModel = deviceModel
        self.endpointId = endpointId
        attributeValueMap = dict()

    def getEndpointId(self):
        """Return :attr:`.endpointId`."""
        return self.endpointId

    def getDeviceModel(self):
        """Return :attr:`.deviceModel`."""
        return self.deviceModel

    def setAttributeValue(self, attribute, value):
        """
        Add the `attribute` with `value` to :attr:`.attributeValueMap`.

        :param attribute: An attribute name
        :param value: the value for attribute must not be `None`
        :raises: :class:`.ArgumentExceptio` if `value` is `None` or not
            an appropriate :class:`.DeviceModelAttribute.Type` type.
        """
        if attribute is None:
            return
        if value is None:
            raise ArgumentException("value cannot be null")
        deviceModelAttribute = \
            self.deviceModel.getDeviceModelAttributes().get(attribute)
        if deviceModelAttribute is None:
            raise ArgumentException(f'setAttributeValue(): '
                                    f'{self.deviceModel.getURN()} does not contain attribute {attribute}')
        dmtype = deviceModelAttribute.getType()
        badValue = False
        if dmtype == DeviceModelAttribute.Type.NUMBER:
            badValue = not (isinstance(value, (numbers.Number, )))
        elif dmtype == DeviceModelAttribute.Type.STRING:
            pass
        elif dmtype == DeviceModelAttribute.Type.URI:
            badValue = not (isinstance(value, (str, )))
        elif dmtype == DeviceModelAttribute.Type.BOOLEAN:
            badValue = not (isinstance(value, (bool, )))
        elif dmtype == DeviceModelAttribute.Type.INTEGER:
            badValue = not (isinstance(value, (int, )))
        elif dmtype == DeviceModelAttribute.Type.DATETIME:
            badValue = not (isinstance(value, (int, )))
        else:
            raise ArgumentException(f'setAttributeValue(): '
                                    f'unknown type "{dmtype}"')
        if badValue:
            raise ArgumentException(f'setAttributeVale(): '
                                    f'cannot set "{self.deviceModel.getURN()}:attribute/{attribute}" to {value}')
        self.attributeValueMap[attribute] = value

    def getAttributeValue(self, attribute):
        """
        Return the value of `attribute`.

        :param attribute:
        :return: the value of `attribute` or its default value.
        """
        deviceModelAttribute = None
        if attribute in self.deviceModel.getDeviceModelAttributes():
            deviceModelAttribute = \
                self.deviceModel.getDeviceModelAttributes()[attribute]
        if deviceModelAttribute is None:
            raise ArgumentException(f'getAttributeValue(): '
                                    f'{self.deviceModel.getURN()} does not contain attribute {attribute}')

        value = None
        if attribute in self.attributeValueMap:
            value = self.attributeValueMap[attribute]
        if value is None:
            value = deviceModelAttribute.getDefaultValue()
        return value

    def call(self, actionName, args):
        """
        Invoke the device model action `actionName`.

        :param actionName: a device model action
        :param args: a `list` of action arguments
        """
        deviceModelActionMap = self.deviceModel.getDeviceModelActions()
        if deviceModelActionMap == None:
            self.getLogger().warning(f'call(): {self.deviceModel.getURN()} '
                                     f'does not contain action "{actionName}"')
            return
        deviceModelAction = None
        if actionName in deviceModelActionMap:
            deviceModelAction = deviceModelActionMap[actionName]
        if deviceModelAction == None:
            self.getLogger().warning(f'call(): {self.deviceModel.getURN()} '
                                     f'does not contain action "{actionName}"')
            return

        # What this is doing is pushing a request message into the message
        # queue for the requested action. To the LL, it is handled just like
        # any other RequestMessage. But we don't want this RM going to the
        # server, so the source and destination are set to be the same
        # endpoint id. In SendReceiveImpl, if there is a RequestMessage
        # with a source == destination, it is treated it specially.
        requestMessage = RequestMessage(source=self.getEndpointId(),
                                        destination=self.getEndpointId(),
                                        url=f'deviceModels/{self.getDeviceModel().getURN()}/actions/{actionName}',
                                        method="POST")

        argumentList = deviceModelAction.getArguments()
        try:
            actionArgs = dict()
            for argument in argumentList:
                # deal with this now, instead of in validate.
                if argument.getArgType() is None:
                    self.getLogger().warning(
                        f'call(): Action {actionName} not called, argument type is None')
                    return
                value = None
                if argument.getName() in args:
                    value = args[argument.getName()]
                if value is None:
                    value = argument.getDefaultValue()
                    if value is None:
                        self.getLogger().warning(
                            f'call(): Action {actionName} not called, missing argument {argument.getName()}')
                        return
                # value is some value
                if not argument.validate(value):
                    self.getLogger().warning(
                        f'call(): Action {actionName} not called, value {value} for argument {argument.getName()}'
                        f' is not {argument.getArgType().name}')
                    return
                # check bounds
                if argument.getArgType() is DeviceModelAttribute.Type.NUMBER or \
                        argument.getArgType() is DeviceModelAttribute.Type.INTEGER:
                    try:
                        argument.checkBounds(value)
                    except Exception as e:
                        self.getLogger().warning(
                            f'call(): Action {actionName} {e}')
                        return
                # convert datetime
                if argument.getArgType() is DeviceModelAttribute.Type.DATETIME:
                    # If a datetime covert to float
                    if isinstance(value, (datetime.datetime, )):
                        value = int(value.timestamp * 1000)
                    else:
                        # its already an int, make it float
                        value = float(value)

                # Even if validated, a URI attribute in this context must be ExternalObject,
                # validate accepts str or ExternalObject
                if argument.getArgType() is DeviceModelAttribute.Type.URI and value is not ExternalObject:
                    self.getLogger().warning(
                        f'call(): Action {actionName} not called, value for argument {argument.getName()}'
                        f' is not an ExternalObject {argument.getArgType().name}')
                    return

                actionArgs[argument.getName()] = value

            # RequestMessage setBody converts pyton to json and then to bytearray
            requestMessage.setBody(actionArgs)


        except Exception as e:
            self.getLogger().warning(
                f'call(): Action {actionName} caught exception {e}')
            return

        useLongPolling = not Util.getConfig().getboolean("device_messaging",
                                                         "disable_long_polling",
                                                         False)
        if useLongPolling:
            try:
                responseMessage = RequestDispatcher.getInstance().dispatch(requestMessage)
                if responseMessage.getStatusCode() == StatusCode.NOT_FOUND:
                    self.getLogger().info(f'call(): Endpoint '
                                          f'{self.getEndpointId()} has no handler for '
                                          f'{requestMessage.getURL()}')
            except Exception as e:
                self.getLogger().warning(f'call(): {e}')
        else:
            try:
                self.directlyConnectedDevice.send(requestMessage)
            except Exception as e:
                self.getLogger().warning(f'call(): {e}')


    def queueMessage(self, message):
        """
        Queue a message.
        :param message: the message to queue.
        """
        try:
            self.directlyConnectedDevice.offer(message)
        except Exception as e:
            self.getLogger().info(f'queue(): {e}')


    @classmethod
    def getLogger(cls):
        """Return a logger for :class:`.DeviceAnalogImpl`."""
        return logging.getLogger('iotcs.messaging.client.device.DeviceAnalogImpl')
# com.oracle.iot.client.impl.device

class DevicePolicy():
    """Detailed information on a device policy."""

    class Function():
        """A cLass to encapsulate :class:.`DeviceFunction`s."""
        __slots__ = (
            'id',
            # The :class:`.DeviceFunction` id.
            'parameters'
            # The :class:`.DeviceFunction` parameters.
        )

        def __init__(self, id, parameters):
            """
            Create an instance of :class:`.Function`

            :param id: the function id
            :param parameters: the function parameters
            """
            self.id = id
            if parameters is not None and len(parameters) != 0:
                self.parameters = deepcopy(parameters)
            else:
                self.parameters = dict()

        def getId(self):
            """Return the function id."""
            return self.id

        def getParameters(self):
            """Return the function parameters."""
            return self.parameters


        def __eq__(self, other):
            """
            Return `True` if `other` equals `self`.
            :param other: the `Function` instance to compare.
            :return: `True` if `other` equals `self`.
            """
            if self is other:
                return True
            if self.__class__ != other.__class__:
                return False
            # TODO: What about parameters ?
            return self.id == other.id

    __slots__ = (
        'id',
        # The :class:`.DevicePolicy` id
        'deviceModelURN',
        # The :class:`.DeviceModel` urn of this policy
        'pipelines',
        # The policy function pipelines.
        'description',
        # The policy description.
        'enabled',
        # The if `enabled` is `True` the policy is in effect, if `False` it is
        # not in effect.
        'lastModified'
        # The date the policy was last modified.
    )

    ALL_ATTRIBUTES = "*"

    def __init__(self, id, deviceModelURN, pipelines, description,
                 lastModified, enabled):
        """
        Create a :class:`.DevicePolicy` instance.

        :param id: the policy id
        :param deviceModelURN: the :classs:`.DeviceModel` urn
        :param pipelines: the policy function pipelines
        :param description: the policy description
        :param lastModified: the policy's last modification date
        :param enabled: `True` if the policy is in effect, else `False`
        """
        self.id = id
        self.deviceModelURN = deviceModelURN
        if pipelines is not None:
            self.pipelines = deepcopy(pipelines)
        else:
            self.pipelines = dict()
        self.description = description
        self.lastModified = lastModified
        self.enabled = enabled if enabled is not None else False

    def getId(self):
        """
        Return the policy id.

        :return: the policy id
        """
        return self.id

    def getDeviceModelURN(self):
        """
        Return the target device model URN.

        :return: the URN of the target device model
        """
        return self.deviceModelURN

    def getPipeline(self, attributeName):
        """
        Return the function pipeline of this policy for `attribute`.

        If `attributeName` is `None` return the piplines for all attributes.
        See :attrr:`.ALL_ATTRIBUTES`

        :param attributeName: a device model attibute
        :return: the function pipeline of this policy for `attribute`
        """
        if attributeName is None:
            attributeName = DevicePolicy.ALL_ATTRIBUTES

        if attributeName in self.pipelines:
            return self.pipelines[attributeName]

    def getPipelines(self):
        """
        Return the pipelines of this policy.

        The map key is an attribute name, the value is a `list` of `Function`
        for that attribute.

        :return: the pipelines of this policy
        """
        return self.pipelines

    # 
    #      * Get
    #      *
    #      * @return the description of the model
    #      
    def getDescription(self):
        """Return the free form description of the device policy."""
        return self.description

    # 
    #      * Get the {@code enabled} state of the device policy
    #      * @return {@code true} if the policy is enabled, {@code false} otherwise
    #      
    def isEnabled(self):
        """
        Return the `enabled` state of the device policy.

        :return: `True` if the policy is in effect else `False`
        """
        return self.enabled

    def getLastModified(self):
        """
        Return the date of last modification.

        :return: the date of last modification
        """
        return self.lastModified

    def __eq__(self, other):
        """
        Return `True` if `other` equals `self`.
        :param other: the :class:`.DevicePolicy` instance to compare.
        :return: `True` if `other` equals `self`.
        """
        if self is other:
            return True
        if self.__class__ != other.__class__:
            return False
        return self.id == other.id

    def __str__(self):
        """Return a string representation of this policy."""
        dumpstring = f'{self.getDeviceModelURN()}\n"pipelines":\n['
        pipelines = self.getPipelines()
        if pipelines is not None and len(pipelines) != 0:
            first = True
            for key,value in pipelines.items():
                dumpstring = self.dumpPipeline(dumpstring, key,
                                               value, first)
                first = False

        dumpstring = f'{dumpstring}\n]'
        return dumpstring

    def dumpPipeline(self, dumpstring, attributeName, pipeline, first):
        """
        Return a string representation of a pipeline.

        :param dumpstring: the pipelines string representation
        :param attributeName: the attribute association with a pipeline
        :param pipeline: the `Function` pipeline
        :param first: append a comma if `first` is `False`
        """
        pstr = f'\n{{\n    "attributeName":"{attributeName}",'\
                      f'\n    "pipeline":\n    ['
        if not first:
            pstr = f'{pstr},'

        pstr = f'{pstr}\n'\
            f'  {{\n'\
            f'    "attributeName":"{attributeName}",\n'\
            f'    "pipeline":\n'\
            f'    ['

        if pipeline is not None and len(pipeline) != 0:
            for deviceFunction in pipeline:
                id = deviceFunction.getId()
                parameters = deviceFunction.getParameters()
                deviceFunction = DeviceFunction.getDeviceFunction(id)
                if deviceFunction is not None:
                    pstr = \
                        f'{pstr}\n       {deviceFunction.getDetails(parameters)}'
                else:
                    pstr = f'{pstr}`{self.id}` not found!'

            pstr = f'{pstr}\n'

        return pstr


# com.oracle.iot.client.impl.device

class DevicePolicyManager():
    """
    Policy manager for /iot/privateapi/v2/devicePolicies.
    """
    CAPABILITY_URN ="urn:oracle:iot:dcd:capability:device_policy"
    PATH = f'deviceModels/{CAPABILITY_URN}/policyChanged'

    class ChangeListener(ABC):
        """
        An interface to respond notification assigned or unassigned policies.
        """
        @abstractmethod
        def policyAssigned(self, devicePolicy, assignedDevices):
            """
            Called if a policy has been assigned to some devices.
            
            The listener should check to see if the device is in the set
            of assigned devices. If the device is not in the set of assigned
            devices, the listener should take no action.

            param devicePolicy: the policies that have been assigned.
            param assignedDevices: the set of devices that to which the 
                policy has been assigned.
            """

        @abstractmethod
        def policyUnassigned(self, devicePolicy, unassignedDevices):
            """
            Called if a policy has been unassigned from some devices.
            
            The listener should check to see if the device is in the set of unassigned devices. If
            the device is not in the set of unassigned devices, the listener should take no action.
            :param devicePolicy: the policies that have been unassigned.
            :param unassignedDevices: the set of devices from which the policy has been unassigned.
            """

    # TODO: Consolidate operations on XbyX data structures to simplify
    # TODO: locking semantics, or ensure that they are not modified unless
    # TODO: within the policyChanged lock.
    # TODO: note that although a XbyX data structure is accessed within
    # TODO: a lock, the acquired data member is subject to modification
    # TODO: while outside the lock.

    __slots__ = (
        'changelistenerslock',
        # A lock to control listeners map.,
        'changelisteners',
        'policiesbydeviceidlock',
        'policiesbydeviceid',
        # A device id to a map of device model urn to a policy id.
        # *{<device-id>:{<device-model-urn>:<policy-id>}}*
        'policychangelock',
        'policiesbypolicyid',
        # A policy map, mapping a device policy id to its
        # :class:`.DevicePolicy`. *{<policy-id>:<device-policy>}*
        'policiesbypolicyidlock',
        'policiesbydevicemodelurn',
        # A policy map, mapping a device model urn to a map, mapping a policy
        # id to a set of device id's.
        # *{<device-model-urn>>:{<policy-id>:[<device-id>,...]}*
        'policiesbydevicemodelurnlock',
        'secureConnection'
    )

    PERSISTENCE_LOCK = threading.RLock()
    # A lock to synchronize the "device-associations" file and locally
    # persistend policies

    def __init__(self, secureConnection):
        """Return a :class:`.DevicePolicyManager` instance."""

        self.secureConnection = secureConnection
        self.policiesbydeviceid = dict()
        self.policiesbydeviceidlock = threading.RLock()
        self.policiesbypolicyid = dict()
        self.policiesbypolicyidlock = threading.RLock()
        self.policiesbydevicemodelurn = dict()
        self.policiesbydevicemodelurnlock = threading.RLock()
        self.changelisteners = []
        self.changelistenerslock = threading.RLock()
        self.policychangelock = threading.RLock()

    def addChangeListener(self, changelistener):
        """
        Add a listener, called when policies are assigned or unassigned.
        """
        if changelistener is None:
            return
        with self.changelistenerslock:
            self.changelisteners.append(changelistener)

    def removeChangeListener(self, changelistener):
        """
        Remove a change listener.
        """
        if changelistener is None:
            return
        with self.changelistenerslock:
            self.changelisteners.remove(changelistener)

    def notifyPolicyAssigned(self, devicePolicy, assignedDevices):
        """
        Invoke :meth:`.policyAssigned` method on change listeners.

        :param devicePolicy: the assigned policy
        :param assignedDevices: the devices to which the policy was assigned.
        """
        
        if devicePolicy is None or len(assignedDevices) == 0:
            return
        with self.changelistenerslock:
            for changelistener in self.changelisteners:
                try:
                    changelistener.policyAssigned(devicePolicy, assignedDevices)
                except Exception as e:
                    #  listener may throw exception.
                    # SEVERE
                    self.getLogger().critical(f'notifyPolicyAssigned(): '
                        f'changeListener.policyAssigned() raised exception {e}')

    def notifyPolicyUnassigned(self, devicePolicy, unassignedDevices):
        """
        Invoke :meth:`.policyUnassigned` method on change listeners.

        :param devicePolicy: the assigned policy
        :param unassignedDevices: the devices to which the policy was assigned.
        """

        if devicePolicy is None or len(unassignedDevices) == 0:
            return
        with self.changelistenerslock:
            for changelistener in self.changelisteners:
                try:
                    changelistener.policyUnassigned(devicePolicy, unassignedDevices)
                except Exception as e:
                    #  listener may throw exception.
                    self.getLogger().critical(f'notifyPolicyUnassigned(): '
                        f'changeListener.policyUnassigned() raised exception {e}')

    @classmethod
    def getDevicePolicyManager(cls, directlyConnectedDevice):
        """
        Return a :class:`.DevicePolicyManager` for `directlyConnectedDevice`.

        :param directlyConnectedDevice; the directly connected device
            associated with the returned :class:`.DevicePolicyManager`
        :return: Return a :class:`.DevicePolicyManager` for
            `directlyConnectedDevice` else `None`
            if there no policy manager registerd for this client.
        """
        persistenceStore = PersistenceStoreManager.getPersistenceStore(
            directlyConnectedDevice.getEndpointId())
        dpmObj = persistenceStore.get(DevicePolicyManager.__name__, None)
        if dpmObj is None:
            # SEVERE
            cls.getLogger().critical(f'getDevicePolicyManager(): '
                f'cannot access DevicePolicyManager for {directlyConnectedDevice.getEndpointId()}')
            return None
        return dpmObj

    def getPolicy(self, deviceModelURN, deviceId):
        """
        Return the :classs:`.DevicePolicy` for the given device model and device id.

        :param deviceModelURN the device model URN, which may not be {@code null}
        :param deviceId the device id, which may not be {@code null}
        :return a {@code DevicePolicy}, possibly {@code null}
        """
        #  If we already have the policy for this device model and device, return it.
        #  Otherwise, look it up. If there isn't an entry in policiesbydeviceid
        #  for the device model urn and device id when this method is called,
        #  there will be by the time this method completes.
        # 
        with self.policiesbydeviceidlock:
            #  policiesbydeviceid looks like { <device-id> : { <device-model-urn> : <policy-id> }}
            devicePolicies = None
            if deviceId in self.policiesbydeviceid:
                devicePolicies = self.policiesbydeviceid[deviceId]
            #  If devicePolicies is null, we drop through and do the lookup.
            if devicePolicies is not None:
                # 
                #  If the deviceModelURN is not found in the map, we drop through and do the lookup.
                #  There may be a mapping for the device model urn, but the value may be null,
                #  which means that there is no policy for the combination of device model and device.
                # 
                if deviceModelURN in devicePolicies:
                    policyId = devicePolicies[deviceModelURN]
                    devicePolicy = None
                    if policyId is not None:
                        #  Very important that the client has gotten
                        #  the policy before we get here!
                        with self.policiesbypolicyidlock:
                            if policyId in self.policiesbypolicyid:
                                devicePolicy = \
                                    self.policiesbypolicyid[policyId]
                    else:
                        devicePolicy = None
                    return devicePolicy
        #  stop policyChanged while doing this lookup.
        with self.policychangelock:
            #  If we get to here, then there was no mapping for the deviceId
            # in policiesbydeviceid, or there was a mapping for the deviceId,
            # but not for the device model. So we need to do a lookup and
            # update policiesbydeviceid
            devicePolicy = self.lookupPolicyForDevice(deviceModelURN, deviceId)
            #  Add the mapping so the code doesn't try to fetch a policy for
            # this device again. The only way the device will get a policy
            # after this is from an "assigned" policyChanged, or when the
            # client restarts.
            policies = None
            with self.policiesbydeviceidlock:
                if deviceId in self.policiesbydeviceid:
                    policies = self.policiesbydeviceid[deviceId]
                if policies is None:
                    policies = dict()
                    self.policiesbydeviceid[deviceId] = policies
                # Note: devicePolicy may be null, but the entry is made
                # anyway.This just means the device has no policy for this
                # device model. Adding null prevents another lookup.
                policyId = devicePolicy.getId() \
                    if devicePolicy is not None else None
                policies[deviceModelURN] = policyId
            if devicePolicy is not None:
                assignedDevices = set()
                assignedDevices.add(deviceId)
                self.notifyPolicyAssigned(devicePolicy, assignedDevices)
            return devicePolicy

    def policyChanged(self, directlyConnectedDevice, requestMessage):
        """
        Handle the policy changed resource request.

        *deviceModels/urn:oracle:iot:dcd:capability:device_policy/policyChanged*

        :param directlyConnectedDevice: the device client
        :param requestMessage: the RequestMessage from the server
        :raise: Exception if there is an exception from handling the request
        """

        # 
        # The server will not send a policyChanged to a device for which the
        # policy is not intended. If this is a DCD, then the policy  is meant
        # for this DCD. If this is a GW, then the policy is meant for one or
        # more of its ICDs.
        # 
        #  RequestMessage body looks like:
        #  [{
        #     "deviceModelUrn": "urn:com:oracle:iot:device:humidity_sensor",
        #     "id": "547B66F3-5DC8-4F60-835F-7B7773C8EE7A",
        #     "lastModified": 1511805927387,
        #     "op": "changed"
        #  }]
        # 
        #  Where op is:
        #    "changed"    - The policy pipeline was changed. The client needs
        #                   to GET the policy.
        #    "assigned"   - The policy was assigned to device(s). The policy
        #                   pipeline itself has not changed. The server will
        #                   not send this to the client unless the client has
        #                   the device(s). A gateway client needs to get a list
        #                   of devices the policy applies to, but a directly
        #                   connected device can assume the policy is for it.
        #                   If necessary, the client will GET the policy.
        #    "unassigned" - The policy was unassigned from device(s). The
        #                   policy pipeline itself has not changed. The server
        #                   will not send this to the client unless the client
        #                   has the device(s). A gateway client needs to get
        #                   a new list of devices the policy applies to, but a
        #                   directly connected device can assume the policy is
        #                   for it.
        #
        # Need to make sure that "isGatewayDevice" is always set appropriately.
        # Cannot use isintance GatewayDevice, because of circular import.
        # check to see if the "isGatewayDevice" attribute is exists on the
        # dcd, if it does and is True it's a gateway, else it's not.
        # Can't add a method, since JCL does not have such a method.
        dcdisgatewaydevice = getattr(directlyConnectedDevice,
                                     'isGatewayDevice', False)
        endpointId = directlyConnectedDevice.getEndpointId()
        body = requestMessage.getBody()
        if body is None:
            self.getLogger().warning(f'policyChanged(): requestMessage body is None')
            return
        #  block second half of getPolicy while doing an update.
        with self.policychangelock:
            try:
                items = json.loads(body)
                for item in items:
                    op = "changed"
                    if "op" in item:
                        op = item["op"]
                    deviceModelUrn = None
                    if "deviceModelUrn" in item:
                        deviceModelUrn = item["deviceModelUrn"]
                    policyId = None
                    if "id" in item:
                        policyId = item["id"]
                    lastModified = None
                    if "lastModified" in item:
                        lastModified = item["lastModified"]

                    if "unassigned" == op:
                        self.processUnassign(deviceModelUrn, policyId,
                                             endpointId, dcdisgatewaydevice,
                                             lastModified)
                    elif "assigned" == op:
                        self.processAssign(deviceModelUrn, policyId,
                                           endpointId, dcdisgatewaydevice,
                                           lastModified)
                    elif "changed" == op:
                        policyBeforeChange = None
                        with self.policiesbypolicyidlock:
                            if policyId in self.policiesbypolicyid:
                                policyBeforeChange = \
                                    self.policiesbypolicyid[policyId]
                        # If device policy is null, then something is wrong
                        # in our mappings. Remove the references to this device
                        #  model urn an policy id.
                        policies = None
                        if policyBeforeChange is None:
                            # Remove the policy id from the map, nothing good
                            # can come of having a policyId without a policy.
                            with self.policiesbypolicyidlock:
                                del(self.policiesbypolicyid[policyId])
                            with self.policiesbydevicemodelurnlock:
                                if deviceModelUrn in self.policiesbydevicemodelurn:
                                    policies = \
                                        self.policiesbydevicemodelurn[deviceModelUrn]
                            if policies is not None:
                                assignedDevices = None
                                if policyId in policies:
                                    assignedDevices = policies.pop(policyId)
                                if assignedDevices is not None:
                                    with self.policiesbydeviceidlock:
                                        for deviceId in assignedDevices:
                                            if deviceId not in self.policiesbydeviceid:
                                                continue
                                            devicePolicies = \
                                                self.policiesbydeviceid[deviceId]
                                            if devicePolicies is not None:
                                                del(devicePolicies[policyId])
                            continue 
                        # Before updating the policy, notify the devices the
                        # policy is unassigned. This gives the code in
                        # VirtualDeviceImpl or MessagingPolicyImpl a chance
                        # to clean up the exisiting pipeline before the new
                        # pipeline comes in.
                        assignedDevices = None
                        with self.policiesbydevicemodelurnlock:
                            if deviceModelUrn in self.policiesbydevicemodelurn:
                                policies = \
                                    self.policiesbydevicemodelurn[deviceModelUrn]
                            if policies is not None:
                                if policyId in policies:
                                    assignedDevices = policies[policyId]
                            if assignedDevices is not None:
                                if policyBeforeChange is not None:
                                    self.notifyPolicyUnassigned(policyBeforeChange,
                                                                assignedDevices)
                            self.processPipelineChanged(directlyConnectedDevice,
                                                        deviceModelUrn,
                                                        policyId, lastModified)
                            if assignedDevices is not None:
                                policyAfterChange = None
                                with self.policiesbypolicyidlock:
                                    if policyId in self.policiesbypolicyid:
                                        policyAfterChange = self.policiesbypolicyid[policyId]
                                if policyAfterChange is not None:
                                    self.notifyPolicyAssigned(policyAfterChange,
                                                              assignedDevices)
                    else:
                        self.getLogger().warning(f'policyChanged(): '
                        f'{requestMessage.getURL()} invalid operation: {item}')
            except Exception as e:
                self.getLogger().critical(f'policyChanged(): '
                    f'{requestMessage.getURL()} body={body} : {e}')
                responseMessage = ResponseMessage(requestMessage=requestMessage,
                                                  statusCode=StatusCode.BAD_REQUEST,
                                                  body=f'{e}'.encode(encoding='utf-8'))
                return responseMessage

        responseMessage = ResponseMessage(requestMessage=requestMessage,
                                          statusCode=StatusCode.OK)
        return responseMessage

    def processPipelineChanged(self, directlyConnectedDevice, deviceModelUrn,
                               policyId, lastModified):
        """
        Process a change to a policy pipeline.

        The method needs to fetch the policy.
        :param directlyConnectedDevice: the directly connected device affected
            by the policy
        :param deviceModelUrn: the device model urn of the policy that has
            changed
        :param policyId: the id of the policy that has changed
        :param lastModified: the date of the modification
        :raises: TransportException from :class:`.SecureConnection`
        :raises: SecurityException from :class:`.SecureConnection`
        """

        # first, check to see if we have a copy,
        # and if so, whether or not it is more recent.
        with self.policiesbypolicyidlock:
            currentDevicePolicy = self.policiesbypolicyid.get(policyId)
        if currentDevicePolicy is not None:
            if lastModified < currentDevicePolicy.getLastModified():
                # our copy is more recent, return.
                return
        devicePolicy = None
        # Download the policy
        # block getPolicy from accessing policiesByDeviceId while the policy
        # is being updated.
        with self.policiesbydeviceidlock:
            devicePolicy = self.downloadPolicy(deviceModelUrn, policyId)
        if devicePolicy is not None:
            # FINE
            self.getLogger().debug("processPipelineChanged(): " +
                directlyConnectedDevice.getEndpointId() +
                " : Policy changed : \"" + devicePolicy.__str__())

    def processAssign(self, deviceModelUrn, policyId, endpointId,
                      dcdisgatewaydevice, lastModified):
        """
        Assign a policy to a device.

        Process the "assign" operation from policyChanged.
        The method needs to notify listeners that the policy was assigned,
        then update data structures and persistence to add the association
        of the policy to the device.

        :param deviceModelUrn: the device model urn of the policy
            being assigned
        :param policyId: the id of the policy being assigned.
        :param endpointId: the endpointId of the device receiving the
            policy assignment.
        :param dcdisgatewaydevice: `True` if the device is a gateway, else
            `False`
        :param lastModified: the date of the modification
        """

        # get the set of devices to which this policy is assigned.
        # If the directly connected device is a gateway, then get
        # the assigned devices from the server. Otherwise, the
        # assigned device is the directly connected device with endpointId.
        #
        # Note that if the call to get the ICD ids that are assigned
        # to the policy throws an exception, unassign every device
        # with that device model urn. In such a case, we  remove the
        # mapping from device id the device-model urn from
        # policiesByDeviceId. If there is no mapping, getPolicy
        # will try to create a new mapping by getting the policy
        # for the device from the server.
        # set()
        assignedDevices = None
        if dcdisgatewaydevice:
            couldNotGetIcds = True
            try:
                assignedDevices = \
                    self.getIndirectlyConnectedDeviceIdsForPolicy(
                        deviceModelUrn, policyId, endpointId)
                couldNotGetIcds = False
            except IOError as e:
                pass
            except SecurityException as e:
                pass
            finally:
                if couldNotGetIcds:
                    # remove the mappings for all devices that reference this policy
                    # since we no longer know which policy they refer to. On next call
                    # to getPolicy, this should self-correct.
                    # {<policy-id>:[<device-id>,...]} [device-id] is a set()
                    deviceModelPolicies = None
                    with self.policiesbydevicemodelurnlock:
                        deviceModelPolicies = \
                            self.policiesbydevicemodelurn.get(deviceModelUrn)
                    if deviceModelPolicies is not None:
                        assignedDeviceIds = None
                        if policyId in deviceModelPolicies:
                            assignedDeviceIds = deviceModelPolicies.pop(policyId)
                        if assignedDeviceIds is not None:
                            for deviceId in assignedDeviceIds:
                                # {<device-id>:{<device-model-urn>:<policy-id>}}
                                assignedPolicies = None
                                with self.policiesbydeviceidlock:
                                    if deviceId in self.policiesbydeviceid:
                                        assignedPolicies = \
                                            self.policiesbydeviceid[deviceId]
                                if assignedPolicies is not None:
                                    if deviceModelUrn in assignedPolicies:
                                        del(assignedPolicies[deviceModelUrn])
                    return
        else:
            assignedDevices = set()
            assignedDevices.add(endpointId)
        if assignedDevices is None or len(assignedDevices) == 0:
            return
        devicePolicy = None
        with self.policiesbydeviceidlock:
            # Download the policy. The reason we have to download again
            # on assign is that the policy may have been modified
            # while it was not assigned to this device or ICDs.
            newPolicy = self.downloadPolicy(deviceModelUrn, policyId)
            with self.policiesbypolicyidlock:
                self.policiesbypolicyid[policyId] = newPolicy
                for deviceId in assignedDevices:
                    self.assignPolicyToDevice(deviceModelUrn, policyId,
                                              deviceId, lastModified)
                if policyId in self.policiesbypolicyid:
                    devicePolicy = self.policiesbypolicyid[policyId]

        if devicePolicy is not None:
            self.notifyPolicyAssigned(devicePolicy, assignedDevices)

    def assignPolicyToDevice(self, deviceModelUrn, policyId, deviceId,
                             lastModified):
        """
        Assign a policy to a device.

        :param deviceModelUrn: the device model urn of the policy
            being assigned
        :param policyId: the id of the policy that being assigned.
        :param deviceId: the device id receiving the policy assigned.
        :param lastModified: the date of last modification
        """

        policies = None
        currentPolicyId = None

        # If the device has the policy assigned,
        # it needs to be unassigned.
        with self.policiesbydeviceidlock:
            # *{<device-id>:{<device-model-urn>:<policy-id>}}*
            if deviceId in self.policiesbydeviceid:
                policies = self.policiesbydeviceid[deviceId]

        if policies is not None:
            if deviceModelUrn in policies:
                currentPolicyId = policies[deviceModelUrn]
        else:
            currentPolicyId = None

        # if the current policy is the same as the policy that is being
        # assigned, no need to continue.
        if policyId == currentPolicyId:
            return

        # Make sure we have the policy
        # before assigning the new policy.
        devicePolicy = None
        with self.policiesbypolicyidlock:
            if policyId in self.policiesbypolicyid:
                devicePolicy = self.policiesbypolicyid[policyId]

        if devicePolicy is None:
            devicePolicy = self.downloadPolicy(deviceModelUrn, policyId)
        # TODO: can devicePolicy be None at this point ?
        self.policiesbypolicyid[policyId] = devicePolicy
        if devicePolicy is not None:
            # FINE
            self.getLogger().debug(
                f'assignPolicyToDevice(): {deviceId} : Policy : '
                f'{devicePolicy.getId()}\n{devicePolicy.__str__()}')

        if devicePolicy is not None:
            devicePolicies = None
            with self.policiesbydeviceidlock:
                if deviceId in self.policiesbydeviceid:
                    devicePolicies = self.policiesbydeviceid[deviceId]
                if devicePolicies is None:
                    devicePolicies = dict()
                    self.policiesbydeviceid[deviceId] = devicePolicies
                devicePolicies[deviceModelUrn] =  policyId
                with self.policiesbydevicemodelurnlock:
                    deviceModelPolicies = None
                    if deviceModelUrn in self.policiesbydevicemodelurn:
                        deviceModelPolicies = \
                            self.policiesbydevicemodelurn[deviceModelUrn]
                    if deviceModelPolicies is None:
                        deviceModelPolicies = dict()
                        self.policiesbydevicemodelurn[deviceModelUrn] = \
                            deviceModelPolicies
                    assignedDevices = None
                    if policyId in deviceModelPolicies:
                        assignedDevices = deviceModelPolicies[policyId]
                    if assignedDevices is None:
                        assignedDevices = set()
                        deviceModelPolicies[policyId] = assignedDevices
                    assignedDevices.add(deviceId)

    def processUnassign(self, deviceModelUrn, policyId, endpointId,
                        dcdisgatewaydevice, lastModified):
        """
        Unassign a policy from a device.

        # Process the "unassign" operation from policyChanged.
        # The method updates the data structures and persistence to remove the association
        # of the policy to the device.
        :param deviceModelUrn: the device model urn of the policy being
            unassigned
        :param policyId: the id of the policy being unassigned
        :param endpointId: the endpointId of the device receiving the
            policy changed request
        :param dcdisgatewaydevice: `True` if the device is a
            :class:`.GatewayDevice` else `False`
        :param lastModified: the date of last modification

        """
        # Get the set of devices to which this policy is assigned. This
        # This will be the difference of the set of devices the client
        # says are assigned and the set of devices the server says are
        # assigned (the server doesn't say who was unassigned, we can
        # only ask who is assigned).
        # set()
        unassignedDevices = None
        policies = None
        with self.policiesbydevicemodelurnlock:
            if deviceModelUrn in self.policiesbydevicemodelurn:
                # *{<device-model-urn>>:{<policy-id>:[<device-id>,...]}*
                policies = self.policiesbydevicemodelurn[deviceModelUrn]
        if policies is not None:
            if policyId in policies:
                unassignedDevices = policies[policyId]
            if unassignedDevices is None:
                # client doesn't have any devices assigned to this policy
                return
        else:
            # client doesn't have any devices assigned to this policy
            return

        # If we aren't a gateway device, then make sure the
        # assigned devices contains the directly connected device
        # endpoint id, and ensure that the only element of
        # unassignedDevices is the directly connected device
        # endpont id.
        if not dcdisgatewaydevice:
            if endpointId not in unassignedDevices:
                # this endpoint is not currently assigned to the policy
                return
            unassignedDevices.clear()
            unassignedDevices.add(endpointId)

        # Now get the set of devices to which this policy is assigned,
        # according to the server. Remove the set of server assigned
        # devices from the client assigned devices so that
        # unassignedDevices is now the set of devices that have
        # been unassigned from this policy.
        #
        # If the directly connected device is not a gateway, then we
        # know that the subject of the unassign is the directly connected
        # device and there is no need to make a call to the server for
        # the assigned devices.
        #
        # Note that if the call to get the set of ICD ids that are assigned
        # to the policy might fail, throwing an exception. In this case,
        # there is no way to tell what devices belong to the policy or not.
        # To handle this situation, every device on the client that has
        # this policy will be be will unassign from the policy _and_
        # the device's mapping to the device model urn in policiesByDeviceId
        # will be removed. Removing the mapping ensures that getPolicy
        # will fetch the policy anew and the mapping will self correct.
        # The flag "couldNotGetIcds" lets us know that the call failed.
        couldNotGetIcds = dcdisgatewaydevice
        if dcdisgatewaydevice:
            try:
                serverAssignedDevices = \
                    self.getIndirectlyConnectedDeviceIdsForPolicy(
                        deviceModelUrn, policyId, endpointId)
                # returned without exception, couldNotGetIcds is false.
                couldNotGetIcds = False
                for icd in serverAssignedDevices:
                    unassignedDevices.discard(icd)
                # if unassignedDevices is empty now that we removed
                # all the ICD ids from the server, we should return
                # since there are no devices on the client affected
                # by the change.
                if len(unassignedDevices) == 0:
                    return
            except Exception as e:
                pass

        with self.policiesbypolicyidlock:
            devicePolicy = None
            if policyId in self.policiesbypolicyid:
                devicePolicy = self.policiesbypolicyid[policyId]
            # TODO: throw or do what ?
            #assert devicePolicy is not None
            self.notifyPolicyUnassigned(devicePolicy, unassignedDevices)

            with self.policiesbydeviceidlock:
                # Now unassignedDevices is the set of device ids that
                # have been unassigned from this policy.
                for deviceId in unassignedDevices:
                    self.unassignPolicyFromDevice(deviceModelUrn, policyId,
                                                  deviceId, lastModified)
                    if couldNotGetIcds:
                        # unassignPolicyFromDevice takes care of the entry in
                        # policiesByDeviceModelUrn
                        # {<device-id>:{<device-model-urn>:<policy-id>}}
                        devicePolicies = None
                        if deviceId in self.policiesbydeviceid:
                            devicePolicies = self.policiesbydeviceid[deviceId]
                        if devicePolicies is not None:
                            if deviceModelUrn in devicePolicies:
                                del(devicePolicies[deviceModelUrn])

    def unassignPolicyFromDevice(self, deviceModelUrn, policyId, deviceId,
                                 lastModified):
        """
        Unassign policy from device.

        Handle the logic for unassigning a policy from a device. The only
        reason for this method to return false is if the client has a more
        recent change than what it was told by the server.

        :param deviceModelUrn: the device model urn from which the policy is
            unassigned
        :param policyId: the policy id of the policy that is unassigned
        :param deviceId: the device id having the policy unassigned
        :param lastModified: the date of last modification
        :return: `False` if the cached policy is newer than policy being
            unassigned, else `True`
        """
        # Sanity check... does this device have the unassigned policy?
        currentPolicyId = None
        policies = None
        with self.policiesbydeviceidlock:
            # {<device-id>:{<device-model-urn>:<policy-id>}}
            if deviceId in self.policiesbydeviceid:
                policies = self.policiesbydeviceid.get(deviceId)
        if policies is not None:
            if deviceModelUrn in policies:
                currentPolicyId = policies[deviceModelUrn]
        else:
            currentPolicyId = None

        if currentPolicyId is None:
            # device doesn't have a policy id right now, move on.
            return True

        # if badMapping is set to true, the policiesByDeviceId entry for
        # the device-model urn of this device is removed. On the next
        # call to getPolicy, the map will auto-correct.
        badMapping = False
        if not policyId == currentPolicyId:
            # server is telling me to unassign a policy id
            # that the client doesn't have assigned. If
            # the policy that is assigned is newer than
            # lastModified, then the client is right and
            # we move on. Otherwise, unassign whatever
            # policy the device has and let the state
            # auto-correct on the next call to getPolicy.
            devicePolicy = None
            with self.policiesbypolicyidlock:
                if currentPolicyId in self.policiesbypolicyid:
                    devicePolicy = self.policiesbypolicyid[currentPolicyId]
                if devicePolicy is not None:
                    if devicePolicy.getLastModified() > lastModified:
                        # client info is newer, move on to the next device id
                        return False
                    badMapping = True
                else:
                    # Oh my. The device maps to some policy that
                    # the client doesn't know about. Remove the mapping
                    # and policiesByPolicyId will self correct for this
                    # device the next time getPolicy is called.
                    # Note that since devicePolicy is null, getPolicy
                    # will return null for this device and device model anyway,
                    # so taking an axe to policiesByPolicyId is okay here.
                    #
                    # policiesByDeviceId is
                    # { <device-id> : { <device-model-urn> : <policy-id> } }
                    devicePolicies = None
                    with self.policiesbydeviceidlock:
                        if deviceId in self.policiesbydeviceid:
                            devicePolicies = self.policiesbydeviceid[deviceId]
                        if devicePolicies is not None:
                            if deviceModelUrn in devicePolicies:
                                del(devicePolicies[deviceModelUrn])
                    return True

        # If the sanity check passes, then we are good to remove
        # the mapping to the device-model-urn from policiesByDeviceId
        # for this device.
        if policies is not None:
            if not badMapping:
                # If nothing is wrong in our mapping, then
                # set the current policy for this device and
                # device model urn to null. This state causes
                # getPolicy to return null without futher lookup.
                policies[deviceModelUrn] = None
            else:
                # if there was something bad in our mapping,
                # the remove the deviceModelUrn entry altogether.
                # On the next call to getPolicy for this device
                # and device model, the map will be corrected.
                if deviceModelUrn in policies:
                    del(policies[deviceModelUrn])

        return True

    def downloadPolicy(self, deviceModelUrn, policyId):
        """
        Return the :class:`.DevicePolicy` for `policyId`.

        GET iot/privateapi/v2/deviceModels/{deviceModelUrn}/devicePolicies/{policyId}.
        The gotten policy is persisted, the policiesByPolicyId map
        is updated, and an entry is made (if necessary) in the
        policiesByDeviceModelUrn map.

        :param deviceModelUrn: the device model urn
        :param policyId: the policy id
        :raise: :class:`.TransportException` if request fails
        :return: the :class:`.DevicePolicy` for `policyId`
        """
        data = None
        try:
            urn = urllib.parse.quote(deviceModelUrn)
            fields = "id,pipelines,enabled,lastModified"
            query = {'fields': fields}
            uri = f'{RestApi.V2.getPrivateRoot()}/deviceModels/{urn}/devicePolicies/{policyId}'
            data = None
            res = self.secureConnection.get(uri, query, None, None)
            status = res.getStatus()
            if status != StatusCode.OK.value:
                raise TransportException(status, uri)
            data = res.getData()
        except Exception as e:
            self.getLogger().warning(f'GET {uri}: {e}')

        if data is None or len(data) == 0:
            return None

        devicePolicy = None
        try:
            policyJson = json.loads(data)
            devicePolicy = self.devicePolicyFromJSON(deviceModelUrn,
                                                     policyJson)
            with self.policiesbypolicyidlock:
                self.policiesbypolicyid[policyId] = devicePolicy
                with self.policiesbydevicemodelurnlock:
                    policies = None
                    if deviceModelUrn in self.policiesbydevicemodelurn:
                        policies = \
                            self.policiesbydevicemodelurn[deviceModelUrn]
                    if policies is None:
                        policies = dict()
                        self.policiesbydevicemodelurn[deviceModelUrn] = \
                            policies
            if devicePolicy is not None:
                # FINE
                self.getLogger().debug(f'downloadPolicy(): {policyId} '
                    f' : Policy : {devicePolicy.getId()}\n'
                    f'{devicePolicy.__str__()}')
        except Exception as e:
            # SEVERE
            self.getLogger().critical(f'{e}')
        return devicePolicy

    def lookupPolicyForDevice(self, deviceModelUrn, deviceId):
        """
        Return a :class:`.DevicePolicy` for `deviceModelUrn` and `deviceId`
        :param deviceModelUrn:
        :param deviceId:
        :return:
        """
        policies = None
        devicePolicy = None
        with self.policiesbydevicemodelurnlock:
            # *{<device-model-urn>>:{<policy-id>:[<device-id>,...]}*
            if deviceModelUrn in self.policiesbydevicemodelurn:
                policies = self.policiesbydevicemodelurn[deviceModelUrn]
            if policies is not None:
                for policyId,deviceIds in policies.items():
                    if deviceId in deviceIds:
                        with self.policiesbydeviceidlock:
                            devicePolicy = self.policiesbypolicyid.get(policyId)
                        if devicePolicy is not None:
                            return devicePolicy

        urn = urllib.parse.quote(deviceModelUrn)
        # Looks like quoting the query or its arguments is not necessary
        devicesDotId = f'{{"devices.id":"{deviceId}"}}'
        fields = "id,pipelines,enabled,lastModified"
        query = { "q": devicesDotId, "fields": fields}
        uri = f'{RestApi.V2.getPrivateRoot()}/deviceModels/{urn}/devicePolicies'
        data = None
        try:
            res = self.secureConnection.get(uri, query)
            status = res.getStatus()
            if status != StatusCode.OK.value:
                m = f'{RestApi.V2.getPrivateRoot()}/deviceModels/{deviceModelUrn}/devicePolicies'\
                    f'{repr(query)} returned status {status}'
                raise TransportException(status, m)
            data = res.getData()
        except Exception as e:
            self.getLogger().warning(f'GET {e}')

        if data is not None and len(data) > 0:
            try:
                devicePolicies = json.loads(data)
                items = None
                if "items" in devicePolicies:
                    items = devicePolicies["items"]
                if items is not None and len(items) > 0:
                    item = items[0]
                    devicePolicy = self.devicePolicyFromJSON(deviceModelUrn, item)
            except Exception as e:
                # SEVERE
                self.getLogger().critical(f'{e}')

        if devicePolicy is None:
            return None

        policyId = devicePolicy.getId()
        with self.policiesbypolicyidlock:
            if policyId not in self.policiesbypolicyid:
                self.policiesbypolicyid[policyId] = devicePolicy
                # FINE (conditional if logging FINE)
                if devicePolicy is not None:
                    self.getLogger().debug(f'adding policy : {policyId} : Policy : '
                        f'{devicePolicy.getId()}\n{devicePolicy.__str__()}')
        with self.policiesbydevicemodelurnlock:
            if deviceModelUrn in self.policiesbydevicemodelurn:
                policies = self.policiesbydevicemodelurn[deviceModelUrn]
            if policies is None:
                policies = dict()
                self.policiesbydevicemodelurn[deviceModelUrn] = policies
            deviceIds = None
            if policyId in policies:
                deviceIds = policies[policyId]
            if deviceIds is None:
                deviceIds = set()
                policies[policyId] = deviceIds
            deviceIds.add(deviceId)

        return devicePolicy

    def getIndirectlyConnectedDeviceIdsForPolicy(self, deviceModelUrn,
                                                 policyId,
                                                 directlyConnectedOwner):
        """
        Return the ids of indirectly connected devices with assigned `policyId`.
        :param deviceModelUrn: the device model urn
        :param policyId: the policy id
        :param directlyConnectedOwner: the gateway device
        :return: a `set` of indirectly connected device ids
        """
        urn = None
        try:
            urn = urllib.parse.quote(deviceModelUrn)
        except Exception as cannot_happen:
            raise cannot_happen
        query = None
        try:
            icdFilter = \
                f'{{"directlyConnectedOwner":"{directlyConnectedOwner}"}}'
            query = { "q": icdFilter, "fields": "id"}
        except Exception as cannot_happen:
            raise cannot_happen

        uri = f'{RestApi.V2.getPrivateRoot()}/deviceModels/{urn}/devicePolicies/{policyId}/devices'
        res = self.secureConnection.get(uri, query)
        status = res.getStatus()
        if status != StatusCode.OK.value:
            self.getLogger().warning(res.getVerboseStatus("GET", uri))
        data = res.getData()
        if data is None or len(data) <= 2:
            return set()
        icdIds = set()
        try:
            items = None
            jsonObject = json.loads(data)
            if "items" in jsonObject:
                items = jsonObject["items"]
            if items is None:
                return icdIds
            for item in items:
                icdId = None
                if "id" in item:
                    icdId = item["id"]
                else:
                    continue
                icdIds.add(icdId)
        except Exception as e:
            # SEVERE
            self.getLogger().critical(f'{e}')

        return icdIds

    @classmethod
    def devicePolicyFromJSON(cls, deviceModelURN, jsonObject):
        """
        Return a :class:`.DevicePolicy` instance from `jsonObject`.

        :param deviceModelURN: the device model urn
        :param jsonObject:  a device policy json object
        :return: a :class:`.DevicePolicy` instance
        """

        # In case this is an old format policy
        if "pipelines" not in jsonObject:
            return None

        pipelineArray = jsonObject["pipelines"]
        if pipelineArray is None or 0 == len(pipelineArray):
            return None

        pipelines = dict()

        id = None
        if "id" in jsonObject:
            id = jsonObject["id"]
        description = None
        if "description" in jsonObject:
            description = jsonObject["description"]
        enabled = None
        if "enabled" in jsonObject:
            enabled = jsonObject["enabled"]
        lastModified = None
        if "lastModified" in jsonObject:
            lastModified = jsonObject["lastModified"]

        for attributePipeline in pipelineArray:
            attributeName = DevicePolicy.ALL_ATTRIBUTES
            if "attributeName" in attributePipeline:
                attributeName = attributePipeline["attributeName"]
            functions = []
            pipelines[attributeName] = functions
            pipeline = None
            if "pipeline" in attributePipeline:
                pipeline = attributePipeline["pipeline"]
            if pipeline is None:
                continue
            for devicefunction in pipeline:
                functionId = None
                if "id" in devicefunction:
                    functionId = devicefunction["id"]
                # functionId should never be None if we get here
                # but if there is no FunctionId what good is a parameterMap
                # or a DevicePolicy.Function ? just continue
                if functionId is None:
                    continue

                parameterMap = dict()
                dpfunction = DevicePolicy.Function(functionId, parameterMap)
                functions.append(dpfunction)
                # since Function does a deepcopy of parameterMap
                # get it back
                parameterMap = dpfunction.getParameters()
                parameters = None
                if "parameters" in devicefunction:
                    parameters = devicefunction["parameters"]
                # if there are no parameters continue
                if parameters is None:
                    continue
                for parameterName,parameterValue in parameters.items():
                    if "action" == parameterName:
                        actionObject = parameterValue
                        actionName = None
                        if "name" in actionObject:
                            actionName = actionObject["name"]
                        parameterMap["name"] = actionName
                        actionArguments = None
                        if "arguments" in actionObject:
                            actionArguments = actionObject["arguments"]
                        if actionArguments is not None and \
                                len(actionArguments) > 0:
                            argumentList = []
                            parameterMap["arguments"] = argumentList
                            for argument in actionArguments:
                                argumentList.append(argument)
                    elif "alert" == parameterName:
                        alertObject = parameterValue
                        urn = None
                        if "urn" in alertObject:
                            urn = alertObject["urn"]
                        parameterMap["urn"] = urn
                        fields = None
                        if "fields" in alertObject:
                            fields = alertObject["fields"]
                        fieldList = dict()
                        if fields is not None:
                            for fieldName, fieldValue in fields.items():
                                fieldList[fieldName] = fieldValue
                        parameterMap["fields"] = fieldList
                        if "severity" in alertObject:
                            parameterMap["severity"] = alertObject["severity"]
                    else:
                        parameterMap[parameterName] = parameterValue

        return DevicePolicy(id, deviceModelURN, pipelines, description,
                            lastModified, enabled)

    @classmethod
    def getLogger(cls):
        """
        Return a logger for `.DevicePolicyManager`
        
        param from: a tag for the log location.
        return: a logger for `.DevicePolicyManager`
        """
        return logging.Logger(
            f'iotcs.messaging.client.device.DevicePolicyManager')

# com.oracle.iot.client.impl.device


class MessagingPolicy(DevicePolicyManager.ChangeListener):
    """
    MessagingPolicy supports device policy messaging *offer* methods.

    The *offer* methods are :meth:`.DirectlyConnectedDevice.offer` and
    :meth:`.MessageDispatcher.offer`.
    This class uses the devicePolicy REST API.

    The :meth:`.DirectlyConnectedDevice.offer` API invokes the
    :meth:`.applyPolicies` method and then calls
    :meth:`.DirectlyConnectedDevice.send` with the resulting messages.

    The :meth:`.MessageDispatcher.offer` API invokes the
    :meth:`.applyPolicies` method and then calls
    :meth:`.MessageDispatcher.queue` with the resulting messages.
    """

    __slots__ = (
        'directlyConnectedDevice',
        'messagesFromExpiredPolicies',
        'computedMetricTriggers',
        # *{<devicemodelurn>: {<attribute>:(<triggerattributes>,...)}}*
        # Key is device model urn, value is attribute -> trigger attributes
        # (a trigger attribute is a referenced attribute in a computedMetric formula).
        'windowMap',
        # *{<deviceModelUrn:attributename:functionid>:<starttime>}*
        # deviceModelUrn:attribute:deviceFunctionId -> start time of last
        # window. For a window policy, this maps the policy target plus the
        # function to when the window started. When the attribute for a timed
        # function is in the message, we can compare this start time to the
        # elapsed time to determine if the window has expired. If the window
        # has expired, the value computed by the function is passed to the
        # remaining functions in the pipeline.
        'deviceAnalaogMap',
        # *{ <deviceModelUrn>: <DeviceAnalog> }*
        # We need more than one DeviceAnalog because the DCD can have more than
        #  one device model.
        'damaplock',
        'pipelineDataCache'
        # *{<attributename> : [<devicefunction>, ...]}*
        # Data pertaining to this virtual device and its attributes for
        # computing policies. The key is attribute name (or null for the device
        # model policies), value is a list. Each element in the list
        # corresponds to a function in the pipeline for the attribute, and the
        # map is used by the function to hold data between calls.
        # Note that there is a 1:1 correspondence between the pipeline
        # configuration data for the attribute, and the pipeline data for the
        # attribute.
        'pldclock'
    )

    def __init__(self, directlyConnectedDevice):
        """
        Return an instance of `MessagingPolicy`.

        Constructed in :class:`.DirectlyConnectedDevice`
        :param directlyConnectedDevice: the DirectlyConnectedDevice
        """
        super(MessagingPolicy, self).__init__()
        self.directlyConnectedDevice = directlyConnectedDevice
        # {<attributename> : [<devicefunction>, ...]}
        self.pipelineDataCache = dict()
        #  { <deviceModelUrn>: <DeviceAnalog> }
        self.deviceAnalogMap = dict()
        # {<deviceModelUrn:attributename:functionid>:<starttime>}
        self.windowMap = dict()
        # {<devicemodelurn>: {<attribute>:(<triggerattributes>,...)}}
        self.computedMetricTriggers = dict()
        self.messagesFromExpiredPolicies = []

        self.damaplock = threading.RLock()
        self.pldclock = threading.RLock()

    def applyPolicies(self, message):
        """
        Return a `list` of messages after applying device policies.

        This is the method that applies whatever policies there may be to the
        message. The method returns zero or more messages, depending on the
        policies that have been applied to the message. The caller is
        responsible for sending or queuing the returned messages. The data
        items in the returned are messages are possibly modified the same
        message is returned with the temperature value replaced by the
        avearage temperature. A returned message may also be one that is
        created by a policy function (such as a computedMetric). Or the
        returned messages may be messages that have been batched. If no policy
        applies to the message, the message itself is returned.
        :param message: A message of any kind
        :return: a `list` of  messages to be delivered
        :raises: :class:`.TransportException` possibly thrown during network
            communications
        :raises: :class:`.SecurityException` possibly thrown accessing or
            using the secure connection
        """
        if message is None:
            return []

        # Pass the current millisecond time down so that functions that have
        # the same or intersecting timeout get expired at the same time. This
        # ensures the attributes are grouped.
        currentTimeMillis = TimeManager.currentTimeMillis()
        resultingMessages = []
        if message.getType() == Message.Type.DATA:
            dataMessage = \
                self.applyAttributePolicies(message, currentTimeMillis)
            if dataMessage is not None:
                resultingMessages.append(dataMessage)
        else:
            resultingMessages.append(message)
        messageList = []
        if len(self.messagesFromExpiredPolicies) > 0:
            messageList.extend(self.messagesFromExpiredPolicies)
            self.messagesFromExpiredPolicies.clear()

        for m in resultingMessages:
            messagesFromDevicePolicy = \
                self.applyDevicePolicies(m, currentTimeMillis)
            messageList.extend(messagesFromDevicePolicy)

        return messageList

    def policyAssigned(self, devicePolicy, assignedDevices):
        """

        :param devicePolicy:
        :param assignedDevices:
        :return:
        """
        pass

    def policyUnassigned(self, devicePolicy, assignedDevices):
        """

        :param devicePolicy:
        :param assignedDevices:
        """
        currentTimeMillis = TimeManager.currentTimeMillis()
        messages = self.expirePolicy(devicePolicy, currentTimeMillis)
        if messages is not None and len(messages) > 0:
            self.messagesFromExpiredPolicies.extend(messages)
        #  TODO:  Need to figure out how to handle accumulated values.
        #         For now, just clear out the various maps, which
        #         effectively means "start from scratch"
        with self.damaplock:
            self.deviceAnalogMap.clear()
        with self.pldclock:
            self.pipelineDataCache.clear()
        self.computedMetricTriggers.clear()
        self.windowMap.clear()


    def applyAttributePolicies(self, dataMessage, currentTimeMillis):
        """
        Return a :class:`.DataMessage` instance after appying device policies.

        Apply policies that are targeted to an attribute
        :param dataMessage: the message to apply device policies
        :param currentTimeMillis: the current time
        :return: a :class:`.DataMessage` after applying device policies
        """
        #  A data message format cannot be null or empty
        #  (enforced in DataMessage(Builder) constructor
        format = dataMessage.getFormat()
        deviceModelUrn = format[0:len(":attributes") - len(format)]
        deviceModel = \
            self.directlyConnectedDevice.getDeviceModel(deviceModelUrn)
        if not isinstance(deviceModel, (DeviceModelImpl, )):
            return dataMessage

        endpointId = dataMessage.getSource()
        deviceAnalog = None
        if endpointId not in self.deviceAnalogMap:
            deviceAnalog = DeviceAnalogImpl(self.directlyConnectedDevice,
                                            deviceModel, endpointId)
            with self.damaplock:
                self.deviceAnalogMap[endpointId] = deviceAnalog
        else:
            deviceAnalog = self.deviceAnalogMap.get(endpointId)

        triggerMap = None
        # TODO: Need lock on triggers ?
        if deviceModelUrn not in self.computedMetricTriggers:
            # {<attribute>:<set>}
            triggerMap = dict()
            self.computedMetricTriggers[deviceModelUrn] = triggerMap
            deviceModelAttributeMap = deviceModel.getDeviceModelAttributes()
            for deviceModelAttribute in \
                    deviceModelAttributeMap.values():
                # TODO: isn't this the key in deviceModelAttributeMap ?
                attribute = deviceModelAttribute.getName()
                devicePolicyManager = \
                    DevicePolicyManager.getDevicePolicyManager(
                        self.directlyConnectedDevice)
                devicePolicy = devicePolicyManager.getPolicy(deviceModelUrn,
                                                             endpointId)
                if devicePolicy is None:
                    continue 
                pipeline = devicePolicy.getPipeline(attribute)
                if pipeline is None or len(pipeline) == 0:
                    continue

                #  If a computedMetric is the first function in the pipeline,
                #  then see if the formula refers to other attributes. If so,
                #  then try this pipeline after all others.
                plfunction = pipeline[0]
                deviceFunctionId = plfunction.getId()
                parameters = plfunction.getParameters()
                if "computedMetric" == deviceFunctionId:
                    formula = parameters["formula"]
                    triggerAttributes = set()
                    # TODO: Use Formula.getAttributeReferences
                    # returns -1 if not found
                    pos = formula.find("$(")
                    while pos != -1:
                        end = formula.find(')', pos + 1)
                        # check for '$$'
                        if pos == 0 or formula[pos - 1] != '$':
                            attr = formula[pos + len('$('): end]
                            if not attr == attribute:
                                triggerAttributes.add(attr)
                        pos = formula.indexOf("$(", end + 1)
                    if len(triggerAttributes) > 0:
                        triggerMap[attribute] = triggerAttributes
        else:
            triggerMap = self.computedMetricTriggers[deviceModelUrn]

        #  getDataItems is returns an unmodifiable list
        dataMessageDataItems = dataMessage.getDataItems()
        #  DataItems resulting from policies
        policyDataItems = []
        #  DataItems that have no policies
        skippedDataItems = []
        #  If no policies are found, we will return the original message
        noPoliciesFound = True
        for dataItem in dataMessageDataItems:
            attribute = dataItem.getKey()
            deviceModelAttribute = None
            if attribute in deviceModel.getDeviceModelAttributes():
                deviceModelAttribute = \
                    deviceModel.getDeviceModelAttributes()[attribute]
            if deviceModelAttribute is None:
                self.getLogger().info(f'applyAttributePolicies(): '
                    f'{deviceModel.getURN()} does not contain attribute {attribute}')
                skippedDataItems.append(dataItem)
                continue

            devicePolicyManager = DevicePolicyManager.getDevicePolicyManager(
                self.directlyConnectedDevice)
            devicePolicy = devicePolicyManager.getPolicy(deviceModelUrn,
                                                         endpointId)
            if devicePolicy is None:
                deviceAnalog.setAttributeValue(attribute, dataItem.getValue())
                skippedDataItems.append(dataItem)
                continue

            pipeline = devicePolicy.getPipeline(attribute)
            #  No policies for this attribute, retain the data item.
            if pipeline is None or len(pipeline) == 0:
                deviceAnalog.setAttributeValue(attribute, dataItem.getValue())
                skippedDataItems.append(dataItem)
                continue

            noPoliciesFound = False
            #  if this is a computed metric, skip it for now
            if attribute in triggerMap:
                continue

            policyDataItem = self.applyAttributePolicy(deviceAnalog, dataItem,
                                                       pipeline,
                                                       currentTimeMillis)
            if policyDataItem is not None:
                policyDataItems.append(policyDataItem)

        #  if no policies were found, return the original message
        if noPoliciesFound:
            return dataMessage
        #  if policies were found, but there are no policyDataItems
        #  and no skipped data items, then return null.
        if len(policyDataItems) == 0 and len(skippedDataItems) == 0:
            return None
        #  This looks like a good place to check for computed metrics, too.
        if len(policyDataItems) > 0:
            self.checkComputedMetrics(policyDataItems, deviceAnalog,
                                      triggerMap, currentTimeMillis)

        #  Since we can't modify the original data message, we have to create a copy.
        policyDataItems.extend(skippedDataItems)
        diagnostics = None
        if dataMessage.getDiagnostics() is not None:
            # TODO: make sure this "copy" does the right thing.
            diagnostics = dataMessage.getDiagnostics().copy()

        newDataMessage = \
            DataMessage(format=format, clientId=dataMessage.getClientId(),
                        source=dataMessage.getSource(),
                        destination=dataMessage.getDestination(),
                        priority=dataMessage.getPriority(),
                        reliability=dataMessage.getReliability(),
                        eventTime=dataMessage.getEventTime(),
                        properties=dataMessage.getProperties(),
                        sender=dataMessage.getSender(),
                        diagnostics=diagnostics,
                        items=policyDataItems)

        return newDataMessage

    def applyAttributePolicy(self, deviceAnalog, dataItem, pipeline, currentTimeMillis):
        """
        Return :class:`.DataItem if it should be included in the Message.

        :param deviceAnalog:
        :param dataItem:
        :param pipeline:
        :param currentTimeMillis:
        :return: :class:`.DataItem if it should be included in the Message,
            else `None`
        """
        attribute = dataItem.getKey()
        policyValue = dataItem.getValue()
        deviceModel = deviceAnalog.getDeviceModel()
        pipelineData = None
        with self.pldclock:
            if attribute in self.pipelineDataCache:
                pipelineData = self.pipelineDataCache[attribute]
            # Create the pipeline data for this attribute
            if pipelineData is None:
                pipelineData = []
                self.pipelineDataCache[attribute] = pipelineData

        DeviceFunction.putInProcessValue(deviceAnalog.getEndpointId(),
                                         deviceModel.getURN(), attribute,
                                         policyValue)
        index = 0
        maxIndex = len(pipeline)
        while index < maxIndex:
            plfunction = pipeline.get(index)
            functionData = None
            if index < len(pipelineData):
                functionData = pipelineData.get(index)
            else:
                functionData = dict()
                pipelineData.add(functionData)
            key = plfunction.getId()
            parameters = plfunction.getParameters()
            deviceFunction = DeviceFunction.getDeviceFunction(key)
            if deviceFunction is None:
                index += 1
                continue 
            windowExpired = bool()
            window = DeviceFunction.getWindow(parameters)
            slide = DeviceFunction.getSlide(parameters, window)
            if window > 0:
                # This could be more succinct, but it makes the key easy to
                # read in the debugger.
                k = f'{deviceModel.getURN()}:{attribute}:{deviceFunction.getId()}.expiry'
                expiry = None
                if k in self.windowMap:
                    expiry = self.windowMap[k]
                if expiry is None:
                    expiry = currentTimeMillis + window
                    self.windowMap[k] = expiry
                windowExpired = expiry <= currentTimeMillis
                if windowExpired:
                    self.windowMap[k] = expiry + slide
            else:
                windowExpired = False

            if deviceFunction.apply(deviceAnalog, attribute, parameters,
                                    functionData, policyValue) or \
                    windowExpired:
                valueFromPolicy = deviceFunction.get(deviceAnalog, attribute,
                                                     parameters, functionData)
                if valueFromPolicy is not None:
                    policyValue = valueFromPolicy
                    DeviceFunction.putInProcessValue(
                        deviceAnalog.getEndpointId(), deviceModel.getURN(),
                        attribute, policyValue)
                else:
                    self.getLogger().warning(f'applyAttributePolicy(): '
                        f'{attribute} got None value from policy '
                        f'{deviceFunction.getDetails(parameters)}')
                    break
            else:
                # apply returned value
                policyValue = None
                break

            index += 1

        # After the policy loop, if the policyValue is null, then the policy
        # either filtered out the attribute, or the policy parameters have not
        # been met (e.g., sampleQuality rate is not met).
        # If it is not null, then create a new DataItem to replace the old
        # in the data message.
        policyDataItem = None
        if policyValue is not None:
            deviceModelAttribute = None
            if attribute in deviceModel.getDeviceModelAttributes():
                deviceModelAttribute = \
                    deviceModel.getDeviceModelAttributes()[attribute]
            # TODO: can this happen ?
            if deviceModelAttribute is None:
                self.getLogger().warning(f'appyAttributePolicy(): '
                    f'attribute {attribute} not in device model.')
            else:
                deviceAnalog.setAttributeValue(attribute, policyValue)
                policyDataItem = self.createDataItem(dataItem.getType(),
                                                     dataItem.getKey(),
                                                     policyValue)
        else:
            policyDataItem = None

        DeviceFunction.removeInProcessValue(deviceAnalog.getEndpointId(),
                                            deviceModel.getURN(), attribute)

        return policyDataItem

    def checkComputedMetrics(self, dataItems, deviceAnalog, triggerMap,
                             currentTimeMillis):
        """

        :param dataItems:
        :param deviceAnalog:
        :param triggerMap:
        :param currentTimeMillis:
        :return:
        """
        if len(triggerMap) == 0 or len(dataItems) == 0:
            return

        updatedAttributes = set()
        for dataItem in dataItems:
            updatedAttributes.add(dataItem.getKey())
        endpointId = deviceAnalog.getEndpointId()
        deviceModel = deviceAnalog.getDeviceModel()
        deviceModelAttributes = deviceModel.getDeviceModelAttributes()
        deviceModelUrn = deviceModel.getURN()

        # {<string>:<set>}
        # This is the set of attributes that the formula refers to.
        for attribute,key in triggerMap.items():
            # If the set of attributes that the formula refers to
            # is a subset of the updated attributes, then compute
            # the value of the computedMetric.
            if not updatedAttributes.issubset(key):
                continue
            deviceModelAttribute = None
            if attribute in deviceModelAttribute:
                deviceModelAttribute = deviceModelAttributes[attribute]

            attributeValue = deviceAnalog.getAttributeValue(attribute)
            if attributeValue is None:
                attributeValue = deviceModelAttribute.getDefaultValue()
            dataItem = None
            if deviceModelAttribute.getType() == DeviceModelAttribute.Type.NUMBER or \
                    deviceModelAttribute.getType() == DeviceModelAttribute.Type.INTEGER:
                number = float(attributeValue) \
                    if attributeValue is not None else 0.0
                dataItem = DataItem(attribute, number)
            elif deviceModelAttribute.getType() == DeviceModelAttribute.Type.STRING or \
                    deviceModelAttribute.getType() == DeviceModelAttribute.Type.URI:
                string = str(attributeValue) \
                    if attributeValue is not None else ""
                dataItem = DataItem(attribute, string)
            elif deviceModelAttribute.getType() == DeviceModelAttribute.Type.BOOLEAN:
                boolvalue = bool(attributeValue) \
                    if attributeValue is not None else False
                dataItem = DataItem(attribute, boolvalue)
            elif deviceModelAttribute.getType() == DeviceModelAttribute.Type.DATETIME:
                value = 0
                if isinstance(attributeValue, (datetime.datetime, )):
                    value = int(attributeValue.timestamp() * 1000.0)
                else:
                    value = int(attributeValue) \
                        if attributeValue is not None else 0
                dataItem = DataItem(attribute, value)
            else:
                self.getLogger().warning(f'checkComputedMetric(): '
                    f'unknown device model attribute type: {deviceModelAttribute.getType()}')
                continue

            devicePolicyManager = DevicePolicyManager.getDevicePolicyManager(
                self.directlyConnectedDevice)
            devicePolicy = devicePolicyManager.getPolicy(deviceModelUrn,
                                                         endpointId)
            if devicePolicy is None:
                continue

            pipeline = devicePolicy.getPipeline(attribute)
            if pipeline is None or len(pipeline) == 0:
                continue

            policyDataItem = self.applyAttributePolicy(deviceAnalog, dataItem,
                                                       pipeline,
                                                       currentTimeMillis)
            if policyDataItem is not None:
                dataItems.append(policyDataItem)

    def applyDevicePolicies(self, message, currentTimeMillis):
        """
        Return a `list` of messages by applying device policies to `message`.

        Apply policies that are targeted to a device model.

        :param message:
        :param currentTimeMillis:
        :return: a `list` of messages by applying device policies to `message`
        """

        # A data message or alert format cannot be null or empty
        # (enforced in Data/AlertMessage(Builder) constructor)
        format = None
        deviceModelUrn = None
        endpointId = message.getSource()
        if isinstance(message, (DataMessage, )):
            format = message.getFormat()
            deviceModelUrn = format[0: len(":attributes") - len(format)]
        elif isinstance(message, (AlertMessage, )):
            format = message.getFormat()
            # Must contain ":" or get ValueError
            deviceModelUrn = format[0: format.rindex(':')]
        else:
            return [message]

        deviceAnalog = None
        with self.damaplock:
            if endpointId in self.deviceAnalogMap:
                deviceAnalog = self.deviceAnalogMap[endpointId]
        if deviceAnalog is None:
            deviceModel = \
                self.directlyConnectedDevice.getDeviceModel(deviceModelUrn)
            if isinstance(deviceModel, (DeviceModelImpl, )):
                deviceAnalog = DeviceAnalogImpl(self.directlyConnectedDevice,
                                                deviceModel, endpointId)
                with self.damaplock:
                    self.deviceAnalogMap[endpointId] =  deviceAnalog
            # TODO: what to do if deviceAnalog is null?
            if deviceAnalog is None:
                return [message]

        devicePolicyManager = DevicePolicyManager.getDevicePolicyManager(
            self.directlyConnectedDevice)
        devicePolicy = devicePolicyManager.getPolicy(deviceModelUrn,
                                                     endpointId)
        if devicePolicy is None:
            return [message]

        pipeline = devicePolicy.getPipeline(DevicePolicy.ALL_ATTRIBUTES())
        # No policies for this device model, retain the data item.
        if pipeline is None or len(pipeline) == 0:
            return [message]

        logmessage = ""
        if len(pipeline) > 1:
            index = 0
            maxIndex = len(pipeline)
            while index < maxIndex:
                plfunction = pipeline[index]
                id = plfunction.getId()
                parameters = plfunction.getParameters()
                deviceFunction = DeviceFunction.getDeviceFunction(id)
                if index == 0:
                    logmessage = f'applyDevicePolicies(): '\
                        f'Only one function allowed for all-attribute pipeline.'\
                        f'\n\tApplying: {deviceFunction.getDetails(parameters)}'
                else:
                    logmessage = f'{logmessage}'\
                        f'\n\tIgnoring: {deviceFunction.getDetails(parameters)}'
                index += 1

        self.getLogger().warning(logmessage)

        # Create the pipeline data for this device model
        pipelineData = None
        with self.pldclock:
            pipelineData = self.pipelineDataCache[DevicePolicy.ALL_ATTRIBUTES]
        if pipelineData is None:
            pipelineData = []
            # Use the ALL_ATTRIBUTES token, not None, as in JCL
            with self.pldclock:
                self.pipelineDataCache[DevicePolicy.ALL_ATTRIBUTES] = \
                    pipelineData

        # Handle pipleline for device policy
        # TODO: handle more than one function in a pipeline for all-attributes
        # Note: A warning is printed out, several lines above this, if there
        # is more than one function for an all-attributes policy.
        # Leaving original line here, commented-out, for future reference.
        # for (int index = 0,
        #      maxIndex = pipeline.size(); index < maxIndex; index++)
        index = 0
        maxIndex = 1
        while index < maxIndex:
            plfunction = pipeline[index]
            functionData = None
            if index < len(pipelineData):
                functionData = pipelineData[index]
            else:
                functionData = dict()
                pipelineData.append(functionData)
            key = plfunction.getId()
            parameters = plfunction.getParameters()
            deviceFunction = DeviceFunction.getDeviceFunction(key)
            if deviceFunction is None:
                index += 1
                continue 
            windowExpired = False
            window = DeviceFunction.getWindow(parameters)
            slide = DeviceFunction.getSlide(parameters, window)
            if window > 0:
                k = f'{deviceModelUrn}::{deviceFunction.getId()}.expiry'
                expiry = None
                if k in self.windowMap:
                    expiry = self.windowMap.get(k)
                if expiry is None:
                    expiry = currentTimeMillis + window
                    self.windowMap[k] = expiry
                windowExpired = expiry <= currentTimeMillis
                if windowExpired:
                    self.windowMap[k] = expiry + slide
            else:
                windowExpired = False

            alertOverridesPolicy = False
            if isinstance(message, (AlertMessage, )):
                alertMessage = message
                alertMessageSeverity = alertMessage.getSeverity()
                configuredSeverity = AlertMessage.Severity.CRITICAL
                criterion = None
                if "alertSeverity" in parameters:
                    criterion = str(parameters["alertSeverity"])
                if criterion is not None:
                    try:
                        criterion = criterion.upper()
                        configuredSeverity = AlertMessage.Severity[criterion]
                    except Exception as e:
                        configuredSeverity = AlertMessage.Severity.CRITICAL
                alertOverridesPolicy = \
                    configuredSeverity.value() <= alertMessageSeverity.value()
            else:
                alertOverridesPolicy = False

            # VirtualDevice stores Pair<Message,StorageObjectImpl>,
            # do the same here.
            pair = Pair(message, None)
            if deviceFunction.apply(deviceAnalog, None, parameters,
                                    functionData, pair) or \
                    windowExpired or \
                    alertOverridesPolicy:
                valueFromPolicy = deviceFunction.get(deviceAnalog, None,
                                                     parameters, functionData)
                if valueFromPolicy is not None:
                    messageList = valueFromPolicy
                    messages = []
                    # element is a Pair
                    for element in messageList:
                        # For LL, we don't care about StorageObject since it
                        # isn't queued with message.
                        # For HL, we do care because MessageDispatcherImpl
                        # logically ties the two together.
                        # But this is LL code here, so we only want the
                        # message. Besides, when this Pair is created (just
                        # before calling apply), the value side of the Pair
                        # is passed as null.
                        messages.append(element.getKey())
                    return messages

            index += 1

        return []

    def expirePolicy(self, devicePolicy, currentTimeMillis):
        """
        Return a `list` of messages from expired pipelines in `devicePolicy`.

        :param devicePolicy:
        :param currentTimeMillis:
        :return: a `list` of messages from expired pipelines in `devicePolicy`.
        """

        # consolidate messages
        consolidatedMessageList = []
        messageList = self._expirePolicy_0(devicePolicy)
        if len(messageList) == 0:
            return consolidatedMessageList

        dataItemMap = dict()
        for message in messageList:
            if isinstance(message, (DataMessage, )):
                endpointId = message.getSource()
                dataItems = None
                if endpointId in dataItemMap:
                    dataItems = dataItemMap[endpointId]
                if dataItems is None:
                    dataItems = []
                    dataItemMap[endpointId] = dataItems
                dataItems.extend(message.getDataItems())
            else:
                consolidatedMessageList.append(message)

        for endpointId,dataItems in dataItemMap.items():
            deviceAnalog = None
            with self.damaplock:
                if endpointId in self.deviceAnalogMap:
                    deviceAnalog = self.deviceAnalogMap[endpointId]
            if deviceAnalog is None:
                continue
            format = deviceAnalog.getDeviceModel().getURN()
            # TODO: Need a lock for triggers ?
            if len(self.computedMetricTriggers) > 0:
                triggerMap = None
                if format in self.computedMetricTriggers:
                    triggerMap = self.computedMetricTriggers[format]
                if triggerMap is not None and len(triggerMap) > 0:
                    try:
                        self.checkComputedMetrics(dataItems, deviceAnalog,
                                                  triggerMap,
                                                  currentTimeMillis)
                    except Exception as e:
                        self.getLogger().warning(f'{e}')
            dataMessage = DataMessage(source=deviceAnalog.getEndpointId(),
                            format=deviceAnalog.getDeviceModel().getURN(),
                            items=dataItems)
            consolidatedMessageList.append(dataMessage)


    #@expirePolicy.register(object, DevicePolicy)
    def _expirePolicy_0(self, devicePolicy):
        """
        Return a `list` of messages from expired pipelines in `devicePolicy`.

        :param devicePolicy:
        :return: a `list` of messages from expired pipelines in `devicePolicy`
        """
        messageList = []

        deviceAnalogs = None
        with self.damaplock:
            if len(self.deviceAnalogMap) == 0:
                return messageList
            deviceAnalogs = self.deviceAnalogMap.values()

        for deviceAnalog in deviceAnalogs:
            messages = self._expirePolicy_1(devicePolicy, deviceAnalog)
            if messages is not None and len(messages) > 0:
                messageList.extend(messages)
        return messageList

    #@expirePolicy.register(object, DevicePolicy, DeviceAnalog)
    def _expirePolicy_1(self, devicePolicy, deviceAnalog):
        """
        Return a `list` of messages from expired pipelines in `devicePolicy`.

        :param devicePolicy:
        :param deviceAnalog:
        :return: Return a `list` of messages from expired pipelines in
            `devicePolicy`.
        """
        messageList = []
        pipelines = devicePolicy.getPipelines()
        if pipelines is None:
            return messageList

        for attributeName,pipeline in pipelines.items():
            messages = self._expirePolicy_2(attributeName, pipeline,
                                           deviceAnalog)
            if messages is not None:
                messageList.extend(messages)
        return messageList

    #@expirePolicy.register(object, str, List, DeviceAnalog)
    def _expirePolicy_2(self, attributeName, pipeline, deviceAnalog):
        """
        Return a `list` of messages resulting from the expired `pipeline`.

        :param attributeName:
        :param pipeline:
        :param deviceAnalog:
        :return: a `list` of messages from expired `pipeline` functions.
        """
        if pipeline is None or len(pipeline) == 0:
            return None

        # TODO: Note that pipelineData is not a copy here, and even though
        # TODO: there was a lock when it was obtained, if changed elsewhere
        # TODO: it will affect the results of this method.
        # TODO: Can this ever happen ?
        pipelineData = None
        with self.pldclock:
            pipelineData = self.pipelineDataCache.pop(attributeName)
        if pipelineData is None:
            return None

        index = 0
        maxIndex = len(pipeline)
        while index < maxIndex:
            plfunction = pipeline[index]
            if plfunction is None:
                index += 1
                continue 
            deviceFunction = \
                DeviceFunction.getDeviceFunction(plfunction.getId())
            if deviceFunction is None:
                return None

            # Looking for the first policy function in the pipeline that has a
            # "window".
            # If there isn't one, we'll drop out of the loop and return null.
            # If there is one, we process the remaining pipeline from there.
            window = DeviceFunction.getWindow(plfunction.getParameters())
            if window == -1:
                index += 1
                continue 
            functionData = pipelineData[index] \
                if index < len(pipelineData) else None
            if functionData is None:
                # there is no data for this function, so return
                return None
            valueFromPolicy = \
                deviceFunction.get(deviceAnalog, attributeName,
                                   plfunction.getParameters(), functionData)
            if valueFromPolicy is None:
                return None

            next = index + 1
            while next < maxIndex:
                plfunction = pipeline[next]
                if plfunction is None:
                    return None
                deviceFunction = \
                    DeviceFunction.getDeviceFunction(plfunction.getId())
                if deviceFunction is None:
                    return None
                functionData = pipelineData[next] \
                    if next < len(pipelineData) else None
                if functionData is None:
                    return None

                if deviceFunction.apply(deviceAnalog, attributeName,
                                        plfunction.getParameters(),
                                        functionData, valueFromPolicy):
                    valueFromPolicy = \
                        deviceFunction.get(deviceAnalog, attributeName,
                                           plfunction.getParameters(),
                                           functionData)
                    if valueFromPolicy is None:
                        return None
                else:
                    return None

                next += 1

            # If we get here, valueFromPolicy is not null.
            if isinstance(valueFromPolicy, (list, )):
                return valueFromPolicy

            deviceModel = deviceAnalog.getDeviceModel()
            dataMessage = DataMessage(
                source=deviceAnalog.getEndpointId(),
                format=deviceModel.getURN())
            deviceModelAttribute = None
            if attributeName in deviceModel.getDeviceModelAttributes():
                deviceModelAttribute = \
                    deviceModel.getDeviceModelAttributes()[attributeName]
            # TODO: Can this happen ?
            if deviceModelAttribute is None:
                continue
            if deviceModelAttribute.getType() == DeviceModelAttribute.Type.DATETIME or \
                    deviceModelAttribute.getType() == DeviceModelAttribute.Type.NUMBER or \
                    deviceModelAttribute.getType() == DeviceModelAttribute.Type.INTEGER:
                dataMessage.addItem(attributeName, float(valueFromPolicy))
            elif deviceModelAttribute.getType() == DeviceModelAttribute.Type.BOOLEAN:
                dataMessage.addItem(attributeName, float(bool(valueFromPolicy)))
            elif deviceModelAttribute.getType() == DeviceModelAttribute.Type.STRING or \
                    deviceModelAttribute.getType() == DeviceModelAttribute.Type.URI:
                dataMessage.addItem(attributeName, str(valueFromPolicy))

            messages = []
            messages.append(dataMessage)
            return messages

        return None

    def createDataItem(self, itemtype, key, newValue):
        """
        Return a :class:`.DataItem` instance.
        :param self:
        :param itemtype:
        :param key:
        :param newValue:
        :return: a :class:`.DataItem` instance.
        """
        dataItem = None
        if itemtype == DataItem.Type.DOUBLE:
            dataItem = DataItem(key, float(newValue))
        elif itemtype == DataItem.Type.STRING:
            dataItem = DataItem(key, str(newValue))
        elif itemtype == DataItem.Type.BOOLEAN:
            dataItem = DataItem(key, bool(newValue))
        return dataItem

    @classmethod
    def getLogger(cls):
        """Return a logger for  :class:`.MessagingPolicy`."""
        return logging.getLogger('iotcs.messaging.client.device.MessagingPolicy')

# com.oracle.iot.client.impl.device

class ScheduledPolicyData():
    """A class to encapsulate scheduled policy data."""
    
    class Key():
        """A class to represent a key that represents ScheduledPolicyData."""

        __slots__ = (
            'window',
            'slide'
        )

        def __init__(self, window, slide):
            """

            :param window:
            :param slide:
            """
            self.window = window
            self.slide = slide

        def __eq__(self, other):
            if self is other:
                return True
            if self.__class__ != other.__class__:
                return False

            return self.window == self.window and self.slide == self.slide

        def __hash__(self):
            """Return a *hash* for this instance."""
            return hash((self.window, self.slide))

    __slots__ = (
        'expiry',
        #  Millisecond time in the future at which the policy value should be
        # computed
        'slide',
        #  How much the window moves is used to calculate expiry.
        'pipelineIndices',
        #  { attributeName : pipelineIndex }
        'pipelineIndicesLock'
    )

    def __init__(self, window, slide, timeZero):
        """
        Return an instance of `ScheduledPolicyData`
        :param window:
        :param slide:
        :param timeZero:
        """
        self.slide = slide
        self.pipelineIndices = dict()
        #  Initial expiry is window milliseconds past time zero.
        #  Tenth of a millisecond resolution helps group
        #  intersecting slide values (10 second and 20 second,
        #  for example).
        self.expiry = ((window + timeZero) / 10) * 10

        self.pipelineIndicesLock = threading.RLock()

    def addAttribute(self, attributeName, pipelineIndex):
        """
        Add an entry to `pipelineIndices`.

        :param attributeName:
        :param pipelineIndex:
        """
        with self.pipelineIndicesLock:
            self.pipelineIndices[attributeName] = pipelineIndex

    def removeAttribute(self, attributeName, pipelineIndex):
        """ generated source for method removeAttribute """
        with self.pipelineIndicesLock:
            if attributeName in self.pipelineIndices:
                del(self.pipelineIndices[attributeName])

    def isEmpty(self):
        """
        Return `True` if `pipelineIndices` is empty.
        :return:
        """
        with self.pipelineIndicesLock:
            return len(self.pipelineIndices) == 0

    def getDelay(self, now):
        """
        Return the offset (delay) between :attr:`.expiry` and `now`.
        :param now:
        :return: the offset (delay) between :attr:`.expiry` and `now` in
            fractional seconds.
        """
        delay = float((self.expiry - now) / 1000.0)
        return delay if delay > 0 else 0

    def processExpiredFunction(self, virtualDeviceImpl, updatedAttributes,
                               timeZero):
        """
        Perform maintenance for the expired function.
        :param virtualDeviceImpl:
        :param updatedAttributes:
        :param timeZero:
        """
        try:
            self.handleExpiredFunction(virtualDeviceImpl, updatedAttributes)
        finally:
            # ensure expiry is reset
            # tenth of a millisecond resolution
            self.expiry = ((self.slide + timeZero) / 10) * 10

    def handleExpiredFunction(self, virtualDeviceImpl, updatedAttributes):
        """

        :param virtualDeviceImpl:
        :param updatedAttributes:
        """
        devicePolicy = virtualDeviceImpl._VirtualDeviceImpl__devicePolicyManager.getPolicy(
            virtualDeviceImpl.getDeviceModel().getURN(),
            virtualDeviceImpl.getEndpointId())
        if devicePolicy is None:
            # SEVERE
            self.getLogger().critical(f'handleExpiredFunction(): '
                f'could not find {virtualDeviceImpl.getDeviceModel().getURN()} '
                f'in policy configuration')
            return

        pipelineIndicesCopy = None
        with self.pipelineIndicesLock:
            pipelineIndicesCopy = self.pipelineIndices.copy()
        self._handleExpiredFunction_0(virtualDeviceImpl, updatedAttributes,
                                   devicePolicy, pipelineIndicesCopy)


    def _handleExpiredFunction_0(self, virtualDeviceImpl, updatedAttributes,
                                devicePolicy, indices):
        """

        :param virtualDeviceImpl:
        :param updatedAttributes:
        :param devicePolicy:
        :param indices:
        """
        for attributeName,pipelineIndex in indices.items():
            self._handleExpiredFunction_1(virtualDeviceImpl, updatedAttributes,
                                       devicePolicy, attributeName,
                                       pipelineIndex)

    def _handleExpiredFunction_1(self, virtualDeviceImpl, updatedAttributes,
                                devicePolicy, attributeName, pipelineIndex):
        """

        :param virtualDeviceImpl:
        :param updatedAttributes:
        :param devicePolicy:
        :param attributeName:
        :param pipelineIndex:
        """
        pipeline = devicePolicy.getPipeline(attributeName)
        if pipeline is None or len(pipeline) == 0:
            return

        if len(pipeline) <= pipelineIndex:
            # SEVERE
            self.getLogger().critical(f'_handleExpiredPolicy_1(): '
                f'pipeline does not match configuration')
            return

        pipelineData = virtualDeviceImpl.getPipelineData(attributeName)
        if len(pipelineData) <= pipelineIndex:
            self.getLogger().critical(f'_handleExpiredPolicy_1(): '
                f'pipeline data does not match configuration')
            return

        remainingPipelineConfigs = pipeline[pipelineIndex: len(pipeline)]
        remainingPipelineData = pipelineData[pipelineIndex: len(pipelineData)]
        if not DevicePolicy.ALL_ATTRIBUTES == attributeName:
            virtualDeviceImpl.processExpiredFunction(updatedAttributes,
                                                     attributeName,
                                                     remainingPipelineConfigs,
                                                     remainingPipelineData)
        else:
            virtualDeviceImpl._processExpiredFunction(remainingPipelineConfigs,
                                                     remainingPipelineData)

    def __hash__(self):
        """Return a hash."""
        return hash((self.slide, self.pipelineIndicesLock))

    def __eq__(self, obj):
        """ generated source for method equals """
        if self == obj:
            return True
        if obj == None or obj.__class__ != self.__class__:
            return False

        return (self.slide == obj.slide)

    def getLogger(self):
        """Return a logger for :class:`.ScheduledPolicyData`."""
        return logging.getLogger(
            "iotcs.messaging.client.device.ScheduledPolicyData")

#  TODO: can we have only one of these threads for all virtual devices?
class TimedPolicyThread(threading.Thread):
    """A thread to run timed policies."""

    __slots__ = (
        'scheduledPolicyData',
        'scheduledPolicyDataLock',
        'cancelled',
        'virtualDevice'
    )

    THREADCOUNTLOCK = threading.RLock()
    THREADCOUNT = 0

    def __init__(self, virtualDevice):
        """Return an instance of :class:`.TimedPolicyThread."""
        super(TimedPolicyThread, self).__init__(
            name=f'iotcs.messaging.client.device.timed-policy-thread'
                f'{TimedPolicyThread.THREADCOUNT}')
        with TimedPolicyThread.THREADCOUNTLOCK:
            TimedPolicyThread.THREADCOUNT += 1
        self.setDaemon(True)
        self.scheduledPolicyDataLock = threading.Condition()
        self.scheduledPolicyData = []
        self.cancelled = False
        self.virtualDevice = virtualDevice

    def addTimedPolicyData(self, data):
        """
        Add `data` to :attr:`.scheduledPolicyData`.
        :param data: a :class:`.ScheduledPolicyData` instance.
        """
        with self.scheduledPolicyDataLock:
            index = -1
            try:
                index = self.scheduledPolicyData.index(data)
            except Exception as e:
                # data is not in the list
                pass
            if index == -1:
                self.scheduledPolicyData.append(data)
            else:
                self.scheduledPolicyData[index] = data
            self.scheduledPolicyDataLock.notify_all()

    def removeTimedPolicyData(self, data):
        """
        Remove `data` from :attr:`.scheduledPolicyData`.
        :param data:  a :class:`.ScheduledPolicyData` instance.
        :return:
        """
        with self.scheduledPolicyDataLock:
            self.scheduledPolicyData.remove(data)
            self.scheduledPolicyDataLock.notify_all()

    def isCancelled(self):
        """Return `True` if the thread is cancelled."""
        return self.cancelled

    #  TODO: never used. Do we need cancelled and cancel()?
    def cancel(self):
        """ generated source for method cancel """
        self.cancelled = True

    def run(self):
        """ generated source for method run """
        while not self.cancelled:
            delay = -1
            with self.scheduledPolicyDataLock:
                #  Wait for a timed function to be added
                while not self.cancelled and \
                        len(self.scheduledPolicyData) == 0:
                    try:
                        # TODO: wait with timeout
                        self.scheduledPolicyDataLock.wait()
                    except Exception as e:
                        #  Restore interrupt status
                        break

                now = TimeManager.currentTimeMillis()
                def keyfunc(element):
                    return element.getDelay(now)
                self.scheduledPolicyData.sort(key=keyfunc)
                delay = self.scheduledPolicyData[0].getDelay(now)
                #  May have been cancelled while waiting for data to be
                #  added to scheduledPolicyData
                if not self.cancelled and delay > 0:
                    try:
                        self.scheduledPolicyDataLock.wait(delay)
                    except Exception as e:
                        #  Restore the interrupted status
                        break

                currentTimeMillis = TimeManager.currentTimeMillis()

                # List<Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>>
                updatedAttributes = []
                for policyData in self.scheduledPolicyData:
                    #  Run through all the timed function data
                    if policyData.getDelay(currentTimeMillis) <= 0:
                        policyData.processExpiredFunction(
                            self.virtualDevice, updatedAttributes,
                            currentTimeMillis)
                if len(updatedAttributes) > 0:
                    # 
                    #  Call updateFields to ensure the computed metrics get run,
                    #  and will put all attributes into one data message.
                    # 
                    self.virtualDevice.updateFields(updatedAttributes)

    def getLogger(self):
        """Return a logger for :class:`.TimedPolicyThread`."""
        return logging.getLogger("iotcs.messaging.client.device.TimedPolicyThread")

