# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod
import logging
import threading


from iotcs.common import Singleton
from iotcs.common import StatusCode
from iotcs.message import ResponseMessage

# com.oracle.iot.client.device.util

class RequestHandler(ABC):
    """A `RequestHandler` handles request messages from an IoT server."""

    __slots__ = ()

    @abstractmethod
    def handleRequest(self, requestMessage):
        """
        Handle the :class:`iotcs.shared.message.RequestMessage`.

        After handling the reqest from the IoT server, return a
        :class:`iotcs.shared.message.ResponseMessage`
        :param requestMessage: A device request.
        :return: :class:`iotcs.shared.message.ResponseMessage`
        """
        pass

# TODO: Need better solution than Singleton
# TODO: Make "module" object

@Singleton
class RequestDispatcher:
    """"
    `RequestDispatcher` is a utility that allows the caller to register handlers
    for requests coming from the IoT server. The `RequestDispatcher` handles all
    of the logic of matching the resource to a handler and invoking the handler.

    The `hendler` definition  `handler(requestMessage:RequestMessage) -> ResponseMessage`
    """

    class DefaultRequestHandler(RequestHandler):
        def handleRequest(self, requestMessage):
            return ResponseMessage(requestMessage=requestMessage,statusCode=StatusCode.NOT_FOUND)

    class RequestHandlerRegistration:
        """
        `RequestHandlerRegistration` encapsulates the association of a `handler` to
        an `endpointId` and resource `path`
        """

        __slots__ = (
            'endpointId',
            'path',
            'handler'
        )

        def __init__(self, endpointId, path, handler):
            self.endpointId = "*" if (endpointId == None) or (len(endpointId) == 0) else endpointId
            self.path = "*" if (path == None) or (len(path) == 0) else path
            self.handler = handler


        # Note that other code depends on RequestHandlerRegistrations being
        #sorted such that longer contexts come first.
        def compareTo(self,other):

            if (self.endpointId < other.endpointId):
                return -1

            if (self.endpointId > other.endpointId):
                return 1

            #same endpoint
            diff = len(other.path) - len(self.path)
            if (diff != 0):
                return diff

            if (self.path < other.path):
                return -1

            if (self.path > other.path):
                return 1

            # both the same
            return 0

        def __eq__(self,other):

            # equals and compareTo have to agree.
            if (other == None or not isinstance(other,self.__class__)):
                return False

            if (not self.endpointId.__eq__(other.endpointId)):
                return False

            return self.path.__eq__(other.path)

        def __lt__(self, other):
            return self.compareTo(other) < 0

        def __gt__(self, other):
            return self.compareTo(other) > 0

        def __str__(self):
            return "RequestHandlerRegistration{" + self.endpointId + ", " + self.path + "}"

    __slots__ = (
        '__registrationsList',
        '__registrationsListLock',
        '__defaultHandler',
        '__DEFAULT_HANDLER'
    )

    def __init__(self):
        self.__registrationsList = []
        # TODO: Should this be reentrant RLock ?
        self.__registrationsListLock = threading.Lock()
        self.__DEFAULT_HANDLER = self.DefaultRequestHandler()
        self.__defaultHandler = self.__DEFAULT_HANDLER


    def registerRequestHandler(self, endpointId, path, handler):
        """
        Register a handler for the given endpoint and path. If endpoint is null,
        empty, or a wildcard ("*"), then the handler is registered as being used
        for all endpoints. If path is null, empty, or a wildcard ("*")
        then the handler is registered as the handler for any requests that do
        not have a more specific handler.
        The {@code handler} parameter may not be null.

        Registering a handler with
            `registerRequestHandler("*", "*", handler)`
        will register a default, catch-all handler. Note that this utility class
        already provides a reasonable default handler that can be replaced with this
        technique. The implementation ensures there is always a default handler.

        endpointId the endpoint id that has the given the resource path,
                   or a wildcard ("*") for any endpoint.
        path       the resource path, or a wildcard ("*") for any resource
        handler    the handler to invoke for requests to this endpoint and path
        throws IllegalArgumentException if handler is null
        """


        if handler is None:
            raise ValueError(__name__ + ": handler may not be null")

        normalizedEndpointId = "*" if (endpointId == None) or (len(endpointId) == 0) else endpointId
        normalizedPath = "*" if (path == None) or (len(path) == 0) else path

        # synchronized (requestHandlerRegistrations) {
        with self.__registrationsListLock:

            # special case for default handler
            if "*" == normalizedEndpointId and "*" == normalizedPath:
                self.__defaultHandler = handler
                return

            # RequestHandlerRegistration is immutable, so make a new one
            registration = self.RequestHandlerRegistration(normalizedEndpointId, normalizedPath, handler)

            found = False
            for n in range(len(self.__registrationsList)):

                other = self.__registrationsList[n]
                # assignment in the conditional is intentional
                found = registration.__eq__(other)
                if found:
                    # replace the existing registration
                    self.__registrationsList[n] = registration
                    break

            if not found:
                self.__registrationsList.append(registration)
                self.__registrationsList.sort()

    def unregisterRequestHandler(self, handler, endpointId='-', path='-'):
        """
        Un-register the handler for the given `endpoint` and `path`. Either or both
        of the parameters may be a wildcard ("*"). `None` or empty parameters are
        interpreted as a wildcard.
        Note that the wildcard is not used to match any endpoint or path, but
        is used to match a handler that was registered with the wildcard endpoint
        and/or path. Thus, a `handler` registered by calling
        `registerRequestHandler("0-AB", "resource", handler)`
        will not be un-registered by calling
        `unregisterRequestHandler("*", "resource")`.
        Nor will calling `unregisterRequestHandler("0-AB", "*")` result
        in un-registering all handlers for endpoint "0-AB". In short, the
        arguments to `registerRequestHandler` should match the
        arguments to `unregisterRequestHandler`.

        A default handler can be registered by calling
        `registerRequestHandler("*", "*", handler)`. Calling
        `unregisterRequestHandler("*", "*")` will restore the
        default handler. The implementation ensures there is always a
        default handler.

        If `endpointId` and `path` are not specified as arguments then the `handler`
        is removed from all registrations that is is defined in.

        endpointId the endpoint id that has the given the resource path,
                           or a wildcard ("*").
        path       the resource path, or a wildcard ("*")
        handler the requestHandler to register
        """

        # special case to remove all registrations of this handler
        # was unregisterRequestHandler(self, handler)
        if endpointId == '-' and path == '-':
            with self.__registrationsListLock:
                # special case for default handler
                if handler == self.__defaultHandler and self.__defaultHandler != self.__DEFAULT_HANDLER:
                    self.__defaultHandler = self.__DEFAULT_HANDLER
                    return

                for n in range(len(self.__registrationsList)):
                    registration = self.__registrationsList[n]
                    if registration.handler == handler:
                        self.__registrationsList.pop(n)
                return None

        if endpointId == '-':
            endpointId = None
        if path == '-':
            endpointId = None

        normalizedEndpointId = "*" if (endpointId == None) or (len(endpointId) == 0) else endpointId
        normalizedPath = "*" if (path == None) or (len(path) == 0) else path

        with self.__registrationsListLock:
            # special case for default handler
            if "*" == normalizedEndpointId and "*" == normalizedPath:
                self.__defaultHandler = self.__DEFAULT_HANDLER
                return

            # RequestHandlerRegistration is immutable, so make a new one
            registrationToUnregister = self.RequestHandlerRegistration(normalizedEndpointId, normalizedPath, handler)

            found = False
            for n in range(len(self.__registrationsList)):

                registration = self.__registrationsList[n]
                # assignment in the conditional is intentional
                found = registration.__eq__(registrationToUnregister)
                if found:
                    self.__registrationsList.pop(n)
                    return

    def dispatch(self, requestMessage):
        """
        Match the request to a registered handler and invoke the handler.

        requestMessage The request message to be dispatched
        return The `ResponseMessage` of the invoked handler. If the invoked
        handler throws an exception or returns None, a reasonable
        `ResponseMessage` will be returned by this method. This method will
        not return None.

        raises `ValueError` if requestMessage is null
        """

        if requestMessage is None:
            raise ValueError(__name__ + ": requestMessage may not be None")

        endpointId = requestMessage.destination
        path = requestMessage.getURL()
        requestHandler = self.getRequestHandler(endpointId, path)

        try:
            responseMessage = requestHandler.handleRequest(requestMessage)
            if responseMessage is None:
                if logging.getLogger("iotcs.device.impl.RequestDispatcher").isEnabledFor(logging.CRITICAL):
                    logging.getLogger().critical(__name__ + ":" + requestHandler.str() + " returned None")

                responseMessage = self.copyResponseMessage(ResponseMessage(requestMessage))
                responseMessage.statusCode = StatusCode.INTERNAL_SERVER_ERROR

            return responseMessage
        except Exception as exp:
            if logging.getLogger().isEnabledFor(logging.CRITICAL):
                logging.getLogger().critical(str(requestHandler) + " raised " + str(exp))

            responseMessage = ResponseMessage(requestMessage, StatusCode.INTERNAL_SERVER_ERROR)
            #responseMessage.source = requestMessage.destination
            #responseMessage.requestId = requestMessage.id
            responseMessage.body = str(exp)
            return responseMessage

    def copyResponseMessage(self, message):
        responseMessage = ResponseMessage()
        if message is ResponseMessage:
            responseMsg = message
            responseMessage = self.copyMessage(responseMsg)
            oldHeader = responseMsg.getHeaders()
            for (h,v) in oldHeader:
                responseMessage.header(h,v)

            #responseMessage =
            #   responseMessage.body(Arrays.copyOf(responseMsg.getBody(),
            #                                   len(responseMsg.getBody())))
            responseMessage.statusCode = responseMsg.getStatusCode()
            responseMessage.url = responseMsg.getURL()
            responseMessage.requestId = responseMsg.getRequestId()
        else:
            raise ValueError(__name__ + ": not copy a different type of message")

        return responseMessage

    def copyMessage(self, responseMessage, message):
        responseMessage.id = message.getId()
        responseMessage.clientId = message.getClientId()
        responseMessage.source = message.getSource()
        responseMessage.destination = message.getDestination()
        responseMessage.priority = message.getPriority()
        responseMessage.properties = message.getProperties()
        responseMessage.eventTime = message.getEventTime().longValue()
        responseMessage.reliability = message.getReliability()
        responseMessage.sender = message.getSender()

        responseMessage = self.copyMap(responseMessage, message.getDiagnostics())
        responseMessage.direction = message.getDirection()
        responseMessage.receivedTime = message.getReceivedTime()
        responseMessage.sentTime = message.getSentTime()
        return responseMessage

    def copyMap(self, responseMessage, map):

        if map is None:
            return responseMessage

        responseMessageCopy = None
        for (k,v) in map:
            responseMessageCopy = responseMessage.diagnostic(k, v)

        return responseMessageCopy

    def getRequestHandler(self, endpointId, path):
        """
        Lookup a RequestHandler for the given endpointId and path. The lookup
        will return the most specific handler for the endpointId and path. If
        there is not a specific match for endpointId, the lookup will match
        try to match '("*", path)'. Failing that, the lookup will return
        a handler for '("*", "*")', or the default handler if no handler
        for '("*","*")' has been registered.
        Will not return null.

        endpointId the endpoint id that has the given the resource path,
                  or a wildcard ("*").
        path the resource path, or a wildcard ("*")
        return The RequestHandler for the given endpointId and path.
        """
        normalizedEndpointId = "*" if endpointId is None or (0 == len(endpointId)) else endpointId
        normalizedPath = "*" if (path is None) or (len(path) == 0) else  path

        with self.__registrationsListLock:

            # special case for default handler
            if "*" == normalizedEndpointId and "*" == normalizedPath:
                return self.__defaultHandler

            #
            # Note that this loop depends on requestHandlerRegistrations being
            # sorted such that longer paths come first in the list.
            #
            for n in range(len(self.__registrationsList)):

                registration = self.__registrationsList[n]
                otherEndpointId = registration.endpointId

                if normalizedEndpointId == otherEndpointId or "*" == otherEndpointId:
                    otherPath = registration.path
                    if "*" == otherPath:
                        return registration.handler

                    diff = len(otherPath) - len(normalizedPath)
                    if diff == 0:
                        if normalizedPath == otherPath:
                            return registration.handler

                    elif diff < 0:
                        if normalizedPath.startswith(otherPath):
                            return registration.handler
                    #else:
                    #    diff > 0, which means that registration.resource
                    #    is more specific than resource
                    #

        return self.__defaultHandler
