# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import abstractmethod
import collections
import hashlib
import json
import logging
import pathlib
import threading
from urllib.parse import urlparse
import urllib3

from iotcs.exception import ArgumentException
from iotcs.exception import NetworkException
from iotcs.exception import SecurityException
from iotcs.exception import StateException
from iotcs.exception import TransportException
from iotcs.messaging.client import ExternalObject
from iotcs.common import Closeable
from iotcs.common import RestApi
from iotcs.common import Util

# com.oracle.iot.client.impl


class StorageAuthenticationResponse():
    """Encapsulates the cloud storage service the authentication response."""

    __slots__ = (
        'storageContainerUrl',
        # The url of the storage service container.
        'authToken'
        # The storage server authentication token.
    )

    FIELD_STORAGE_CONTAINER_URL = "storageContainerUrl"
    FIELD_AUTH_TOKEN = "authToken"

    def __init__(self):
        """Create a :class:`.StorageAuthenticationResponse` instance."""
        super(StorageAuthenticationResponse, self).__init__()
        self.storageContainerUrl = None
        self.authToken = None

    def getAuthToken(self):
        """
        Return the storage authentication token.

        :return: the storage authentication token
        """
        return self.authToken

    def setAuthToken(self, authToken):
        """Set the storage authentication token."""
        self.authToken = authToken

    def getStorageContainerUrl(self):
        """
        Return the :attr:`.storageContainerUrl`.

        :return: the storage container url
        """
        return self.storageContainerUrl

    def setServicePath(self, servicePath):
        """Set the storage service path."""
        self.storageContainerUrl = servicePath

    def __str__(self):
        """Return a string representation of this instance."""
        s = f'StorageAuthenticationResponse{{'\
            f'storageContainerUrl=\'{self.storageContainerUrl}\','\
            f'authToken={self.authToken}}}'
        return s

    @classmethod
    def fromJson(cls, jsonObject):
        """
        Return a :class:`.StorageAuthenticationResponse` instance.

        :param jsonObject: a json formatted storage authentication response
        :raises: :class:`.NetworkException` if `jsonObject` is not a valid
            storage authentication response
        :return: a :class:`.StorageAuthenticationResponse` instance
        """
        url = None
        token = None
        try:
            url = jsonObject[cls.FIELD_STORAGE_CONTAINER_URL]
            token = jsonObject[cls.FIELD_AUTH_TOKEN]
        except Exception as e:
            raise NetworkException(f'Invalid storage authentication: {e}')

        response = StorageAuthenticationResponse()
        response.setAuthToken(token)
        response.setServicePath(url)
        return response

    def toJson(self):
        """
        Return a `json` object for this instance.

        :return: a `json` object
        """
        jsonObject = dict()

        jsonObject[
            StorageAuthenticationResponse.FIELD_STORAGE_CONTAINER_URL] = \
            self.storageContainerUrl
        jsonObject[StorageAuthenticationResponse.FIELD_AUTH_TOKEN] = \
            self.authToken

        return jsonObject

# com.oracle.iot.client.impl


class StorageConnection(Closeable):
    """
    A class to represent a connection to the Storage Cloud Service.

    The StorageConnection transfers content to the Storage Cloud Service.
    StorageConnection is used by the client internally.
    """

    @abstractmethod
    def sync(self, storageObject):
        """
        Synchronize content with the Storage Cloud Service.

        :param storageObject: The StorageObject that identifies the content.
        :raises OSerror: if there is an I/O exception raised by the runtime,
            or a failure reported by the storage cloud
        :raises: :class:`.SecurityException` if there is an exception
            establishing a secure connection to the storage cloud
        :raises: :class:`.ArgumentException` if the
            :class:`storageObject <.StorageObject>` does not return an
            input stream or an output stream
        """
        pass



# com.oracle.iot.client.impl

class StorageConnectionBase(StorageConnection):
    """
    A class to transfer content to and from the Storage Cloud Service.

    There is one StorageConnection instance per client.
    """

    __slots__ = (
        'secureConnection',
        'authToken',
        'authTokenLock'
        'closed',
        'closedLock'
        'http'
    )

    _AUTH_TOKEN_HEADER = "X-Auth-Token"
    _BAD_CHECKSUM_STATUS = -1
    _TRANSFER_CANCELLED = -2

    _REST_STORAGE_AUTHENTICATION = \
        f'{RestApi.V2.getReqRoot()}/provisioner/storage'
    """REST resource for storage authentication."""
    _RETRY_LIMIT = Util.getConfig().getint("media_storage",
                                          "storage_connection_retry_limit")
    """Number of times to retry a transfer if the token expires."""
    STORAGE_CLOUD_HOST = Util.getConfig().get("media_storage",
                                              "storage_connection_host_name")

    # negate the value of "disable_storage_object_prefix" to get an
    # affirmative sense with "_virtualStorageDirectories" i.e. "allow"
    # directory prefix for storage urls.
    _virtualStorageDirectories = not Util.getConfig().getboolean(
        "media_storage", "disable_storage_object_prefix")

    def __init__(self, secureConnection):
        """
        Create a :class:`.StorageConnectionBase` instance.
        
        :param secureConnection: a :class`.SecureConection` instance
        """
        super(StorageConnectionBase, self).__init__()
        self.secureConnection = secureConnection
        self.closed = False
        self.closedLock = threading.RLock()
        self.authToken = None
        self.authTokenLock = threading.RLock()
        self._authenticate()

        # TODO: encapsulate this somewhere else
        self.http = urllib3.PoolManager()

    def sync(self, storageObject):
        """
        Synchronize content with the Storage Cloud Service.

        :param storageObject: The StorageObject that identifies the content.
        :raises TransportException: if there is an I/O exception raised by the runtime,
            or a failure reported by the storage cloud
        :raises: :class:`.SecurityException` if there is an exception
            establishing a secure connection to the storage cloud
        :raises: :class:`.ArgumentException` if the
            :class:`storageObject <.StorageObject>` does not return an
            input stream or an output stream
        """
        if storageObject.getInputStream() is not None:
            self._upload(storageObject)
        elif storageObject.getOutputStream() is not None:
            self._download(storageObject)
        else:
            raise ArgumentException(
                "InputStream and OutputStream are not set.")

    def _upload(self, storageObject):
        """
        Upload `storageObject` to cloud storage.
        
        :param storageObject: a :class:`.StorageObject` instance.
        :raises: :class:`.TransportException` if upload fails
        """
        responseCode = self._transfer(storageObject)
        if responseCode != 201:
            raise TransportException(responseCode,
                                     f'Upload failed {responseCode}')

    def _download(self, storageObject):
        """
        Download `storageObject` from cloud storage.

        :param storageObject: a :class:`.StorageObject` instance
        :raises: :class:`.TransportException` if download fails
        """
        responseCode = self._transfer(storageObject)
        if responseCode != 200 and responseCode != 206:
            raise TransportException(responseCode,
                                     f'Download failed {responseCode}')

    # TODO: There are 3 createStorageObject methods
    # TODO: public createStorageObject(clientId, name, contentType)
    # TODO:     called from messaging.DirectlyConnectedDevice.createStorageObject(name,contentType)
    # TODO:     called from virtualization.EnterpriseClient.createStorageObject(name,contentType)
    # TODO;     this method calls createStorageObject(storageContainerUrl, name, 
    # TODO:                              contentType, type_, encoding, date, length)
    # TODO: public createStorageObject(storageUrl)
    # TODO:     called from messaging.DirectlyConnectedDevice.createStorageObject(url)
    # TODO:     called from virtualization.EnterpriseClient.createStorageObject(url)
    # TODO:     this method calls createStorageObject(storageContainerUrl, name, 
    # TODO:                              contentType, type_, encoding, date, length)
    # TODO: 
    # TODO: protected abstract createStorageObject(containerUrl, name, contentType,
    # TODO:                                        encoding, date, length)
    def createStorageObject(self, clientId=None, name=None, contentType=None,
                            storageUrl=None):
        """
        Create a :class:`.StorageObject`.
        
        If `storageUrl` is not `None`, `clientId`, `name`, and `contentType`
        are ignored.

        If any of `clientId`, `name`, or `contentType` are not `None`, 
        `storageUrl` is ignored.

        Information about the transfer of `storageObject` is obtained from a
        `HEAD` HTTP requrest and retried if `SecureConnectionBase._RETRY_LIMIT`
        times. See :mod:`.config` and the congifuration property in 
        *storage_connection_retry_limit* in section *media_storage*.

        :param clientId:
        :param name:
        :param contentType:
        :param storageUrl:

        :raises: :class:`.SecurityException` if `storageUrl` does not start
            with :meth:`container url <.getContainerUrl>`.
        :raises: :class:`.ArgumentException` if `storageUral` is `None` or
            empty or invalid
        """
        # TODO: What if name is None, JCL does not check.
        if clientId is not None or name is not None or contentType is not None:
            path = f'/{name}' if name is not None else ''
            if StorageConnectionBase._virtualStorageDirectories and \
                    clientId is not None:
                path = f'/{clientId}{path}'
            contentStorageLocation = f'{self.getStorageContainerUrl()}{path}'
            storageObject = self._createStorageObject(
                contentStorageLocation, name, contentType, None, None, -1)
            return storageObject
        else:
            if storageUrl is None or len(storageUrl) == 0:
                raise ArgumentException(
                    f'StorageConnectedBase.createStorageObject(): '
                    f'storageUrl must not be None or empty')

            # clientId, name, contentType are None
            url = None
            name = None
            try:
                # validate the url
                url = urlparse(storageUrl)
                fullContainerUrl = self.getStorageContainerUrl() + "/"
                if not storageUrl.startswith(fullContainerUrl):
                    raise SecurityException("Storage container URL does not match.")
                # if _virtualStorageDirectories is true use the complete path
                # to the file, everything after the "fullContainerUrl"
                # else just the "tail"
                # TODO: It's not clear if "name" is required to only ever be
                # TODO: "foo.jpg" and never "boo/baz/foo.jpg". Otherwise
                # TODO: this code (and JCL) will not recover the original name.
                if StorageConnectionBase._virtualStorageDirectories:
                    # just the tail
                    name = pathlib.Path(url.path).name
                else:
                    # the path less the container url.
                    name = storageUrl[len(fullContainerUrl):]
            except Exception as ex:
                raise ArgumentException(
                    f'StorageConnectionBase.createStorageObject(): '
                    f'Storage Cloud URL {storageUrl} is invalid, {ex}')
            i = 0
            while i < self._RETRY_LIMIT:
                headtuple = self.__head(storageUrl)
                length = headtuple.length
                contentType = headtuple.contentType
                encoding = headtuple.encoding
                date = headtuple.date
                responseCode = headtuple.status
                if responseCode == 200:
                    return self._createStorageObject(storageUrl, name,
                                                     contentType, encoding,
                                                     date, length)
                elif responseCode != 401:
                    raise TransportException(responseCode,
                        'StorageConnectionBase.createStorageObject(): HEAD failed')  # noqa: E501
                else:
                    self._authenticate()
                i += 1
            raise TransportException(401,
                f'StorageConnectionBase.createStorageObject(): '
                f'Exceeded retry limit {StorageConnectionBase._RETRY_LIMIT}')

    # TODO: protected abstract
    @abstractmethod
    def _createStorageObject(self, storageContainerUrl, name, contentType,
                               encoding, date, length):
        """
        Return an appropriate storage object delegate.

        There are two implementations, one in messaging which returns a
        :class:`.messaging.client.StorageObjectDelegate` and one for enterprise
        that returns `.enterprise.client.StorageObjectDelegate`.

        :param storageContainerUrl:
        :param name:
        :param contentType:
        :param encoding:
        :param date:
        :param length:
        :return: a storage object delegate
        """
        pass

    @classmethod
    def isStorageCloudURI(cls, uri):
        """
        Return `True` if `uri` contains `STORAGE_CLOUD_HOST`.

        See :mod:`.config` and property *storage_connection_host_name* in
        section *media_storage*

        :return: `True` if `uri` contains
            `StorageConnectionBase.STORAGE_CLOUD_HOST` else `False`
        """
        try:
            url = urlparse(uri)
            if uri.netloc.startswith(StorageConnectionBase.STORAGE_CLOUD_HOST):
                return True
        except Exception as ex:
            return False

    def close(self):
        """Close the connection."""
        # TODO: Where is this ever used ?
        with self.closedLock:
            if not self.closed:
                super(StorageConnectionBase, self).close()
                self.closed = True
                # TODO: encapsulate somewhere else
                self.http.clear()

    def getStorageContainerUrl(self):
        """
        Return the cloud storage service container url.

        The storage cloud service container url is obtained from the
        :class:`.StorageAuthenticationResponse` in the process of
        authenticating. An authentication is attempted if the connection
        had not authenticated previously.

        See :meth:`.authenticate`
        See :meth:`._getStorageAuthentication`

        :return: the cloud storage service container url
        """
        with self.authTokenLock:
            if self.authToken is None:
                self.authToken = self._getStorageAuthentication()

            return self.authToken.getStorageContainerUrl()


    def _transfer(self, storageObject):
        """
        Transfer `storageObject` to or from the cloud storage service.
        As many as `SecureConnectionBase._RETRY_LIMIT` retries are attempted.
        See :mod:`.config` and the congifuration property in 
        *storage_connection_retry_limit* in section *media_storage*.

        If the authorization token expires, an authentication attempt is made.
        See :meth:`.authenticate`
        
        :return: the HTTTP response code
        :raises: :class:`.TransportException` if an exception occurs
            transferring content.
        """
        responseCode = 0
        if self.authToken is None:
            if not self._authenticate():
                responseCode = 401
        inputStream = storageObject.getInputStream()
        upload = inputStream is not None
        # TODO: What is the python equivalent of "markSupported" ?
        # if upload and inputStream.markSupported():
        #     inputStream.mark(sys.maxsize)

        storageContainerUrl = self.getStorageContainerUrl()

        scPath = f'{storageContainerUrl}/storageObject.getName()' \
            if storageObject.getURI() is None else \
                storageObject.getURI()
        retries = 0
        while retries < self._RETRY_LIMIT:
            if self.authToken is not None:
                responseCode = self._transferContent(scPath, storageObject)
            if (upload and responseCode == 201) or \
                    (not upload and (responseCode == 200 or responseCode == 206)):
                break
            if responseCode == 401:
                self._authenticate()
            elif responseCode == self._BAD_CHECKSUM_STATUS:
                # try the transfer again
                pass
            else:
                # don't retry anymore
                break
            if upload:
                # # TODO: What is the python equivalent of markSupported and reset ?
                # if inputStream.markSupported():
                #     inputStream.reset()
                # # TODO: python FileInputStream equivalent
                # elif not (isinstance(inputStream, (FileInputStream, ))):
                #     return responseCode
                # else:
                #     # TODO: python "getChannel" equivalent ?
                #     fs = inputStream
                #     fc = fs.getChannel()
                #     fc.position(0)
                pass
            retries += 1
        return responseCode

    def _transferContent(self, scPath, storageObject):
        """
        Transfer :class:`storageObject <.StorageObject>` to the cloud storage service.
        
        Upload or download content to or from the cloud storage service.
        if :meth:`.StorageObject.getInputStream` returns `None` perform a download else
        perform an upload.

        :return: the HTTP status of the transfer.
        """
        contentType = storageObject.getContentType()
        encoding = storageObject.getEncoding()

        inputStream = storageObject.getInputStream()
        outputStream = storageObject.getOutputStream()
        upload = inputStream is not None
        responseCode = int()

        # TODO: Deal with exception do the right thing
        url = urlparse(scPath)

        if upload:
            responseCode = self.__upload(storageObject, scPath, inputStream,
                                         contentType, encoding)
        else:
            responseCode = self.__download(storageObject, scPath, outputStream)

        return responseCode

    def __getAuthToken(self):
        """Return the `authToken`."""
        with self.authTokenLock:
            if self.authToken is None:
                self.authToken = self._getStorageAuthentication()

            return self.authToken.getAuthToken()

    def _getStorageAuthentication(self):
        """
        Return the :class:`.StorageAuthenticationResponse` instance.
        
        :return: the :class:`.StorageAuthenticationResponse` instance.
        :raises: :class:`.NetworkException` for an invalid response
        :raises: :class:`.TransportException` if there is an HTTP error
        """
        response = \
            self.secureConnection.get(self._REST_STORAGE_AUTHENTICATION, None)
        status = response.getStatus()
        if status != 200:
            raise TransportException(status,
                response.getVerboseStatus("GET",
                                          self._REST_STORAGE_AUTHENTICATION))
        jsonResonse = None
        try:
            jsonResponse = json.loads(response.getData())
        except Exception as e:
            raise NetworkException(f'{e}')
        storageAuthenticationResponse = \
            StorageAuthenticationResponse.fromJson(jsonResponse)
        # FINEST
        self._getLogger().debug(storageAuthenticationResponse.__str__())
        return storageAuthenticationResponse

    def _authenticate(self):
        """
        Authenticate with the cloud storage service.
        
        :return `True` if authenticated else `False`
        """
        with self.authTokenLock:
            try:
                self.authToken = self._getStorageAuthentication()
                return True
            except Exception as e:
                self._getLogger().info(f'IoT storage API cannot be accessed: {e}')
        return False

    # TODO: handle exceptions
    def __download(self, storageObject, url, outputStream):

        digest = hashlib.md5()
        transferredBytes = 0
        headers = dict()
        headers[StorageConnectionBase._AUTH_TOKEN_HEADER] = \
            self.__getAuthToken()
        response = self.http.request('GET', url, headers=headers,
                                preload_content=False)
        chunk = 4096
        with open(outputStream, 'wb') as out:
            while True:
                if storageObject.isCancelled():
                    self._getLogger().debug(
                        f'StorageConnectionBase.__download(): '
                        f'download cancelled')
                    return StorageConnectionBase._TRANSFER_CANCELLED

                data = response.read(chunk)
                if not data:
                    break
                out.write(data)
                transferredBytes += len(data)
                storageObject.setTransferredBytes(transferredBytes)
                if digest is not None:
                    digest.update(data)

        status = response.status
        checksum = response.getheader("ETag")
        response.release_conn()

        if checksum is not None and digest is not None:
            digestOut = digest.digest().hex()
            # TODO: are the types the same ?
            if not checksum == digestOut:
                self._getLogger().debug(
                    f'Storage Cloud Service: download checksum mismatch')
                return StorageConnectionBase._BAD_CHECKSUM_STATUS

        if status == 200 or status == 206:
            for h,v in response.getheaders().items():
                if h.startswith("X-Object-Meta-"):
                    key = h[len("X-Object-Meta-"):]
                    # TODO: What is "v" a list ?
                    storageObject.setCustomMetadata(key, v)

        return status

    # TODO: handle exceptions
    def __upload(self, storageObject, url, inputStream, contentType, encoding):

        transferredBytes = 0
        contentType = contentType if contentType is not None else \
                "application/octet-stream"

        headers = dict()
        headers["Content-Type"] = contentType
        headers[StorageConnectionBase._AUTH_TOKEN_HEADER] = \
            self.__getAuthToken()
        if encoding is not None:
            headers["Content-Encoding"] = encoding

        # Before reading in the file, see if the transfer was cancelled
        if storageObject.isCancelled():
            return StorageConnectionBase._TRANSFER_CANCELLED

        for k,v in storageObject.getCustomMetadata().items():
            headers[f'X-Object-Meta-{k}'] = v

        data = None
        # TODO: Need streaming interface
        with inputStream.open(mode='rb') as uploadfile:
            data = uploadfile.read()

        # TODO: Should empty files be uploaded ?
        if data is None:
            self._getLogger().warning(f'StorageConnectionBase.__upload: '
                                     f' emptyfile')
            return
        transferredBytes = len(data)

        fields = dict()
        fields[inputStream.name] = (inputStream.name,
                                    data,
                                    contentType)
        # Check again
        if storageObject.isCancelled():
            return StorageConnectionBase._TRANSFER_CANCELLED

        # TODO: While the recommended way to upload, does not work.
        # TODO: transfers more bytes the in the original file.
        # try:
        #     response = self.http.request_encode_body('PUT', url, headers=headers,
        #                         fields=fields,
        #                         preload_content=False)
        # except Exception as e:
        #     # TODO: return an appropriate exception
        #     raise e

        try:
            response = self.http.request('PUT', url, headers=headers,
                                fields=None,
                                #preload_content=False,
                                body=data)
        except Exception as e:
            # TODO: return an appropriate exception
            raise e

        status = response.status
        checksum = response.getheader("ETag")
        date = response.getheader("Last-Modified")
        response.release_conn()

        if checksum is not None:
            digest = hashlib.md5()
            digest.update(data)
            if not checksum == digest.digest().hex():
                self._getLogger().debug(
                    f'Storage Cloud Service: upload checksum mismatch')
                return StorageConnectionBase._BAD_CHECKSUM_STATUS

        if status == 201:
            storageObject.setAttributes(date, transferredBytes)

        return status

    def __head(self, storageUrl):
        """
        Return `HEAD` request information for `storageUrl`

        :param storageUrl: the url of a storage cloud service object
        :return: a named tuple HeadTuple of
            *['status', 'date', 'length', 'contentType', 'encoding']*
        """
        headers = dict()
        headers[StorageConnectionBase._AUTH_TOKEN_HEADER] = \
            self.__getAuthToken()

        response = self.http.request('HEAD', storageUrl,
                                fields=None,
                                headers=headers)
        lastmodified = response.getheader('Last-Modified')
        length = int(response.getheader('Content-Length'))
        contenttype = response.getheader('Content-Type')
        encoding = response.getheader('Content-Encoding')
        status = response.status
        response.release_conn()

        HeadTuple = collections.namedtuple('HeadTuple',
                                           ['status', 'date', 'length',
                                            'contentType', 'encoding'])

        return HeadTuple(status=status, date=lastmodified, length=length,
                         contentType=contenttype, encoding=encoding)

    def _getLogger(self):
        """Return a `StorageConnectionBase` logger."""
        return logging.getLogger("StorageConnectionBase")

# package: com.oracle.iot.client


class StorageObject(ExternalObject):
    """StorageObject provides information about content in cloud storage."""

    _DEFAULT_TTL = Util.getConfig()["media_storage"].getint("storage_ttl")

    __slots__ = (
        'name',
        'mimetype',
        'encoding',
        'date',
        'length',
        'inputStream',
        'outputStream',
        'metadata'
    )

    def __init__(self, uri, name, mimetype, encoding, date, length):
        """
        Create a :class:`.StorageObject` instance.

        :param uri: the full URL of the object in the Storage Cloud
        :param name: the object name used in the Storage Cloud
        :param mimetype: the object's content type or `None`
        :param encoding: the object's encoding or `None`
        :param date: the the object's last modified date
        :param length: the object's length in bytes
        """
        super(StorageObject, self).__init__(uri)
        self.name = name
        self.mimetype = mimetype if mimetype is not None else \
            "application/octet-stream"
        self.encoding = encoding
        self.date = date
        self.length = length
        self.metadata = dict()

    def getName(self):
        """
        Return the the name of this object in the storage cloud.

        It is the name and path of the file that was uploaded to the
        storage cloud.

        :return the name of this object in the storage cloud
        """
        return self.name

    def getContentType(self):
        """
        Return the mime type of the content.

        See `IANA Media Types <http://www.iana.org/assignments/media-types/media-types.xhtml>`_  # noqa: E501
        :return: The mime-type of the content
        """
        return self.mimetype

    def getEncoding(self):
        """
        Return the compression scheme of the content.

        :return: the compression scheme of the content, or `None` if
            the content is not compressed
        """
        return self.encoding

    def getLength(self):
        """
        Return the length of the content in bytes.

        This is the number of bytes required to upload or download the content.

        :return: The length of the content, or -1 if unknown
        """
        return self.length

    def getDate(self):
        """
        Return the date and time the content was created or last modified.

        Return the date and time the content was created or last modified in
        the cloud storage in ISO 8601 format. This may be `None` if the
        content has not been uploaded.

        :return: The date the content was last modified in cloud storage
            in ISO 8601 format, or `None` if the content has not been
            uploaded.
        """
        return self.date

    def getInputStream(self):
        """
        Called by the library to obtain the input stream when uploading content.

        :return: The input stream for the content being uploaded
        """
        return self.inputStream

    def setInputStream(self, inputStream):
        """
        Set an input stream for content to be uploaded.

        The implementation allows for either the input stream to be set, or the
        output stream to be set, but not both. If`inputStream` parameter is not
        `None` the output stream will be set to `None`.

        :param inputStream: A stream from which the content will be read
        :raises StateException: if a transfer is currently in progress
        """
        self.inputStream = inputStream
        if inputStream != None:
            self.outputStream = None

    def getOutputStream(self):
        """
        Return `outputStream` for downloaded content.

        Called by the library to obtain the output stream when downloading
        content
        :return: The output stream for the content being downloaded
        """
        return self.outputStream

    def setOutputStream(self, outputStream):
        """
        Set an output stream downloaded content.

        The implementation allows for either the input stream to be set, or the
        output stream to be set, but not both. If `outputStream` is not `None`
        the input stream will be set to `None`.

        :param outputStream: A stream to which the content will be written
        :raises StateException: if a transfer is currently in progress
        """
        self.outputStream = outputStream
        if outputStream != None:
            self.inputStream = None

    def setCustomMetadata(self, key, value):
        """
        Sets the metadata for the StorageObject.

        All metdata will be added to Storage Cloud Service as custom metadata
        with the *X-Object-Meta-KeyName* header.
        :param key: The metadata key
        :param value: The metadata value
        :raises: :class:`.Argument` if the key or value is `None` or empty
        """
        if (key is None or key is not None and len(key) == 0) or \
                (value is None or value is not None and len(value) == 0):
            raise ArgumentException(f'Neither key or value can be None or empty')

        self.metadata[key] = value

    def getCustomMetadata(self):
        """
        This method return unmodifiable copy of metadata.
        :return: a `dict` of metadata. It may be empty.
        """
        return self.metadata

    # TODO: confirm exceptions
    @abstractmethod
    def sync(self):
        """
        Synchronize content with the Storage Cloud Service.

        :raises IOError: if there is an I/O exception raised by the runtime,
        or a failure reported by the storage cloud
        :raises SecurityException: if there is an exception establishing a
        secure connection to the storage cloud
        :raises StateException: if the `StorageObject` does not have an
        input stream or output stream set.
        """
        pass

    def __eq__(self, obj):
        """Return `True` if obj equals this `StorageObject`."""
        if id(self) == id(obj):
            return True
        if obj == None:
            return False
        if not isinstance(obj, StorageObject):
            return False
        other = obj
        if not self.getURI() == other.getURI():
            return False
        if not self.getName() == other.getName():
            return False
        if not self.getContentType() == other.getContentType():
            return False
        return True

    def __hash__(self):
        """Return a hash for this `StorageObject` instance."""
        return hash((self.name, self.mimetype,
                     self.getURI()))

    def setAttributes(self, date, length):
        """
        Set meta data `date`, and `length`.

        :param date: the last modified date
        :param length: the object's length
        """
        self.date = date
        self.length = length

# com.oracle.iot.client.impl


class StorageObjectDelegate(StorageObject):
    """
    `StorageObjectDelegate` provides a storage connection and number of transferred bytes.  # noqa: E501

    See :class:`.messaging.client.StorageObject`
    See :class:`.messaging.client.StorageConnection`
    """
    __slots__ = (
        'storageConnection',
        # a :class:`.StorageConnection` to the cloud storage service
        'transferredBytes'
        # the number bytes transferred
    )

    def __init__(self, storageConnection, uri, name, contentType,
                 contentEncoding, dateOfLastModification, length):
        """
        Create a `StorageObjectDelegate` instance.

        :param storageConnection: the cloud storage service connection
        :param uri: the full URL of the object in the Storage Cloud
        :param name: the object name used in the Storage Cloud
        :param mimetype: the object's content type or `None`
        :param encoding: the object's encoding or `None`
        :param date: the the object's last modified date
        :param length: the object's length in bytes
        """
        super(StorageObjectDelegate, self).__init__(uri, name, contentType,
                                                    contentEncoding,
                                                    dateOfLastModification,
                                                    length)
        self.storageConnection = storageConnection
        self.transferredBytes = 0

    def setAttributes(self, date, length):
        """
        Set meta data `date` and `length`.

        :param date: the last modified date
        :param length: the number of transferred bytes
        """
        super(StorageObjectDelegate, self).setAttributes(date, length)

    # TODO: Confirm exceptions
    def sync(self):
        """
        Synchronize content with the Storage Cloud Service.

        :raises IOError: if there is an I/O exception raised by the runtime,
        or a failure reported by the storage cloud
        :raises SecurityException: if there is an exception establishing a
        secure connection to the storage cloud
        :raises StateException: if the `StorageObject` does not have an
        input stream or output stream set.
        """
        self.storageConnection.sync(self)

    def isCancelled(self):
        """Cancel any pending transfer."""
        #  TODO: this is overridden on the DC side, but the EC side cannot
        #  TODO: cancel. IOT-42145
        return False

    def setTransferredBytes(self, transferredBytes):
        """
        Set the number of bytes transferred.

        :param transferredBytes: the number of bytes transferred
        """
        self.transferredBytes = transferredBytes

