# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC, abstractmethod
from copy import deepcopy
import threading

from iotcs.exception import ArgumentException

# com.oracle.iot.client.impl.device

class PersistenceStore(ABC):
    """ PersistenceStore. """

    class Transaction(ABC):
        """
        Interface for modifying values in a :class:`.PersistenceStore`.
        
        The :class:`.PersistenceStore` itself is not updated until :meth:`.commit()` is called.
        """
        @abstractmethod
        def clear(self):
            """
            Mark all values to be removed from the :class:`.PersistenceStore` object.

            When :meth:`.commit` is called, values are removed before put methods
            are processed.
            
            :return this :class:`.Transaction` object
            """

        @abstractmethod
        def commit(self):
            """
            Commit this transaction.
            
            This method persists the values to the backing store and
            replaces the values in the PersistenceStore} object.
            :return: `True` if the values were persisted
            """

        @abstractmethod
        def put(self, key, value):
            """
            Set a value for the `key`.
            
            `key is written back to the :class:`.PersistenceStore` object
            when :meth:`.commit()` is called.
            :param key: a key to be used to retrieve the value
            :param value: the value
            :return: this :class:`.Transaction` object
            """

        #  TODO: value should be Serializable or some such.
        @abstractmethod
        def remove(self, key):
            """
            Mark all values to be removed from the :class:`.PersistenceStore` object.

            When :meth:`.commit` is called, values are removed before put methods
            are processed.
            
            :param key: a key whose value is to be removed
            :return: this :class:`.Transaction` object
            """

    @abstractmethod
    def getName(self):
        """
        Return the name used by the :class:`.PersistenceManager` to reference this :class:`.PersistenceStore`.
        :return The PersistenceStore} name.
        """

    @abstractmethod
    def openTransaction(self):
        """
        Open a :class:`.Transaction` object for modifying the values of this :class:`.PersistenceStore`.
        :return: a :class:`.Transaction` object
        """

    @abstractmethod
    def contains(self, key):
        """
        Return `True` if this :class:`.PersistenceStore` contains the given key.

        :param key: the key to search for
        :return: `True` if this :class:`.PersistenceStore` contains the given key
        """

    @abstractmethod
    def getAll(self):
        """
        Return a map of all key/value pairs in this :class:`.PersistenceStore`
        """

    @abstractmethod
    def get(self, key, defaultValue=None):
        """
        Return a value for the given `key`.
        :param key: the key to search for
        :param defaultValue: the value to use if this :class:`.PersistenceStore` does
            not contain the key
        :return: the value for the key
        """

    @abstractmethod
    def getLock(self):
        """Return the lock for this `PersistenceStore`."""


#  com.oracle.iot.client.impl.device

class PersistenceStoreManager(object):
    """PersistenceStoreManager."""  

    class InMemoryPersistenceStore(PersistenceStore):
        """An in memory :class:`.PersistenceStore`."""

        __slots__ = (
            'map',
            'maplock',
            'name'
        )
        
        def __init__(self, name):
            """Create an instance of `InMemoryPersistenceStore`."""
            super(PersistenceStoreManager.InMemoryPersistenceStore, self).__init__()
            self.name = name
            self.map = dict()
            self.maplock = threading.RLock()

        def getName(self):
            """Return the `name` of this :class:`.PersistenceStore`."""
            return self.name

        def openTransaction(self):
            """Open a transaction in this :class:`.PersistenceStore`."""
            return PersistenceStoreManager.InMemoryPersistenceStore.Transaction(self)

        def contains(self, key):
            """Return `True` if `key` is in this :class:`.PersistenceStore`."""
            return key in self.map

        def getAll(self):
            """Return a copy of the map of persisted objects."""
            with self.maplock:
                return deepcopy(self.map)

        def get(self, key, defaultValue=None):
            """Return the value for `key` or `defaultValue` from this :class:`.PersistenceStore`."""
            value = defaultValue
            with self.maplock:
                if key in self.map:
                    value = self.map[key]
            return value

        def getLock(self):
            """Return the lock for this `PersistenceStore`."""
            return self.maplock

        class Transaction(PersistenceStore.Transaction):
            """A `Transaction` processing class."""

            __slots__ = (
                'map',
                'maplock',
                'peristencestore'
            )
            def __init__(self, persistencestore):
                """
                Creat a Transaction instance.

                :param persistencestore: open a transaction on this :class:`.PersistenceStore`
                :raise: :class:`.Argument` if `persistencestore` is `None`
                """
                super(PersistenceStoreManager.InMemoryPersistenceStore.Transaction, self).__init__()
                if persistencestore is None:
                    raise ArgumentException(f'Transaction.__init__(): '
                        f' cannot open a transaction without a PersistenceStore instance')
                self.map = dict()
                self.maplock = threading.RLock()
                self.persistencestore = persistencestore
                
            def clear(self):
                """
                Clear all entries in the transaction.
                :return: this instance
                """
                with self.maplock:
                    self.map.clear()
                return self

            def commit(self):
                """
                Commit the transaction.
                :return: True
                """
                with self.maplock:
                    with self.persistencestore.maplock:
                        self.persistencestore.map.update(self.map)

                return True

            def put(self, key, value):
                """
                Add `key` and `value` to the current transaction.
                :return: this instance
                """
                with self.maplock:
                    self.map[key] = value
                return self

            def remove(self, key):
                """Remove `key` from the transaction."""
                with self.maplock:
                    if key in self.map:
                        del(self.map[key])
                return self

    _map = dict()
    _maplock = threading.RLock()

    @classmethod
    def getPersistenceStore(cls, name):
        """Return the :class:`.PersistenceStore` with `name`."""

        persistenceStore = None
        with cls._maplock:
            if name in cls._map:
                persistenceStore = cls._map[name]

        if persistenceStore is None:
            persistenceStore = \
                PersistenceStoreManager.InMemoryPersistenceStore(name)
        with cls._maplock:
            cls._map[name] = persistenceStore

        return persistenceStore
